@@ -815,7 +815,8 @@ def MigrateInstance(opts, args):
                                  target_node=target_node,
                                  allow_failover=opts.allow_failover,
                                  allow_runtime_changes=opts.allow_runtime_chgs,
-                                 ignore_ipolicy=opts.ignore_ipolicy)
+                                 ignore_ipolicy=opts.ignore_ipolicy,
+                                 ignore_hvversions=opts.ignore_hvversions)
   SubmitOrSend(op, cl=cl, opts=opts)
   return 0
 
@@ -1257,37 +1258,52 @@ def _ConvertNicDiskModifications(mods):
       # Add item as last item (legacy interface)
       action = constants.DDM_ADD
       identifier = -1
+    elif identifier == constants.DDM_ATTACH:
+      # Attach item as last item (legacy interface)
+      action = constants.DDM_ATTACH
+      identifier = -1
     elif identifier == constants.DDM_REMOVE:
       # Remove last item (legacy interface)
       action = constants.DDM_REMOVE
       identifier = -1
+    elif identifier == constants.DDM_DETACH:
+      # Detach last item (legacy interface)
+      action = constants.DDM_DETACH
+      identifier = -1
     else:
-      # Modifications and adding/removing at arbitrary indices
+      # Modifications and adding/attaching/removing/detaching at arbitrary
+      # indices
       add = params.pop(constants.DDM_ADD, _MISSING)
+      attach = params.pop(constants.DDM_ATTACH, _MISSING)
       remove = params.pop(constants.DDM_REMOVE, _MISSING)
+      detach = params.pop(constants.DDM_DETACH, _MISSING)
       modify = params.pop(constants.DDM_MODIFY, _MISSING)
 
-      if modify is _MISSING:
-        if not (add is _MISSING or remove is _MISSING):
-          raise errors.OpPrereqError("Cannot add and remove at the same time",
-                                     errors.ECODE_INVAL)
-        elif add is not _MISSING:
-          action = constants.DDM_ADD
-        elif remove is not _MISSING:
-          action = constants.DDM_REMOVE
-        else:
-          action = constants.DDM_MODIFY
-
-      elif add is _MISSING and remove is _MISSING:
-        action = constants.DDM_MODIFY
-      else:
-        raise errors.OpPrereqError("Cannot modify and add/remove at the"
-                                   " same time", errors.ECODE_INVAL)
+      # Check if the user has requested more than one operation and raise an
+      # exception. If no operations have been given, default to modify.
+      action = constants.DDM_MODIFY
+      ops = {
+        constants.DDM_ADD: add,
+        constants.DDM_ATTACH: attach,
+        constants.DDM_REMOVE: remove,
+        constants.DDM_DETACH: detach,
+        constants.DDM_MODIFY: modify,
+      }
+      count = 0
+      for op, param in ops.items():
+        if param is not _MISSING:
+          count += 1
+          action = op
+      if count > 1:
+        raise errors.OpPrereqError(
+          "Cannot do more than one of the following operations at the"
+          " same time: %s" % ", ".join(ops.keys()),
+          errors.ECODE_INVAL)
 
       assert not (constants.DDMS_VALUES_WITH_MODIFY & set(params.keys()))
 
-    if action == constants.DDM_REMOVE and params:
-      raise errors.OpPrereqError("Not accepting parameters on removal",
+    if action in (constants.DDM_REMOVE, constants.DDM_DETACH) and params:
+      raise errors.OpPrereqError("Not accepting parameters on removal/detach",
                                  errors.ECODE_INVAL)
 
     result.append((action, identifier, params))
@@ -1295,6 +1311,27 @@ def _ConvertNicDiskModifications(mods):
   return result
 
 
+def _ParseExtStorageParams(params):
+  """Parses the disk params for ExtStorage conversions.
+
+  """
+  if params:
+    if constants.IDISK_PROVIDER not in params:
+      raise errors.OpPrereqError("Missing required parameter '%s' when"
+                                 " converting to an ExtStorage disk template" %
+                                 constants.IDISK_PROVIDER, errors.ECODE_INVAL)
+    else:
+      for param in params.keys():
+        if (param != constants.IDISK_PROVIDER and
+            param in constants.IDISK_PARAMS):
+          raise errors.OpPrereqError("Invalid parameter '%s' when converting"
+                                     " to an ExtStorage template (it is not"
+                                     " allowed modifying existing disk"
+                                     " parameters)" % param, errors.ECODE_INVAL)
+
+  return params
+
+
 def _ParseDiskSizes(mods):
   """Parses disk sizes in parameters.
 
@@ -1360,13 +1397,33 @@ def SetInstanceParams(opts, args):
 
   disks = _ParseDiskSizes(_ConvertNicDiskModifications(opts.disks))
 
-  if (opts.disk_template and
-      opts.disk_template in constants.DTS_INT_MIRROR and
-      not opts.node):
-    ToStderr("Changing the disk template to a mirrored one requires"
-             " specifying a secondary node")
+  # verify the user provided parameters for disk template conversions
+  if opts.disk_template:
+    if (not opts.node and
+        opts.disk_template in constants.DTS_INT_MIRROR):
+      ToStderr("Changing the disk template to a mirrored one requires"
+               " specifying a secondary node")
+      return 1
+    elif (opts.ext_params and
+          opts.disk_template != constants.DT_EXT):
+      ToStderr("Specifying ExtStorage parameters requires converting"
+               " to the '%s' disk template" % constants.DT_EXT)
+      return 1
+    elif (not opts.ext_params and
+          opts.disk_template == constants.DT_EXT):
+      ToStderr("Provider option is missing, use either the"
+               " '--ext-params' or '-e' option")
+      return 1
+
+  if ((opts.file_driver or
+       opts.file_storage_dir) and
+      not opts.disk_template in constants.DTS_FILEBASED):
+    ToStderr("Specifying file-based configuration arguments requires"
+             " converting to a file-based disk template")
     return 1
 
+  ext_params = _ParseExtStorageParams(opts.ext_params)
+
   if opts.offline_inst:
     offline = True
   elif opts.online_inst:
@@ -1382,6 +1439,9 @@ def SetInstanceParams(opts, args):
                                    hotplug=opts.hotplug,
                                    hotplug_if_possible=opts.hotplug_if_possible,
                                    disk_template=opts.disk_template,
+                                   ext_params=ext_params,
+                                   file_driver=opts.file_driver,
+                                   file_storage_dir=opts.file_storage_dir,
                                    remote_node=opts.node,
                                    pnode=opts.new_primary_node,
                                    hvparams=opts.hvparams,
@@ -1497,6 +1557,8 @@ m_inst_tags_opt = cli_option("--tags", dest="multi_mode",
 
 # this is defined separately due to readability only
 add_opts = [
+  FORTHCOMING_OPT,
+  COMMIT_OPT,
   NOSTART_OPT,
   OS_OPT,
   FORCE_VARIANT_OPT,
@@ -1536,7 +1598,8 @@ commands = {
     MigrateInstance, ARGS_ONE_INSTANCE,
     [FORCE_OPT, NONLIVE_OPT, MIGRATION_MODE_OPT, CLEANUP_OPT, DRY_RUN_OPT,
      PRIORITY_OPT, DST_NODE_OPT, IALLOCATOR_OPT, ALLOW_FAILOVER_OPT,
-     IGNORE_IPOLICY_OPT, NORUNTIME_CHGS_OPT] + SUBMIT_OPTS,
+     IGNORE_IPOLICY_OPT, IGNORE_HVVERSIONS_OPT, NORUNTIME_CHGS_OPT]
+    + SUBMIT_OPTS,
     "[-f] <instance>", "Migrate instance to its secondary node"
     " (only for mirrored instances)"),
   "move": (
@@ -1600,7 +1663,8 @@ commands = {
      OSPARAMS_OPT, OSPARAMS_PRIVATE_OPT, DRY_RUN_OPT, PRIORITY_OPT, NWSYNC_OPT,
      OFFLINE_INST_OPT, ONLINE_INST_OPT, IGNORE_IPOLICY_OPT, RUNTIME_MEM_OPT,
      NOCONFLICTSCHECK_OPT, NEW_PRIMARY_OPT, HOTPLUG_OPT,
-     HOTPLUG_IF_POSSIBLE_OPT, INSTANCE_COMMUNICATION_OPT],
+     HOTPLUG_IF_POSSIBLE_OPT, INSTANCE_COMMUNICATION_OPT,
+     EXT_PARAMS_OPT, FILESTORE_DRIVER_OPT, FILESTORE_DIR_OPT],
     "<instance>", "Alters the parameters of an instance"),
   "shutdown": (
     GenericManyOps("shutdown", _ShutdownInstance), [ArgInstance()],
