@@ -167,7 +167,7 @@ def _TryReadFile(path):
 
 
 def _ReadSshKeys(keyfiles, _tostderr_fn=ToStderr):
-  """Reads SSH keys according to C{keyfiles}.
+  """Reads the DSA SSH keys according to C{keyfiles}.
 
   @type keyfiles: dict
   @param keyfiles: Dictionary with keys of L{constants.SSHK_ALL} and two-values
@@ -186,13 +186,13 @@ def _ReadSshKeys(keyfiles, _tostderr_fn=ToStderr):
     if public_key and private_key:
       result.append((kind, private_key, public_key))
     elif public_key or private_key:
-      _tostderr_fn("Couldn't find a complete set of keys for kind '%s'; files"
-                   " '%s' and '%s'", kind, private_file, public_file)
+      _tostderr_fn("Couldn't find a complete set of keys for kind '%s';"
+                   " files '%s' and '%s'", kind, private_file, public_file)
 
   return result
 
 
-def _SetupSSH(options, cluster_name, node, ssh_port):
+def _SetupSSH(options, cluster_name, node, ssh_port, cl):
   """Configures a destination node's SSH daemon.
 
   @param options: Command line options
@@ -202,8 +202,17 @@ def _SetupSSH(options, cluster_name, node, ssh_port):
   @param node: Destination node name
   @type ssh_port: int
   @param ssh_port: Destination node ssh port
+  @param cl: luxi client
 
   """
+  # Retrieve the list of master and master candidates
+  candidate_filter = ["|", ["=", "role", "M"], ["=", "role", "C"]]
+  result = cl.Query(constants.QR_NODE, ["uuid"], candidate_filter)
+  if len(result.data) < 1:
+    raise errors.OpPrereqError("No master or master candidate node is found.")
+  candidates = [uuid for ((_, uuid),) in result.data]
+  candidate_keys = ssh.QueryPubKeyFile(candidates)
+
   if options.force_join:
     ToStderr("The \"--force-join\" option is no longer supported and will be"
              " ignored.")
@@ -213,7 +222,10 @@ def _SetupSSH(options, cluster_name, node, ssh_port):
   (_, root_keyfiles) = \
     ssh.GetAllUserFiles(constants.SSH_LOGIN_USER, mkdir=False, dircheck=False)
 
-  root_keys = _ReadSshKeys(root_keyfiles)
+  dsa_root_keyfiles = dict((kind, value) for (kind, value)
+                           in root_keyfiles.items()
+                           if kind == constants.SSHK_DSA)
+  root_keys = _ReadSshKeys(dsa_root_keyfiles)
 
   (_, cert_pem) = \
     utils.ExtractX509Certificate(utils.ReadFile(pathutils.NODED_CERT_FILE))
@@ -223,12 +235,23 @@ def _SetupSSH(options, cluster_name, node, ssh_port):
     constants.SSHS_NODE_DAEMON_CERTIFICATE: cert_pem,
     constants.SSHS_SSH_HOST_KEY: host_keys,
     constants.SSHS_SSH_ROOT_KEY: root_keys,
+    constants.SSHS_SSH_AUTHORIZED_KEYS: candidate_keys,
     }
 
-  bootstrap.RunNodeSetupCmd(cluster_name, node, pathutils.PREPARE_NODE_JOIN,
-                            options.debug, options.verbose, False,
-                            options.ssh_key_check, options.ssh_key_check,
-                            ssh_port, data)
+  ssh.RunSshCmdWithStdin(cluster_name, node, pathutils.PREPARE_NODE_JOIN,
+                         ssh_port, data,
+                         debug=options.debug, verbose=options.verbose,
+                         use_cluster_key=False, ask_key=options.ssh_key_check,
+                         strict_host_check=options.ssh_key_check)
+
+  (_, dsa_pub_keyfile) = root_keyfiles[constants.SSHK_DSA]
+  pub_key = ssh.ReadRemoteSshPubKeys(dsa_pub_keyfile, node, cluster_name,
+                                     ssh_port, options.ssh_key_check,
+                                     options.ssh_key_check)
+  # Unfortunately, we have to add the key with the node name rather than
+  # the node's UUID here, because at this point, we do not have a UUID yet.
+  # The entry will be corrected in noded later.
+  ssh.AddPublicKey(node, pub_key)
 
 
 @UsesRPC
@@ -289,16 +312,21 @@ def AddNode(opts, args):
   # read the cluster name from the master
   (cluster_name, ) = cl.QueryConfigValues(["cluster_name"])
 
-  if not readd and opts.node_setup:
+  if opts.node_setup:
     ToStderr("-- WARNING -- \n"
-             "Performing this operation is going to replace the ssh daemon"
-             " keypair\n"
-             "on the target machine (%s) with the ones of the"
-             " current one\n"
-             "and grant full intra-cluster ssh root access to/from it\n", node)
+             "Performing this operation is going to perform the following\n"
+             "changes to the target machine (%s) and the current cluster\n"
+             "nodes:\n"
+             "* A new SSH daemon key pair is generated is generated on\n"
+             "  the target machine.\n"
+             "* The public SSH keys of all master candidates of the cluster\n"
+             "  are added to the target machine's 'authorized_keys' file.\n"
+             "* In case the target machine is a master candidate, its newly\n"
+             "  generated public SSH key will be distributed to all other\n"
+             "  cluster nodes.\n", node)
 
   if opts.node_setup:
-    _SetupSSH(opts, cluster_name, node, ssh_port)
+    _SetupSSH(opts, cluster_name, node, ssh_port, cl)
 
   bootstrap.SetupNodeDaemon(opts, cluster_name, node, ssh_port)
 
@@ -314,7 +342,8 @@ def AddNode(opts, args):
                          vm_capable=opts.vm_capable, ndparams=opts.ndparams,
                          master_capable=opts.master_capable,
                          disk_state=disk_state,
-                         hv_state=hv_state)
+                         hv_state=hv_state,
+                         node_setup=opts.node_setup)
   SubmitOpCode(op, opts=opts)
 
 
@@ -416,7 +445,8 @@ def EvacuateNode(opts, args):
   op = opcodes.OpNodeEvacuate(node_name=args[0], mode=mode,
                               remote_node=opts.dst_node,
                               iallocator=opts.iallocator,
-                              early_release=opts.early_release)
+                              early_release=opts.early_release,
+                              ignore_soft_errors=opts.ignore_soft_errors)
   result = SubmitOrSend(op, opts, cl=cl)
 
   # Keep track of submitted jobs
@@ -1033,11 +1063,6 @@ def ListDrbd(opts, args):
     ToStderr("Please give one (and only one) node.")
     return constants.EXIT_FAILURE
 
-  if not constants.ENABLE_CONFD:
-    ToStderr("Error: this command requires confd support, but it has not"
-             " been enabled at build time.")
-    return constants.EXIT_FAILURE
-
   status = ReplyStatus()
 
   def ListDrbdConfdCallback(reply):
@@ -1113,8 +1138,9 @@ commands = {
     "Add a node to the cluster"),
   "evacuate": (
     EvacuateNode, ARGS_ONE_NODE,
-    [FORCE_OPT, IALLOCATOR_OPT, NEW_SECONDARY_OPT, EARLY_RELEASE_OPT,
-     PRIORITY_OPT, PRIMARY_ONLY_OPT, SECONDARY_ONLY_OPT] + SUBMIT_OPTS,
+    [FORCE_OPT, IALLOCATOR_OPT, IGNORE_SOFT_ERRORS_OPT, NEW_SECONDARY_OPT,
+     EARLY_RELEASE_OPT, PRIORITY_OPT, PRIMARY_ONLY_OPT, SECONDARY_ONLY_OPT]
+    + SUBMIT_OPTS,
     "[-f] {-I <iallocator> | -n <dst>} [-p | -s] [options...] <node>",
     "Relocate the primary and/or secondary instances from a node"),
   "failover": (
