@@ -549,8 +549,7 @@ def _ComputeMinMaxSpec(name, qualifier, ispecs, value):
 
 def ComputeIPolicySpecViolation(ipolicy, mem_size, cpu_count, disk_count,
                                 nic_count, disk_sizes, spindle_use,
-                                disk_template,
-                                _compute_fn=_ComputeMinMaxSpec):
+                                disk_types, _compute_fn=_ComputeMinMaxSpec):
   """Verifies ipolicy against provided specs.
 
   @type ipolicy: dict
@@ -567,13 +566,15 @@ def ComputeIPolicySpecViolation(ipolicy, mem_size, cpu_count, disk_count,
   @param disk_sizes: Disk sizes of used disk (len must match C{disk_count})
   @type spindle_use: int
   @param spindle_use: The number of spindles this instance uses
-  @type disk_template: string
-  @param disk_template: The disk template of the instance
+  @type disk_types: list of strings
+  @param disk_types: The disk template of the instance
   @param _compute_fn: The compute function (unittest only)
   @return: A list of violations, or an empty list of no violations are found
 
   """
   assert disk_count == len(disk_sizes)
+  assert isinstance(disk_types, list)
+  assert disk_count == len(disk_types)
 
   test_settings = [
     (constants.ISPEC_MEM_SIZE, "", mem_size),
@@ -582,14 +583,20 @@ def ComputeIPolicySpecViolation(ipolicy, mem_size, cpu_count, disk_count,
     (constants.ISPEC_SPINDLE_USE, "", spindle_use),
     ] + [(constants.ISPEC_DISK_SIZE, str(idx), d)
          for idx, d in enumerate(disk_sizes)]
-  if disk_template != constants.DT_DISKLESS:
+
+  allowed_dts = set(ipolicy[constants.IPOLICY_DTS])
+  ret = []
+  if disk_count != 0:
     # This check doesn't make sense for diskless instances
     test_settings.append((constants.ISPEC_DISK_COUNT, "", disk_count))
-  ret = []
-  allowed_dts = ipolicy[constants.IPOLICY_DTS]
-  if disk_template not in allowed_dts:
+  elif constants.DT_DISKLESS not in allowed_dts:
+    ret.append("Disk template %s is not allowed (allowed templates %s)" %
+                (constants.DT_DISKLESS, utils.CommaJoin(allowed_dts)))
+
+  forbidden_dts = set(disk_types) - allowed_dts
+  if forbidden_dts:
     ret.append("Disk template %s is not allowed (allowed templates: %s)" %
-               (disk_template, utils.CommaJoin(allowed_dts)))
+               (utils.CommaJoin(forbidden_dts), utils.CommaJoin(allowed_dts)))
 
   min_errs = None
   for minmax in ipolicy[constants.ISPECS_MINMAX]:
@@ -602,8 +609,7 @@ def ComputeIPolicySpecViolation(ipolicy, mem_size, cpu_count, disk_count,
   return ret + min_errs
 
 
-def ComputeIPolicyDiskSizesViolation(ipolicy, disk_sizes,
-                                     disk_template,
+def ComputeIPolicyDiskSizesViolation(ipolicy, disk_sizes, disks,
                                      _compute_fn=_ComputeMinMaxSpec):
   """Verifies ipolicy against provided disk sizes.
 
@@ -614,18 +620,21 @@ def ComputeIPolicyDiskSizesViolation(ipolicy, disk_sizes,
   @param ipolicy: The ipolicy
   @type disk_sizes: list of ints
   @param disk_sizes: Disk sizes of used disk (len must match C{disk_count})
-  @type disk_template: string
-  @param disk_template: The disk template of the instance
+  @type disks: list of L{Disk}
+  @param disks: The Disk objects of the instance
   @param _compute_fn: The compute function (unittest only)
   @return: A list of violations, or an empty list of no violations are found
 
   """
+  if len(disk_sizes) != len(disks):
+    return [constants.ISPEC_DISK_COUNT]
+  dev_types = [d.dev_type for d in disks]
   return ComputeIPolicySpecViolation(ipolicy,
                                      # mem_size, cpu_count, disk_count
                                      None, None, len(disk_sizes),
                                      None, disk_sizes, # nic_count, disk_sizes
                                      None, # spindle_use
-                                     disk_template,
+                                     dev_types,
                                      _compute_fn=_compute_fn)
 
 
@@ -665,10 +674,10 @@ def ComputeIPolicyInstanceViolation(ipolicy, instance, cfg,
   disk_count = len(disks)
   disk_sizes = [disk.size for disk in disks]
   nic_count = len(instance.nics)
-  disk_template = instance.disk_template
+  disk_types = [d.dev_type for d in disks]
 
   return ret + _compute_fn(ipolicy, mem_size, cpu_count, disk_count, nic_count,
-                           disk_sizes, spindle_use, disk_template)
+                           disk_sizes, spindle_use, disk_types)
 
 
 def _ComputeViolatingInstances(ipolicy, instances, cfg):
@@ -1239,9 +1248,10 @@ def CheckIpolicyVsDiskTemplates(ipolicy, enabled_disk_templates):
   allowed_disk_templates = ipolicy[constants.IPOLICY_DTS]
   not_enabled = set(allowed_disk_templates) - set(enabled_disk_templates)
   if not_enabled:
-    raise errors.OpPrereqError("The following disk template are allowed"
+    raise errors.OpPrereqError("The following disk templates are allowed"
                                " by the ipolicy, but not enabled on the"
-                               " cluster: %s" % utils.CommaJoin(not_enabled))
+                               " cluster: %s" % utils.CommaJoin(not_enabled),
+                               errors.ECODE_INVAL)
 
 
 def CheckDiskAccessModeValidity(parameters):
@@ -1293,20 +1303,21 @@ def CheckDiskAccessModeConsistency(parameters, cfg, group=None):
 
     for entry in inst_uuids:
       inst = cfg.GetInstanceInfo(entry)
-      inst_template = inst.disk_template
+      disks = cfg.GetInstanceDisks(entry)
+      for disk in disks:
 
-      if inst_template != disk_template:
-        continue
+        if disk.dev_type != disk_template:
+          continue
 
-      hv = inst.hypervisor
+        hv = inst.hypervisor
 
-      if not IsValidDiskAccessModeCombination(hv, inst_template, access):
-        raise errors.OpPrereqError("Instance {i}: cannot use '{a}' access"
-                                   " setting with {h} hypervisor and {d} disk"
-                                   " type.".format(i=inst.name,
-                                                   a=access,
-                                                   h=hv,
-                                                   d=inst_template))
+        if not IsValidDiskAccessModeCombination(hv, disk.dev_type, access):
+          raise errors.OpPrereqError("Instance {i}: cannot use '{a}' access"
+                                     " setting with {h} hypervisor and {d} disk"
+                                     " type.".format(i=inst.name,
+                                                     a=access,
+                                                     h=hv,
+                                                     d=disk.dev_type))
 
 
 def IsValidDiskAccessModeCombination(hv, disk_template, mode):
@@ -1323,7 +1334,7 @@ def IsValidDiskAccessModeCombination(hv, disk_template, mode):
     return True
 
   if (hv == constants.HT_KVM and
-      disk_template in (constants.DT_RBD, constants.DT_GLUSTER) and
+      disk_template in constants.DTS_HAVE_ACCESS and
       mode == constants.DISK_USERSPACE):
     return True
 
@@ -1552,3 +1563,11 @@ def EnsureKvmdOnNodes(lu, feedback_fn, nodes=None):
     for node_uuid in stop_nodes:
       results[node_uuid].Warn("Failed to stop KVM daemon in node '%s'" %
                               node_uuid, feedback_fn)
+
+
+def WarnAboutFailedSshUpdates(result, master_uuid, feedback_fn):
+  node_errors = result[master_uuid].payload
+  if node_errors:
+    feedback_fn("Some nodes' SSH key files could not be updated:")
+    for node_name, error_msg in node_errors:
+      feedback_fn("%s: %s" % (node_name, error_msg))
