@@ -53,7 +53,7 @@ from ganeti.cmdlib.common import CheckParamsNotGlobal, \
   GetWantedNodes, MapInstanceLvsToNodes, RunPostHook, \
   FindFaultyInstanceDisks, CheckStorageTypeEnabled, GetClientCertDigest, \
   AddNodeCertToCandidateCerts, RemoveNodeCertFromCandidateCerts, \
-  EnsureKvmdOnNodes
+  EnsureKvmdOnNodes, WarnAboutFailedSshUpdates
 
 
 def _DecideSelfPromotion(lu, exceptions=None):
@@ -293,7 +293,7 @@ class LUNodeAdd(LogicalUnit):
     rpcrunner = rpc.DnsOnlyRunner()
     result = rpcrunner.call_version([node_name])[node_name]
     result.Raise("Can't get version information from node %s" % node_name,
-                 prereq=True)
+                 prereq=True, ecode=errors.ECODE_ENVIRON)
     if constants.PROTOCOL_VERSION == result.payload:
       logging.info("Communication to node %s fine, sw version %s match",
                    node_name, result.payload)
@@ -337,6 +337,43 @@ class LUNodeAdd(LogicalUnit):
                  self.new_node.name, ovs_name, ovs_link)
       result.Raise("Failed to initialize OpenVSwitch on new node")
 
+  def _SshUpdate(self, new_node_uuid, new_node_name, is_master_candidate,
+                 is_potential_master_candidate, rpcrunner, readd, feedback_fn):
+    """Update the SSH setup of all nodes after adding a new node.
+
+    @type readd: boolean
+    @param readd: whether or not this node is readded
+
+    """
+    potential_master_candidates = self.cfg.GetPotentialMasterCandidates()
+    master_node = self.cfg.GetMasterNode()
+
+    if readd:
+      # clear previous keys
+      master_candidate_uuids = self.cfg.GetMasterCandidateUuids()
+      remove_result = rpcrunner.call_node_ssh_key_remove(
+        [master_node],
+        new_node_uuid, new_node_name,
+        master_candidate_uuids,
+        potential_master_candidates,
+        True, # from authorized keys
+        True, # from public keys
+        False, # clear authorized keys
+        True) # clear public keys
+      remove_result[master_node].Raise(
+        "Could not remove SSH keys of node %s before readding,"
+        " (UUID: %s)." % (new_node_name, new_node_uuid))
+      WarnAboutFailedSshUpdates(remove_result, master_node, feedback_fn)
+
+    result = rpcrunner.call_node_ssh_key_add(
+      [master_node], new_node_uuid, new_node_name,
+      potential_master_candidates,
+      is_master_candidate, is_potential_master_candidate,
+      is_potential_master_candidate)
+
+    result[master_node].Raise("Could not update the node's SSH setup.")
+    WarnAboutFailedSshUpdates(result, master_node, feedback_fn)
+
   def Exec(self, feedback_fn):
     """Adds the new node to the cluster.
 
@@ -393,7 +430,7 @@ class LUNodeAdd(LogicalUnit):
 
     node_verifier_uuids = [self.cfg.GetMasterNode()]
     node_verify_param = {
-      constants.NV_NODELIST: ([self.new_node.name], {}),
+      constants.NV_NODELIST: ([self.new_node.name], {}, []),
       # TODO: do a node-net-test as well?
     }
 
@@ -439,6 +476,13 @@ class LUNodeAdd(LogicalUnit):
 
     EnsureKvmdOnNodes(self, feedback_fn, nodes=[self.new_node.uuid])
 
+    # Update SSH setup of all nodes
+    if self.op.node_setup:
+      # FIXME: so far, all nodes are considered potential master candidates
+      self._SshUpdate(self.new_node.uuid, self.new_node.name,
+                      self.new_node.master_candidate, True,
+                      self.rpc, self.op.readd, feedback_fn)
+
 
 class LUNodeSetParams(LogicalUnit):
   """Modifies the parameters of a node.
@@ -496,16 +540,15 @@ class LUNodeSetParams(LogicalUnit):
     """Filter for getting affected instances.
 
     """
-    return (instance.disk_template in constants.DTS_INT_MIRROR and
+    disks = self.cfg.GetInstanceDisks(instance.uuid)
+    any_mirrored = utils.AnyDiskOfType(disks, constants.DTS_INT_MIRROR)
+    return (any_mirrored and
             self.op.node_uuid in self.cfg.GetInstanceNodes(instance.uuid))
 
   def ExpandNames(self):
     if self.lock_all:
       self.needed_locks = {
         locking.LEVEL_NODE: locking.ALL_SET,
-
-        # Block allocations when all nodes are locked
-        locking.LEVEL_NODE_ALLOC: locking.ALL_SET,
         }
     else:
       self.needed_locks = {
@@ -522,7 +565,6 @@ class LUNodeSetParams(LogicalUnit):
     self.share_locks = ShareAll()
     self.share_locks[locking.LEVEL_NODE] = 0
     self.share_locks[locking.LEVEL_NODE_RES] = 0
-    self.share_locks[locking.LEVEL_NODE_ALLOC] = 0
 
     if self.lock_instances:
       self.needed_locks[locking.LEVEL_INSTANCE] = \
@@ -815,12 +857,42 @@ class LUNodeSetParams(LogicalUnit):
       if self.old_role == self._ROLE_CANDIDATE:
         RemoveNodeCertFromCandidateCerts(self.cfg, node.uuid)
 
+    EnsureKvmdOnNodes(self, feedback_fn, nodes=[node.uuid])
+
     # this will trigger job queue propagation or cleanup if the mc
     # flag changed
     if [self.old_role, self.new_role].count(self._ROLE_CANDIDATE) == 1:
       self.context.ReaddNode(node)
 
-    EnsureKvmdOnNodes(self, feedback_fn, nodes=[node.uuid])
+      if self.cfg.GetClusterInfo().modify_ssh_setup:
+        potential_master_candidates = self.cfg.GetPotentialMasterCandidates()
+        master_node = self.cfg.GetMasterNode()
+        if self.old_role == self._ROLE_CANDIDATE:
+          master_candidate_uuids = self.cfg.GetMasterCandidateUuids()
+          ssh_result = self.rpc.call_node_ssh_key_remove(
+            [master_node],
+            node.uuid, node.name,
+            master_candidate_uuids, potential_master_candidates,
+            True, # remove node's key from all nodes' authorized_keys file
+            False, # currently, all nodes are potential master candidates
+            False, # do not clear node's 'authorized_keys'
+            False) # do not clear node's 'ganeti_pub_keys'
+          ssh_result[master_node].Raise(
+            "Could not adjust the SSH setup after demoting node '%s'"
+            " (UUID: %s)." % (node.name, node.uuid))
+          WarnAboutFailedSshUpdates(ssh_result, master_node, feedback_fn)
+
+        if self.new_role == self._ROLE_CANDIDATE:
+          ssh_result = self.rpc.call_node_ssh_key_add(
+            [master_node], node.uuid, node.name,
+            potential_master_candidates,
+            True, # add node's key to all node's 'authorized_keys'
+            True, # all nodes are potential master candidates
+            False) # do not update the node's public keys
+          ssh_result[master_node].Raise(
+            "Could not update the SSH setup of node '%s' after promotion"
+            " (UUID: %s)." % (node.name, node.uuid))
+          WarnAboutFailedSshUpdates(ssh_result, master_node, feedback_fn)
 
     return result
 
@@ -890,7 +962,7 @@ def _GetNodeInstances(cfg, node_uuid):
 
   return _GetNodeInstancesInner(cfg,
                                 lambda inst: node_uuid in
-                                  cfg.GetInstanceNodes(inst.uuid.uuid))
+                                  cfg.GetInstanceNodes(inst.uuid))
 
 
 class LUNodeEvacuate(NoHooksLU):
@@ -1055,8 +1127,9 @@ class LUNodeEvacuate(NoHooksLU):
 
     elif self.op.iallocator is not None:
       # TODO: Implement relocation to other group
-      req = iallocator.IAReqNodeEvac(evac_mode=self.op.mode,
-                                     instances=list(self.instance_names))
+      req = iallocator.IAReqNodeEvac(
+          evac_mode=self.op.mode, instances=list(self.instance_names),
+          ignore_soft_errors=self.op.ignore_soft_errors)
       ial = iallocator.IAllocator(self.cfg, self.rpc, req)
 
       ial.Run(self.op.iallocator)
@@ -1252,7 +1325,6 @@ class LUNodeQueryvols(NoHooksLU):
     else:
       self.needed_locks = {
         locking.LEVEL_NODE: locking.ALL_SET,
-        locking.LEVEL_NODE_ALLOC: locking.ALL_SET,
         }
 
   def Exec(self, feedback_fn):
@@ -1328,7 +1400,6 @@ class LUNodeQueryStorage(NoHooksLU):
     else:
       self.needed_locks = {
         locking.LEVEL_NODE: locking.ALL_SET,
-        locking.LEVEL_NODE_ALLOC: locking.ALL_SET,
         }
 
   def _DetermineStorageType(self):
@@ -1497,6 +1568,27 @@ class LUNodeRemove(LogicalUnit):
     assert locking.BGL in self.owned_locks(locking.LEVEL_CLUSTER), \
       "Not owning BGL"
 
+    if modify_ssh_setup:
+      # retrieve the list of potential master candidates before the node is
+      # removed
+      potential_master_candidates = self.cfg.GetPotentialMasterCandidates()
+      potential_master_candidate = \
+        self.op.node_name in potential_master_candidates
+      master_candidate_uuids = self.cfg.GetMasterCandidateUuids()
+      master_node = self.cfg.GetMasterNode()
+      result = self.rpc.call_node_ssh_key_remove(
+        [master_node],
+        self.node.uuid, self.op.node_name,
+        master_candidate_uuids, potential_master_candidates,
+        self.node.master_candidate, # from_authorized_keys
+        potential_master_candidate, # from_public_keys
+        True, # clear node's 'authorized_keys'
+        True) # clear node's 'ganeti_public_keys'
+      result[master_node].Raise(
+        "Could not remove the SSH key of node '%s' (UUID: %s)." %
+        (self.op.node_name, self.node.uuid))
+      WarnAboutFailedSshUpdates(result, master_node, feedback_fn)
+
     # Promote nodes to master candidate as needed
     AdjustCandidatePool(self, [self.node.uuid])
     self.context.RemoveNode(self.cfg, self.node)
