@@ -36,6 +36,7 @@ import logging
 import errno
 import string # pylint: disable=W0402
 import shutil
+import time
 from cStringIO import StringIO
 
 from ganeti import constants
@@ -488,8 +489,11 @@ class XenHypervisor(hv_base.BaseHypervisor):
   REBOOT_RETRY_COUNT = 60
   REBOOT_RETRY_INTERVAL = 10
   _ROOT_DIR = pathutils.RUN_DIR + "/xen-hypervisor"
-  _NICS_DIR = _ROOT_DIR + "/nic" # contains NICs' info
-  _DIRS = [_ROOT_DIR, _NICS_DIR]
+  # contains NICs' info
+  _NICS_DIR = _ROOT_DIR + "/nic"
+  # contains the pidfiles of socat processes used to migrate instaces under xl
+  _MIGRATION_DIR = _ROOT_DIR + "/migration"
+  _DIRS = [_ROOT_DIR, _NICS_DIR, _MIGRATION_DIR]
 
   _INSTANCE_LIST_DELAYS = (0.3, 1.5, 1.0)
   _INSTANCE_LIST_TIMEOUT = 5
@@ -581,6 +585,20 @@ class XenHypervisor(hv_base.BaseHypervisor):
     """
     return utils.PathJoin(self._cfgdir, instance_name)
 
+  @classmethod
+  def _EnsureDirs(cls, extra_dirs=None):
+    """Makes sure that the directories needed by the hypervisor exist.
+
+    @type extra_dirs: list of string or None
+    @param extra_dirs: Additional directories which ought to exist.
+
+    """
+    if extra_dirs is None:
+      extra_dirs = []
+    dirs = [(dname, constants.RUN_DIRS_MODE) for dname in
+            (cls._DIRS + extra_dirs)]
+    utils.EnsureDirs(dirs)
+
   @classmethod
   def _WriteNICInfoFile(cls, instance, idx, nic):
     """Write the Xen config file for the instance.
@@ -589,18 +607,16 @@ class XenHypervisor(hv_base.BaseHypervisor):
 
     """
     instance_name = instance.name
-    dirs = [(dname, constants.RUN_DIRS_MODE)
-            for dname in cls._DIRS + [cls._InstanceNICDir(instance_name)]]
-    utils.EnsureDirs(dirs)
+    cls._EnsureDirs(extra_dirs=[cls._InstanceNICDir(instance_name)])
 
     cfg_file = cls._InstanceNICFile(instance_name, idx)
     data = StringIO()
 
-    data.write("TAGS=%s\n" % r"\ ".join(instance.GetTags()))
+    data.write("TAGS=\"%s\"\n" % r"\ ".join(instance.GetTags()))
     if nic.netinfo:
       netinfo = objects.Network.FromDict(nic.netinfo)
       for k, v in netinfo.HooksDict().iteritems():
-        data.write("%s=%s\n" % (k, v))
+        data.write("%s=\"%s\"\n" % (k, v))
 
     data.write("MAC=%s\n" % nic.mac)
     if nic.ip:
@@ -619,6 +635,27 @@ class XenHypervisor(hv_base.BaseHypervisor):
       raise errors.HypervisorError("Cannot write Xen instance configuration"
                                    " file %s: %s" % (cfg_file, err))
 
+  @staticmethod
+  def VersionsSafeForMigration(src, target):
+    """Decide if migration is likely to suceed for hypervisor versions.
+
+    Given two versions of a hypervisor, give a guess whether live migration
+    from the one version to the other version is likely to succeed. For Xen,
+    the heuristics is, that an increase by one on the second digit is OK. This
+    fits with the current numbering scheme.
+
+    @type src: list or tuple
+    @type target: list or tuple
+    @rtype: bool
+    """
+    if src == target:
+      return True
+
+    if len(src) < 2 or len(target) < 2:
+      return False
+
+    return src[0] == target[0] and target[1] in [src[1], src[1] + 1]
+
   @classmethod
   def _InstanceNICDir(cls, instance_name):
     """Returns the directory holding the tap device files for a given instance.
@@ -633,6 +670,18 @@ class XenHypervisor(hv_base.BaseHypervisor):
     """
     return utils.PathJoin(cls._InstanceNICDir(instance_name), str(seq))
 
+  @classmethod
+  def _InstanceMigrationPidfile(cls, _instance_name):
+    """Returns the name of the pid file for a socat process used to migrate.
+
+    """
+    #TODO(riba): At the moment, we are using a single pidfile because we
+    # use a single port for migrations at the moment. This is because we do not
+    # allow more migrations, so dynamic port selection and the needed port
+    # modifications are not needed.
+    # The _instance_name parameter has been left here for future use.
+    return utils.PathJoin(cls._MIGRATION_DIR, constants.XL_MIGRATION_PIDFILE)
+
   @classmethod
   def _GetConfig(cls, instance, startup_memory, block_devices):
     """Build Xen configuration for an instance.
@@ -1079,6 +1128,36 @@ class XenHypervisor(hv_base.BaseHypervisor):
     """
     return self._ReadConfigFile(instance.name)
 
+  def _UseMigrationDaemon(self, hvparams):
+    """Whether to start a socat daemon when accepting an instance.
+
+    @rtype: bool
+
+    """
+    return self._GetCommand(hvparams) == constants.XEN_CMD_XL
+
+  @classmethod
+  def _KillMigrationDaemon(cls, instance):
+    """Kills the migration daemon if present.
+
+    """
+    pidfile = cls._InstanceMigrationPidfile(instance.name)
+    read_pid = utils.ReadPidFile(pidfile)
+
+    # There is no pidfile, hence nothing for us to do
+    if read_pid == 0:
+      return
+
+    if utils.IsProcessAlive(read_pid):
+      # If the process is alive, let's make sure we are killing the right one
+      cmdline = ' '.join(utils.GetProcCmdline(read_pid))
+      if cmdline.count("xl migrate-receive") > 0:
+        utils.KillProcess(read_pid)
+
+    # By this point the process is not running, whether killed or initially
+    # nonexistent, so it is safe to remove the pidfile.
+    utils.RemoveFile(pidfile)
+
   def AcceptInstance(self, instance, info, target):
     """Prepare to accept an instance.
 
@@ -1090,7 +1169,28 @@ class XenHypervisor(hv_base.BaseHypervisor):
     @param target: target host (usually ip), on this node
 
     """
-    pass
+    if self._UseMigrationDaemon(instance.hvparams):
+      port = instance.hvparams[constants.HV_MIGRATION_PORT]
+
+      # Make sure there is somewhere to put the pidfile.
+      XenHypervisor._EnsureDirs()
+      pidfile = XenHypervisor._InstanceMigrationPidfile(instance.name)
+
+      # And try and kill a previous daemon
+      XenHypervisor._KillMigrationDaemon(instance)
+
+      listening_arg = "TCP-LISTEN:%d,bind=%s" % (port, target)
+      socat_pid = utils.StartDaemon(["socat", listening_arg,
+                                     "SYSTEM:'xl migrate-receive'"],
+                                     pidfile=pidfile)
+
+      # Wait for a while to make sure the socat process has successfully started
+      # listening
+      time.sleep(1)
+      if not utils.IsProcessAlive(socat_pid):
+        raise errors.HypervisorError("Could not start receiving socat process"
+                                     " on port %d: check if port is available" %
+                                     port)
 
   def FinalizeMigrationDst(self, instance, info, success):
     """Finalize an instance migration.
@@ -1108,8 +1208,10 @@ class XenHypervisor(hv_base.BaseHypervisor):
     """
     if success:
       self._WriteConfigFile(instance.name, info)
+    elif self._UseMigrationDaemon(instance.hvparams):
+      XenHypervisor._KillMigrationDaemon(instance)
 
-  def MigrateInstance(self, cluster_name, instance, target, live):
+  def MigrateInstance(self, _cluster_name, instance, target, live):
     """Migrate an instance to a target node.
 
     The migration will not be attempted if the instance is not
@@ -1125,11 +1227,11 @@ class XenHypervisor(hv_base.BaseHypervisor):
     """
     port = instance.hvparams[constants.HV_MIGRATION_PORT]
 
-    return self._MigrateInstance(cluster_name, instance.name, target, port,
-                                 live, instance.hvparams)
+    return self._MigrateInstance(instance.name, target, port, live,
+                                 instance.hvparams)
 
-  def _MigrateInstance(self, cluster_name, instance_name, target, port, live,
-                       hvparams, _ping_fn=netutils.TcpPing):
+  def _MigrateInstance(self, instance_name, target, port, live, hvparams,
+                       _ping_fn=netutils.TcpPing):
     """Migrate an instance to a target node.
 
     @see: L{MigrateInstance} for details
@@ -1143,26 +1245,28 @@ class XenHypervisor(hv_base.BaseHypervisor):
 
     cmd = self._GetCommand(hvparams)
 
-    if (cmd == constants.XEN_CMD_XM and
-        not _ping_fn(target, port, live_port_needed=True)):
-      raise errors.HypervisorError("Remote host %s not listening on port"
-                                   " %s, cannot migrate" % (target, port))
-
     args = ["migrate"]
 
     if cmd == constants.XEN_CMD_XM:
+      # Try and see if xm is listening on the specified port
+      if not _ping_fn(target, port, live_port_needed=True):
+        raise errors.HypervisorError("Remote host %s not listening on port"
+                                     " %s, cannot migrate" % (target, port))
+
       args.extend(["-p", "%d" % port])
       if live:
         args.append("-l")
 
     elif cmd == constants.XEN_CMD_XL:
+      # Rather than using SSH, use socat as Ganeti cannot guarantee the presence
+      # of usable SSH keys as of 2.13
       args.extend([
-        "-s", constants.XL_SSH_CMD % cluster_name,
+        "-s", constants.XL_SOCAT_CMD % (target, port),
         "-C", self._ConfigFileName(instance_name),
         ])
 
     else:
-      raise errors.HypervisorError("Unsupported Xen command: %s" % self._cmd)
+      raise errors.HypervisorError("Unsupported Xen command: %s" % cmd)
 
     args.extend([instance_name, target])
 
