@@ -42,6 +42,7 @@ from ganeti import utils
 
 import ganeti.masterd.instance as gmi
 
+import logging
 
 _STRING_LIST = ht.TListOf(ht.TString)
 _JOB_LIST = ht.TListOf(ht.TListOf(ht.TStrictDict(True, False, {
@@ -141,6 +142,12 @@ class IARequestBase(outils.ValidatedSlots):
     """
     raise NotImplementedError
 
+  def GetExtraParams(self): # pylint: disable=R0201
+    """Gets extra parameters to the IAllocator call.
+
+    """
+    return {}
+
   def ValidateResult(self, ia, result):
     """Validates the result of an request.
 
@@ -167,6 +174,7 @@ class IAReqInstanceAlloc(IARequestBase):
     ("spindle_use", ht.TNonNegativeInt),
     ("disks", ht.TListOf(ht.TDict)),
     ("disk_template", ht.TString),
+    ("group_name", ht.TMaybe(ht.TNonEmptyString)),
     ("os", ht.TString),
     ("tags", _STRING_LIST),
     ("nics", ht.TListOf(ht.TDict)),
@@ -192,11 +200,14 @@ class IAReqInstanceAlloc(IARequestBase):
     done.
 
     """
-    disk_space = gmi.ComputeDiskSize(self.disk_template, self.disks)
+    for d in self.disks:
+      d[constants.IDISK_TYPE] = self.disk_template
+    disk_space = gmi.ComputeDiskSize(self.disks)
 
     return {
       "name": self.name,
       "disk_template": self.disk_template,
+      "group_name": self.group_name,
       "tags": self.tags,
       "os": self.os,
       "vcpus": self.vcpus,
@@ -265,23 +276,24 @@ class IAReqRelocate(IARequestBase):
 
     """
     instance = cfg.GetInstanceInfo(self.inst_uuid)
+    disks = cfg.GetInstanceDisks(self.inst_uuid)
     if instance is None:
       raise errors.ProgrammerError("Unknown instance '%s' passed to"
                                    " IAllocator" % self.inst_uuid)
 
-    if instance.disk_template not in constants.DTS_MIRRORED:
+    if not utils.AllDiskOfType(disks, constants.DTS_MIRRORED):
       raise errors.OpPrereqError("Can't relocate non-mirrored instances",
                                  errors.ECODE_INVAL)
 
     secondary_nodes = cfg.GetInstanceSecondaryNodes(instance.uuid)
-    if (instance.disk_template in constants.DTS_INT_MIRROR and
+    if (utils.AnyDiskOfType(disks, constants.DTS_INT_MIRROR) and
         len(secondary_nodes) != 1):
       raise errors.OpPrereqError("Instance has not exactly one secondary node",
                                  errors.ECODE_STATE)
 
-    inst_disks = cfg.GetInstanceDisks(instance.uuid)
-    disk_sizes = [{constants.IDISK_SIZE: disk.size} for disk in inst_disks]
-    disk_space = gmi.ComputeDiskSize(instance.disk_template, disk_sizes)
+    disk_sizes = [{constants.IDISK_SIZE: disk.size,
+                   constants.IDISK_TYPE: disk.dev_type} for disk in disks]
+    disk_space = gmi.ComputeDiskSize(disk_sizes)
 
     return {
       "name": instance.name,
@@ -357,6 +369,7 @@ class IAReqNodeEvac(IARequestBase):
   REQ_PARAMS = [
     ("instances", _STRING_LIST),
     ("evac_mode", ht.TEvacMode),
+    ("ignore_soft_errors", ht.TBool),
     ]
   REQ_RESULT = _NEVAC_RESULT
 
@@ -369,6 +382,16 @@ class IAReqNodeEvac(IARequestBase):
       "evac_mode": self.evac_mode,
       }
 
+  def GetExtraParams(self):
+    """Get extra iallocator command line options for
+    node-evacuate requests.
+
+    """
+    if self.ignore_soft_errors:
+      return {"ignore-soft-errors": None}
+    else:
+      return {}
+
 
 class IAReqGroupChange(IARequestBase):
   """A group change request.
@@ -752,6 +775,7 @@ class IAllocator(object):
           nic_dict["bridge"] = filled_params[constants.NIC_LINK]
         nic_data.append(nic_dict)
       inst_disks = cfg.GetInstanceDisks(iinfo.uuid)
+      inst_disktemplate = cfg.GetInstanceDiskTemplate(iinfo.uuid)
       pir = {
         "tags": list(iinfo.GetTags()),
         "admin_state": iinfo.admin_state,
@@ -763,16 +787,16 @@ class IAllocator(object):
                  cfg.GetNodeNames(
                    cfg.GetInstanceSecondaryNodes(iinfo.uuid)),
         "nics": nic_data,
-        "disks": [{constants.IDISK_SIZE: dsk.size,
+        "disks": [{constants.IDISK_TYPE: dsk.dev_type,
+                   constants.IDISK_SIZE: dsk.size,
                    constants.IDISK_MODE: dsk.mode,
                    constants.IDISK_SPINDLES: dsk.spindles}
                   for dsk in inst_disks],
-        "disk_template": iinfo.disk_template,
+        "disk_template": inst_disktemplate,
         "disks_active": iinfo.disks_active,
         "hypervisor": iinfo.hypervisor,
         }
-      pir["disk_space_total"] = gmi.ComputeDiskSize(iinfo.disk_template,
-                                                    pir["disks"])
+      pir["disk_space_total"] = gmi.ComputeDiskSize(pir["disks"])
       instance_data[iinfo.name] = pir
 
     return instance_data
@@ -783,14 +807,17 @@ class IAllocator(object):
     """
     request = req.GetRequest(self.cfg)
     disk_template = None
-    if "disk_template" in request:
+    if request.get("disk_template") is not None:
       disk_template = request["disk_template"]
+    elif isinstance(req, IAReqRelocate):
+      disk_template = self.cfg.GetInstanceDiskTemplate(self.req.inst_uuid)
     self._ComputeClusterData(disk_template=disk_template)
 
     request["type"] = req.MODE
     self.in_data["request"] = request
 
     self.in_text = serializer.Dump(self.in_data)
+    logging.debug("IAllocator request: %s", self.in_text)
 
   def Run(self, name, validate=True, call_fn=None):
     """Run an instance allocator and return the results.
@@ -801,6 +828,9 @@ class IAllocator(object):
 
     ial_params = self.cfg.GetDefaultIAllocatorParameters()
 
+    for ial_param in self.req.GetExtraParams().items():
+      ial_params[ial_param[0]] = ial_param[1]
+
     result = call_fn(self.cfg.GetMasterNode(), name, self.in_text, ial_params)
     result.Raise("Failure while running the iallocator script")
 
