@@ -62,19 +62,6 @@ lusExecuting = [0]
 _OP_PREFIX = "Op"
 _LU_PREFIX = "LU"
 
-#: LU classes which don't need to acquire the node allocation lock
-#: (L{locking.NAL}) when they acquire all node or node resource locks
-_NODE_ALLOC_WHITELIST = frozenset([])
-
-#: LU classes which don't need to acquire the node allocation lock
-#: (L{locking.NAL}) in the same mode (shared/exclusive) as the node
-#: or node resource locks
-_NODE_ALLOC_MODE_WHITELIST = compat.UniqueFrozenset([
-  cmdlib.LUBackupExport,
-  cmdlib.LUBackupRemove,
-  cmdlib.LUOobCommand,
-  ])
-
 
 class LockAcquireTimeout(Exception):
   """Exception to report timeouts on acquiring locks.
@@ -164,6 +151,13 @@ class OpExecCbBase(object): # pylint: disable=W0232
 
     """
 
+  def NotifyRetry(self):
+    """Called when we are about to reset an LU to retry again.
+
+    This function is called after PrepareRetry successfully completed.
+
+    """
+
   def Feedback(self, *args):
     """Sends feedback from the LU code to the end-user.
 
@@ -263,43 +257,6 @@ def _FailingSubmitManyJobs(_):
                                " queries) can not submit jobs")
 
 
-def _VerifyLocks(lu, _mode_whitelist=_NODE_ALLOC_MODE_WHITELIST,
-                 _nal_whitelist=_NODE_ALLOC_WHITELIST):
-  """Performs consistency checks on locks acquired by a logical unit.
-
-  @type lu: L{cmdlib.LogicalUnit}
-  @param lu: Logical unit instance
-
-  """
-  if not __debug__:
-    return
-
-  allocset = lu.owned_locks(locking.LEVEL_NODE_ALLOC)
-  have_nal = locking.NAL in allocset
-
-  for level in [locking.LEVEL_NODE, locking.LEVEL_NODE_RES]:
-    # TODO: Verify using actual lock mode, not using LU variables
-    if level in lu.needed_locks:
-      share_node_alloc = lu.share_locks[locking.LEVEL_NODE_ALLOC]
-      share_level = lu.share_locks[level]
-
-      if lu.__class__ in _mode_whitelist:
-        assert share_node_alloc != share_level, \
-          "LU is whitelisted to use different modes for node allocation lock"
-      else:
-        assert bool(share_node_alloc) == bool(share_level), \
-          ("Node allocation lock must be acquired using the same mode as nodes"
-           " and node resources")
-
-      if lu.__class__ in _nal_whitelist:
-        assert not have_nal, \
-          "LU is whitelisted for not acquiring the node allocation lock"
-      elif lu.needed_locks[level] == locking.ALL_SET:
-        assert have_nal, \
-          ("Node allocation lock must be used if an LU acquires all nodes"
-           " or node resources")
-
-
 def _LockList(names):
   """If 'names' is a string, make it a single-element list.
 
@@ -351,8 +308,62 @@ class Processor(object):
     if not self._enable_locks:
       raise errors.ProgrammerError("Attempted to use disabled locks")
 
+  def _RequestAndWait(self, request, timeout):
+    """Request locks from WConfD and wait for them to be granted.
+
+    @type request: list
+    @param request: the lock request to be sent to WConfD
+    @type timeout: float
+    @param timeout: the time to wait for the request to be granted
+    @raise LockAcquireTimeout: In case locks couldn't be acquired in specified
+        amount of time; in this case, locks still might be acquired or a request
+        pending.
+
+    """
+    logging.debug("Trying %ss to request %s for %s",
+                  timeout, request, self._wconfdcontext)
+    if self._cbs:
+      priority = self._cbs.CurrentPriority() # pylint: disable=W0612
+    else:
+      priority = None
+
+    if priority is None:
+      priority = constants.OP_PRIO_DEFAULT
+
+    ## Expect a signal
+    if sighupReceived[0]:
+      logging.warning("Ignoring unexpected SIGHUP")
+    sighupReceived[0] = False
+
+    # Request locks
+    self.wconfd.Client().UpdateLocksWaiting(self._wconfdcontext, priority,
+                                            request)
+    pending = self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
+
+    if pending:
+      def _HasPending():
+        if sighupReceived[0]:
+          return self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
+        else:
+          return True
+
+      pending = utils.SimpleRetry(False, _HasPending, 0.05, timeout)
+
+      signal = sighupReceived[0]
+
+      if pending:
+        pending = self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
+
+      if pending and signal:
+        logging.warning("Ignoring unexpected SIGHUP")
+      sighupReceived[0] = False
+
+    logging.debug("Finished trying. Pending: %s", pending)
+    if pending:
+      raise LockAcquireTimeout()
+
   def _AcquireLocks(self, level, names, shared, opportunistic, timeout,
-                    opportunistic_count=1):
+                    opportunistic_count=1, request_only=False):
     """Acquires locks via the Ganeti lock manager.
 
     @type level: int
@@ -365,8 +376,11 @@ class Processor(object):
     @param opportunistic: Whether to acquire opportunistically
     @type timeout: None or float
     @param timeout: Timeout for acquiring the locks
+    @type request_only: bool
+    @param request_only: do not acquire the locks, just return the request
     @raise LockAcquireTimeout: In case locks couldn't be acquired in specified
-        amount of time
+        amount of time; in this case, locks still might be acquired or a request
+        pending.
 
     """
     self._CheckLocksEnabled()
@@ -384,7 +398,6 @@ class Processor(object):
         expand_fns = {
           locking.LEVEL_CLUSTER: (lambda: [locking.BGL]),
           locking.LEVEL_INSTANCE: self.cfg.GetInstanceList,
-          locking.LEVEL_NODE_ALLOC: (lambda: [locking.NAL]),
           locking.LEVEL_NODEGROUP: self.cfg.GetNodeGroupList,
           locking.LEVEL_NODE: self.cfg.GetNodeList,
           locking.LEVEL_NODE_RES: self.cfg.GetNodeList,
@@ -413,6 +426,10 @@ class Processor(object):
     else:
       request = [[lock, "exclusive"] for lock in locks]
 
+    if request_only:
+      logging.debug("Lock request for level %s is %s", level, request)
+      return request
+
     self.cfg.OutDate()
 
     if timeout is None:
@@ -444,42 +461,7 @@ class Processor(object):
       if locks == []:
         raise LockAcquireTimeout()
     else:
-      logging.debug("Trying %ss to request %s for %s",
-                    timeout, request, self._wconfdcontext)
-      ## Expect a signal
-      if sighupReceived[0]:
-        logging.warning("Ignoring unexpected SIGHUP")
-      sighupReceived[0] = False
-
-      # Request locks
-      self.wconfd.Client().UpdateLocksWaiting(self._wconfdcontext, priority,
-                                              request)
-      pending = self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
-
-      if pending:
-        def _HasPending():
-          if sighupReceived[0]:
-            return self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
-          else:
-            return True
-
-        pending = utils.SimpleRetry(False, _HasPending, 0.05, timeout)
-
-        signal = sighupReceived[0]
-
-        if pending:
-          pending = self.wconfd.Client().HasPendingRequest(self._wconfdcontext)
-
-        if pending and signal:
-          logging.warning("Ignoring unexpected SIGHUP")
-        sighupReceived[0] = False
-
-      logging.debug("Finished trying. Pending: %s", pending)
-      if pending:
-        # drop the pending request and all locks potentially obtained in the
-        # time since the last poll.
-        self.wconfd.Client().FreeLocksLevel(self._wconfdcontext, levelname)
-        raise LockAcquireTimeout()
+      self._RequestAndWait(request, timeout)
 
     return locks
 
@@ -526,7 +508,7 @@ class Processor(object):
   def BuildHooksManager(self, lu):
     return self.hmclass.BuildFromLu(lu.rpc.call_hooks_runner, lu)
 
-  def _LockAndExecLU(self, lu, level, calc_timeout):
+  def _LockAndExecLU(self, lu, level, calc_timeout, pending=None):
     """Execute a Logical Unit, with the needed locks.
 
     This is a recursive function that starts locking the given level, and
@@ -534,17 +516,39 @@ class Processor(object):
     given LU and its opcodes.
 
     """
+    pending = pending or []
+    logging.debug("Looking at locks of level %s, still need to obtain %s",
+                  level, pending)
     adding_locks = level in lu.add_locks
     acquiring_locks = level in lu.needed_locks
 
     if level not in locking.LEVELS:
-      _VerifyLocks(lu)
+      if pending:
+        self._RequestAndWait(pending, calc_timeout())
+        lu.cfg.OutDate()
+        lu.wconfdlocks = self.wconfd.Client().ListLocks(self._wconfdcontext)
+        pending = []
+
+      logging.debug("Finished acquiring locks")
 
       if self._cbs:
         self._cbs.NotifyStart()
 
       try:
         result = self._ExecLU(lu)
+      except errors.OpPrereqError, err:
+        (_, ecode) = err.args
+        if ecode != errors.ECODE_TEMP_NORES:
+          raise
+        logging.debug("Temporarily out of resources; will retry internally")
+        try:
+          lu.PrepareRetry(self.Log)
+          if self._cbs:
+            self._cbs.NotifyRetry()
+        except errors.OpRetryNotSupportedError:
+          logging.debug("LU does not know how to retry.")
+          raise err
+        raise LockAcquireTimeout()
       except AssertionError, err:
         # this is a bit ugly, as we don't know from which phase
         # (prereq, exec) this comes; but it's better than an exception
@@ -561,6 +565,14 @@ class Processor(object):
     # Determine if the acquiring is opportunistic up front
     opportunistic = lu.opportunistic_locks[level]
 
+    dont_collate = lu.dont_collate_locks[level]
+
+    if dont_collate and pending:
+      self._RequestAndWait(pending, calc_timeout())
+      lu.cfg.OutDate()
+      lu.wconfdlocks = self.wconfd.Client().ListLocks(self._wconfdcontext)
+      pending = []
+
     if adding_locks and opportunistic:
       # We could simultaneously acquire locks opportunistically and add new
       # ones, but that would require altering the API, and no use cases are
@@ -591,19 +603,31 @@ class Processor(object):
         if adding_locks:
           needed_locks.extend(_LockList(lu.add_locks[level]))
 
-        self._AcquireLocks(level, needed_locks, share, opportunistic,
-                           calc_timeout(),
-                           opportunistic_count=opportunistic_count)
-        lu.wconfdlocks = self.wconfd.Client().ListLocks(self._wconfdcontext)
+        timeout = calc_timeout()
+        if timeout is not None and not opportunistic:
+          pending = pending + self._AcquireLocks(level, needed_locks, share,
+                                                 opportunistic, timeout,
+                                                 request_only=True)
+        else:
+          if pending:
+            self._RequestAndWait(pending, calc_timeout())
+            lu.cfg.OutDate()
+            lu.wconfdlocks = self.wconfd.Client().ListLocks(self._wconfdcontext)
+            pending = []
+          self._AcquireLocks(level, needed_locks, share, opportunistic,
+                             timeout,
+                             opportunistic_count=opportunistic_count)
+          lu.wconfdlocks = self.wconfd.Client().ListLocks(self._wconfdcontext)
 
-        result = self._LockAndExecLU(lu, level + 1, calc_timeout)
+        result = self._LockAndExecLU(lu, level + 1, calc_timeout,
+                                     pending=pending)
       finally:
         levelname = locking.LEVEL_NAMES[level]
         logging.debug("Freeing locks at level %s for %s",
                       levelname, self._wconfdcontext)
         self.wconfd.Client().FreeLocksLevel(self._wconfdcontext, levelname)
     else:
-      result = self._LockAndExecLU(lu, level + 1, calc_timeout)
+      result = self._LockAndExecLU(lu, level + 1, calc_timeout, pending=pending)
 
     return result
 
