@@ -62,7 +62,8 @@ from socket import AF_INET
 
 
 __all__ = ["ConfigObject", "ConfigData", "NIC", "Disk", "Instance",
-           "OS", "Node", "NodeGroup", "Cluster", "FillDict", "Network"]
+           "OS", "Node", "NodeGroup", "Cluster", "FillDict", "Network",
+           "Filter"]
 
 _TIMESTAMPS = ["ctime", "mtime"]
 _UUID = ["uuid"]
@@ -110,10 +111,9 @@ def FillDiskParams(default_dparams, custom_dparams, skip_keys=None):
   @see: L{FillDict} for parameters and return value
 
   """
-  assert frozenset(default_dparams.keys()) == constants.DISK_TEMPLATES
-
-  return dict((dt, FillDict(default_dparams[dt], custom_dparams.get(dt, {}),
-                             skip_keys=skip_keys))
+  return dict((dt, FillDict(default_dparams.get(dt, {}),
+                            custom_dparams.get(dt, {}),
+                            skip_keys=skip_keys))
               for dt in constants.DISK_TEMPLATES)
 
 
@@ -415,6 +415,7 @@ class ConfigData(ConfigObject):
     "instances",
     "networks",
     "disks",
+    "filters",
     "serial_no",
     ] + _TIMESTAMPS
 
@@ -427,7 +428,8 @@ class ConfigData(ConfigObject):
     """
     mydict = super(ConfigData, self).ToDict(_with_private=_with_private)
     mydict["cluster"] = mydict["cluster"].ToDict()
-    for key in "nodes", "instances", "nodegroups", "networks", "disks":
+    for key in ("nodes", "instances", "nodegroups", "networks", "disks",
+                "filters"):
       mydict[key] = outils.ContainerToDicts(mydict[key])
 
     return mydict
@@ -446,21 +448,21 @@ class ConfigData(ConfigObject):
       outils.ContainerFromDicts(obj.nodegroups, dict, NodeGroup)
     obj.networks = outils.ContainerFromDicts(obj.networks, dict, Network)
     obj.disks = outils.ContainerFromDicts(obj.disks, dict, Disk)
+    obj.filters = outils.ContainerFromDicts(obj.filters, dict, Filter)
     return obj
 
-  def HasAnyDiskOfType(self, dev_type):
+  def DisksOfType(self, dev_type):
     """Check if in there is at disk of the given type in the configuration.
 
     @type dev_type: L{constants.DTS_BLOCK}
     @param dev_type: the type to look for
-    @rtype: boolean
-    @return: boolean indicating if a disk of the given type was found or not
+    @rtype: list of disks
+    @return: all disks of the dev_type
 
     """
-    for disk in self.disks.values():
-      if disk.IsBasedOnDiskType(dev_type):
-        return True
-    return False
+
+    return [disk for disk in self.disks.values()
+            if disk.IsBasedOnDiskType(dev_type)]
 
   def UpgradeConfig(self):
     """Fill defaults for missing configuration values.
@@ -487,6 +489,8 @@ class ConfigData(ConfigObject):
       network.UpgradeConfig()
     for disk in self.disks.values():
       disk.UpgradeConfig()
+    if self.filters is None:
+      self.filters = {}
 
   def _UpgradeEnabledDiskTemplates(self):
     """Upgrade the cluster's enabled disk templates by inspecting the currently
@@ -495,7 +499,9 @@ class ConfigData(ConfigObject):
     """
     if not self.cluster.enabled_disk_templates:
       template_set = \
-        set([inst.disk_template for inst in self.instances.values()])
+        set([d.dev_type for d in self.disks.values()])
+      if any(not inst.disks for inst in self.instances.values()):
+        template_set.add(constants.DT_DISKLESS)
       # Add drbd and plain, if lvm is enabled (by specifying a volume group)
       if self.cluster.volume_group_name:
         template_set.add(constants.DT_DRBD8)
@@ -537,13 +543,21 @@ class NIC(ConfigObject):
       raise errors.ConfigurationError("Missing bridged NIC link")
 
 
+class Filter(ConfigObject):
+  """Config object representing a filter rule."""
+  __slots__ = ["watermark", "priority",
+               "predicates", "action", "reason_trail"] + _UUID
+
+
 class Disk(ConfigObject):
   """Config object representing a block device."""
   __slots__ = [
+    "forthcoming",
     "name",
     "dev_type",
     "logical_id",
     "children",
+    "nodes",
     "iv_name",
     "size",
     "mode",
@@ -585,6 +599,10 @@ class Disk(ConfigObject):
     """Test if this device needs to be opened on a secondary node."""
     return self.dev_type in (constants.DT_PLAIN,)
 
+  def SupportsSnapshots(self):
+    """Test if this device supports snapshots."""
+    return self.dev_type in constants.DTS_SNAPSHOT_CAPABLE
+
   def StaticDevPath(self):
     """Return the device path if this device type has a static one.
 
@@ -1158,6 +1176,7 @@ def PutOSImage(osparams, os_image):
 class Instance(TaggableObject):
   """Config object representing an instance."""
   __slots__ = [
+    "forthcoming",
     "name",
     "primary_node",
     "secondary_nodes",
@@ -1220,6 +1239,10 @@ class Instance(TaggableObject):
       else:
         nlist = []
       bo[attr] = nlist
+
+    if 'disk_template' in bo:
+      del bo['disk_template']
+
     return bo
 
   @classmethod
@@ -1354,6 +1377,9 @@ class ExtStorage(ConfigObject):
     "detach_script",
     "setinfo_script",
     "verify_script",
+    "snapshot_script",
+    "open_script",
+    "close_script",
     "supported_parameters",
     ]
 
@@ -1625,6 +1651,7 @@ class Cluster(TaggableObject):
     "zeroing_image",
     "compression_tools",
     "enabled_user_shutdown",
+    "data_collectors",
     ] + _TIMESTAMPS + _UUID
 
   def UpgradeConfig(self):
