@@ -64,6 +64,7 @@ IMPORT_EXPORT_DAEMON = _constants.PKGLIBDIR + "/import-export"
 KVM_CONSOLE_WRAPPER = _constants.PKGLIBDIR + "/tools/kvm-console-wrapper"
 KVM_IFUP = _constants.PKGLIBDIR + "/kvm-ifup"
 PREPARE_NODE_JOIN = _constants.PKGLIBDIR + "/prepare-node-join"
+SSH_UPDATE = _constants.PKGLIBDIR + "/ssh-update"
 NODE_DAEMON_SETUP = _constants.PKGLIBDIR + "/node-daemon-setup"
 SSL_UPDATE = _constants.PKGLIBDIR + "/ssl-update"
 XEN_CONSOLE_WRAPPER = _constants.PKGLIBDIR + "/tools/xen-console-wrapper"
@@ -87,6 +88,7 @@ SSH_HOST_DSA_PRIV = _constants.SSH_HOST_DSA_PRIV
 SSH_HOST_DSA_PUB = _constants.SSH_HOST_DSA_PUB
 SSH_HOST_RSA_PRIV = _constants.SSH_HOST_RSA_PRIV
 SSH_HOST_RSA_PUB = _constants.SSH_HOST_RSA_PUB
+SSH_PUB_KEYS = DATA_DIR + "/ganeti_pub_keys"
 
 BDEV_CACHE_DIR = RUN_DIR + "/bdev-cache"
 DISK_LINKS_DIR = RUN_DIR + "/instance-disks"
@@ -154,6 +156,8 @@ MASTER_SOCKET = SOCKET_DIR + "/ganeti-master"
 QUERY_SOCKET = SOCKET_DIR + "/ganeti-query"
 #: WConfD socket
 WCONFD_SOCKET = SOCKET_DIR + "/ganeti-wconfd"
+#: Metad socket
+METAD_SOCKET = SOCKET_DIR + "/ganeti-metad"
 
 LOG_OS_DIR = LOG_DIR + "/os"
 LOG_ES_DIR = LOG_DIR + "/extstorage"
