@@ -370,6 +370,11 @@ _BLOCKDEV_CALLS = [
     ("info", None, None),
     ("exclusive_storage", None, None),
     ], None, None, "Request creation of a given block device"),
+  ("blockdev_convert", SINGLE, None, constants.RPC_TMO_SLOW, [
+    ("bdev_src", ED_SINGLE_DISK_DICT_DP, None),
+    ("bdev_dest", ED_SINGLE_DISK_DICT_DP, None),
+    ], None, None,
+    "Request the copy of the source block device to the destination one"),
   ("blockdev_image", SINGLE, None, constants.RPC_TMO_SLOW, [
     ("bdev", ED_SINGLE_DISK_DICT_DP, None),
     ("image", None, None),
@@ -412,6 +417,11 @@ _BLOCKDEV_CALLS = [
     ("instance_name", None, None),
     ("disks", ED_DISKS_DICT_DP, None),
     ], None, None, "Closes the given block devices"),
+  ("blockdev_open", SINGLE, None, constants.RPC_TMO_NORMAL, [
+    ("instance_name", None, None),
+    ("disks", ED_DISKS_DICT_DP, None),
+    ("exclusive", None, None),
+    ], None, None, "Opens the given block devices in required mode"),
   ("blockdev_getdimensions", SINGLE, None, constants.RPC_TMO_NORMAL, [
     ("disks", ED_MULTI_DISKS_DICT_DP, None),
     ], None, None, "Returns size and spindles of the given disks"),
@@ -421,7 +431,6 @@ _BLOCKDEV_CALLS = [
    "Disconnects the network of the given drbd devices"),
   ("drbd_attach_net", MULTI, None, constants.RPC_TMO_NORMAL, [
     ("disks", ED_DISKS_DICT_DP, None),
-    ("instance_name", None, None),
     ("multimaster", None, None),
     ], None, None, "Connects the given DRBD devices"),
   ("drbd_wait_sync", MULTI, None, constants.RPC_TMO_SLOW, [
@@ -442,6 +451,8 @@ _BLOCKDEV_CALLS = [
    " given amount"),
   ("blockdev_snapshot", SINGLE, None, constants.RPC_TMO_NORMAL, [
     ("cf_bdev", ED_SINGLE_DISK_DICT_DP, None),
+    ("snap_name", None, None),
+    ("snap_size", None, None),
     ], None, None, "Export a given disk to another node"),
   ("blockdev_rename", SINGLE, None, constants.RPC_TMO_NORMAL, [
     ("devlist", ED_BLOCKDEV_RENAME, None),
@@ -526,6 +537,37 @@ _NODE_CALLS = [
     ("daemon", None, "Daemon name"),
     ("run", None, "Whether the daemon should be running or stopped"),
     ], None, None, "Ensure daemon is running on the node."),
+  ("node_ssh_key_add", MULTI, None, constants.RPC_TMO_URGENT, [
+    ("node_uuid", None, "UUID of the node whose key is distributed"),
+    ("node_name", None, "Name of the node whose key is distributed"),
+    ("potential_master_candidates", None, "Potential master candidates"),
+    ("to_authorized_keys", None, "Whether the node's key should be added"
+     " to all nodes' 'authorized_keys' file"),
+    ("to_public_keys", None, "Whether the node's key should be added"
+     " to all nodes' public key file"),
+    ("get_public_keys", None, "Whether the node should get the other nodes'"
+     " public keys")],
+    None, None, "Distribute a new node's public SSH key on the cluster."),
+  ("node_ssh_key_remove", MULTI, None, constants.RPC_TMO_URGENT, [
+    ("node_uuid", None, "UUID of the node whose key is removed"),
+    ("node_name", None, "Name of the node whose key is removed"),
+    ("master_candidate_uuids", None, "List of UUIDs of master candidates."),
+    ("potential_master_candidates", None, "Potential master candidates"),
+    ("from_authorized_keys", None,
+     "If the key should be removed from the 'authorized_keys' file."),
+    ("from_public_keys", None,
+     "If the key should be removed from the public key file."),
+    ("clear_authorized_keys", None,
+     "If the 'authorized_keys' file of the node should be cleared."),
+    ("clear_public_keys", None,
+     "If the 'ganeti_pub_keys' file of the node should be cleared.")],
+    None, None, "Remove a node's SSH key from the other nodes' key files."),
+  ("node_ssh_keys_renew", MULTI, None, constants.RPC_TMO_SLOW, [
+    ("node_uuids", None, "UUIDs of the nodes whose key is renewed"),
+    ("node_names", None, "Names of the nodes whose key is renewed"),
+    ("master_candidate_uuids", None, "List of UUIDs of master candidates."),
+    ("potential_master_candidates", None, "Potential master candidates")],
+    None, None, "Renew all SSH key pairs of all nodes nodes."),
   ]
 
 _MISC_CALLS = [
