@@ -223,6 +223,16 @@ class NodeRequestHandler(http.server.HttpServerHandler):
     return backend.BlockdevCreate(bdev, size, owner, on_primary, info,
                                   excl_stor)
 
+  @staticmethod
+  def perspective_blockdev_convert(params):
+    """Copy data from source block device to target.
+
+    """
+    disk_src, disk_dest = params
+    bdev_src = objects.Disk.FromDict(disk_src)
+    bdev_dest = objects.Disk.FromDict(disk_dest)
+    return backend.BlockdevConvert(bdev_src, bdev_dest)
+
   @staticmethod
   def perspective_blockdev_pause_resume_sync(params):
     """Pause/resume sync of a block device.
@@ -368,13 +378,14 @@ class NodeRequestHandler(http.server.HttpServerHandler):
   def perspective_blockdev_snapshot(params):
     """Create a snapshot device.
 
-    Note that this is only valid for LVM disks, if we get passed
+    Note that this is only valid for LVM and ExtStorage disks, if we get passed
     something else we raise an exception. The snapshot device can be
     remove by calling the generic block device remove call.
 
     """
-    cfbd = objects.Disk.FromDict(params[0])
-    return backend.BlockdevSnapshot(cfbd)
+    (disk, snap_name, snap_size) = params
+    cfbd = objects.Disk.FromDict(disk)
+    return backend.BlockdevSnapshot(cfbd, snap_name, snap_size)
 
   @staticmethod
   def perspective_blockdev_grow(params):
@@ -399,6 +410,15 @@ class NodeRequestHandler(http.server.HttpServerHandler):
     disks = [objects.Disk.FromDict(cf) for cf in params[1]]
     return backend.BlockdevClose(params[0], disks)
 
+  @staticmethod
+  def perspective_blockdev_open(params):
+    """Opens the given block devices.
+
+    """
+    disks = [objects.Disk.FromDict(cf) for cf in params[1]]
+    exclusive = params[2]
+    return backend.BlockdevOpen(params[0], disks, exclusive)
+
   @staticmethod
   def perspective_blockdev_getdimensions(params):
     """Compute the sizes of the given block devices.
@@ -438,9 +458,9 @@ class NodeRequestHandler(http.server.HttpServerHandler):
     disk list must all be drbd devices.
 
     """
-    disks, instance_name, multimaster = params
+    disks, multimaster = params
     disks = [objects.Disk.FromDict(disk) for disk in disks]
-    return backend.DrbdAttachNet(disks, instance_name, multimaster)
+    return backend.DrbdAttachNet(disks, multimaster)
 
   @staticmethod
   def perspective_drbd_wait_sync(params):
@@ -907,6 +927,47 @@ class NodeRequestHandler(http.server.HttpServerHandler):
     (daemon_name, run) = params
     return backend.EnsureDaemon(daemon_name, run)
 
+  @staticmethod
+  def perspective_node_ssh_key_add(params):
+    """Distributes a new node's SSH key if authorized.
+
+    """
+    (node_uuid, node_name, potential_master_candidates,
+     to_authorized_keys, to_public_keys, get_public_keys) = params
+    return backend.AddNodeSshKey(node_uuid, node_name,
+                                 potential_master_candidates,
+                                 to_authorized_keys=to_authorized_keys,
+                                 to_public_keys=to_public_keys,
+                                 get_public_keys=get_public_keys)
+
+  @staticmethod
+  def perspective_node_ssh_keys_renew(params):
+    """Generates a new root SSH key pair on the node.
+
+    """
+    (node_uuids, node_names, master_candidate_uuids,
+     potential_master_candidates) = params
+    return backend.RenewSshKeys(node_uuids, node_names,
+                                master_candidate_uuids,
+                                potential_master_candidates)
+
+  @staticmethod
+  def perspective_node_ssh_key_remove(params):
+    """Removes a node's SSH key from the other nodes' SSH files.
+
+    """
+    (node_uuid, node_name,
+     master_candidate_uuids, potential_master_candidates,
+     from_authorized_keys, from_public_keys, clear_authorized_keys,
+     clear_public_keys) = params
+    return backend.RemoveNodeSshKey(node_uuid, node_name,
+                                    master_candidate_uuids,
+                                    potential_master_candidates,
+                                    from_authorized_keys=from_authorized_keys,
+                                    from_public_keys=from_public_keys,
+                                    clear_authorized_keys=clear_authorized_keys,
+                                    clear_public_keys=clear_public_keys)
+
   # cluster --------------------------
 
   @staticmethod
@@ -1051,7 +1112,10 @@ class NodeRequestHandler(http.server.HttpServerHandler):
     name, idata, ial_params_dict = params
     ial_params = []
     for ial_param in ial_params_dict.items():
-      ial_params.append("--" + ial_param[0] + "=" + ial_param[1])
+      if ial_param[1] is not None:
+        ial_params.append("--" + ial_param[0] + "=" + ial_param[1])
+      else:
+        ial_params.append("--" + ial_param[0])
     iar = backend.IAllocatorRunner()
     return iar.Run(name, idata, ial_params)
 
