@@ -653,6 +653,8 @@ class DRBD8Dev(base.BlockDev):
       ldisk_status = constants.LDS_OKAY
     elif stats.is_diskless:
       ldisk_status = constants.LDS_FAULTY
+    elif stats.is_in_resync:
+      ldisk_status = constants.LDS_SYNC
     else:
       ldisk_status = constants.LDS_UNKNOWN
 
@@ -664,7 +666,7 @@ class DRBD8Dev(base.BlockDev):
                                   is_degraded=is_degraded,
                                   ldisk_status=ldisk_status)
 
-  def Open(self, force=False):
+  def Open(self, force=False, exclusive=True):
     """Make the local state primary.
 
     If the 'force' parameter is given, DRBD is instructed to switch the device
@@ -1022,7 +1024,8 @@ class DRBD8Dev(base.BlockDev):
     # can be bigger
     result = utils.RunCmd([constants.DD_CMD,
                            "if=/dev/zero", "of=%s" % dev_path,
-                           "bs=1048576", "count=128", "oflag=direct"])
+                           "bs=%s" % constants.DD_BLOCK_SIZE, "count=128",
+                           "oflag=direct"])
     if result.failed:
       base.ThrowError("Can't wipe the meta device: %s", result.output)
 
