@@ -46,6 +46,7 @@ import qa_config
 import qa_daemon
 import qa_env
 import qa_error
+import qa_filters
 import qa_group
 import qa_instance
 import qa_iptables
@@ -446,6 +447,23 @@ def RunNetworkTests():
   RunTestIf(["network", "tags"], qa_network.TestNetworkTags)
 
 
+def RunFilterTests():
+  """Run tests for job filter management.
+
+  """
+  RunTestIf("filters", qa_filters.TestFilterList)
+  RunTestIf("filters", qa_filters.TestFilterListFields)
+  RunTestIf("filters", qa_filters.TestFilterAddRemove)
+  RunTestIf("filters", qa_filters.TestFilterReject)
+  RunTestIf("filters", qa_filters.TestFilterOpCode)
+  RunTestIf("filters", qa_filters.TestFilterReasonChain)
+  RunTestIf("filters", qa_filters.TestFilterContinue)
+  RunTestIf("filters", qa_filters.TestFilterAcceptPause)
+  RunTestIf("filters", qa_filters.TestFilterWatermark)
+  RunTestIf("filters", qa_filters.TestFilterRateLimit)
+  RunTestIf("filters", qa_filters.TestAdHocReasonRateLimit)
+
+
 def RunGroupRwTests():
   """Run tests for adding/removing/renaming groups.
 
@@ -468,8 +486,7 @@ def RunExportImportTests(instance, inodes):
   # FIXME: export explicitly bails out on file based storage. other non-lvm
   # based storage types are untested, though. Also note that import could still
   # work, but is deeply embedded into the "export" case.
-  if (qa_config.TestEnabled("instance-export") and
-      instance.disk_template not in constants.DTS_FILEBASED):
+  if qa_config.TestEnabled("instance-export"):
     RunTest(qa_instance.TestInstanceExportNoTarget, instance)
 
     pnode = inodes[0]
@@ -494,8 +511,7 @@ def RunExportImportTests(instance, inodes):
 
   # FIXME: inter-cluster-instance-move crashes on file based instances :/
   # See Issue 414.
-  if (qa_config.TestEnabled([qa_rapi.Enabled, "inter-cluster-instance-move"])
-      and (instance.disk_template not in constants.DTS_FILEBASED)):
+  if (qa_config.TestEnabled([qa_rapi.Enabled, "inter-cluster-instance-move"])):
     newinst = qa_config.AcquireInstance()
     try:
       tnode = qa_config.AcquireNode(exclude=inodes)
@@ -646,7 +662,7 @@ def RunCustomSshPortTests():
 
     for node in nodes:
       qa_node.NodeRemove(node)
-    with qa_iptables.RulesContext(nodes) as r:
+    with qa_iptables.RulesContext() as r:
       with qa_group.NewGroupCtx() as group:
         qa_group.ModifyGroupSshPort(r, group, nodes, port)
 
@@ -820,6 +836,9 @@ def IsExclusiveStorageInstanceTestEnabled():
 def RunInstanceTests():
   """Create and exercise instances."""
 
+  requested_conversions = qa_config.get("convert-disk-templates", [])
+  supported_conversions = \
+      set(requested_conversions).difference(constants.DTS_NOT_CONVERTIBLE_TO)
   for (test_name, templ, create_fun, num_nodes) in \
       qa_instance.available_instance_tests:
     if (qa_config.TestEnabled(test_name) and
@@ -843,11 +862,21 @@ def RunInstanceTests():
             RunTestIf("group-rwops", qa_group.TestAssignNodesIncludingSplit,
                       constants.INITIAL_NODE_GROUP_NAME,
                       inodes[0].primary, inodes[1].primary)
+          # This test will run once but it will cover all the supported
+          # user-provided disk template conversions
           if qa_config.TestEnabled("instance-convert-disk"):
-            RunTest(qa_instance.TestInstanceShutdown, instance)
-            RunTest(qa_instance.TestInstanceConvertDiskToPlain,
-                    instance, inodes)
-            RunTest(qa_instance.TestInstanceStartup, instance)
+            if (len(supported_conversions) > 1 and
+                instance.disk_template in supported_conversions):
+              RunTest(qa_instance.TestInstanceShutdown, instance)
+              RunTest(qa_instance.TestInstanceConvertDiskTemplate, instance,
+                      supported_conversions)
+              RunTest(qa_instance.TestInstanceStartup, instance)
+              # At this point we clear the set because the requested conversions
+              # has been tested
+              supported_conversions.clear()
+            else:
+              test_desc = "Converting instance of template %s" % templ
+              ReportTestSkip(test_desc, "conversion feature")
           RunTestIf("instance-modify-disks",
                     qa_instance.TestInstanceModifyDisks, instance)
           RunCommonInstanceTests(instance, inodes)
@@ -952,6 +981,7 @@ def RunQa():
   RunTestBlock(RunGroupListTests)
   RunTestBlock(RunGroupRwTests)
   RunTestBlock(RunNetworkTests)
+  RunTestBlock(RunFilterTests)
 
   # The master shouldn't be readded or put offline; "delay" needs a non-master
   # node to test
@@ -1036,6 +1066,9 @@ def RunQa():
   finally:
     pnode.Release()
 
+  if qa_rapi.Enabled():
+    RunTestIf("filters", qa_rapi.TestFilters)
+
   RunTestIf("cluster-upgrade", qa_cluster.TestUpgrade)
 
   RunTestBlock(RunExclusiveStorageTests)
