@@ -51,8 +51,8 @@ import qa_job_utils
 import qa_logging
 import qa_utils
 
-from qa_utils import AssertEqual, AssertCommand, GetCommandOutput, \
-                     CheckFileUnmodified
+from qa_utils import AssertEqual, AssertCommand, AssertRedirectedCommand, \
+  GetCommandOutput, CheckFileUnmodified
 
 
 # Prefix for LVM volumes created by QA code during tests
@@ -204,6 +204,22 @@ def TestClusterInit(rapi_user, rapi_secret):
   finally:
     fh.close()
 
+  # If we don't modify the SSH setup by Ganeti, we have to ensure connectivity
+  # before
+  if not qa_config.GetModifySshSetup():
+    (key_type, _, priv_key_file, pub_key_file, auth_key_path) = \
+      qa_config.GetSshConfig()
+    AssertCommand("echo -e 'y\n' | ssh-keygen -t %s -f %s -q -N ''"
+                  % (key_type, priv_key_file))
+    AssertCommand("cat %s >> %s" % (pub_key_file, auth_key_path))
+    for node in qa_config.get("nodes"):
+      if node != master:
+        for key_file in [priv_key_file, pub_key_file]:
+          AssertCommand("scp -oStrictHostKeyChecking=no %s %s:%s" %
+                        (key_file, node.primary, key_file))
+        AssertCommand("ssh %s \'cat %s >> %s\'"
+                      % (node.primary, pub_key_file, auth_key_path))
+
   # Initialize cluster
   enabled_disk_templates = qa_config.GetEnabledDiskTemplates()
   cmd = [
@@ -213,11 +229,25 @@ def TestClusterInit(rapi_user, rapi_secret):
     "--enabled-disk-templates=%s" %
       ",".join(enabled_disk_templates),
     ]
+
+  if not qa_config.GetModifySshSetup():
+    cmd.append("--no-ssh-init")
+
   if constants.DT_FILE in enabled_disk_templates:
     cmd.append(
         "--file-storage-dir=%s" %
         qa_config.get("default-file-storage-dir",
                       pathutils.DEFAULT_FILE_STORAGE_DIR))
+  if constants.DT_SHARED_FILE in enabled_disk_templates:
+    cmd.append(
+        "--shared-file-storage-dir=%s" %
+        qa_config.get("default-shared-file-storage-dir",
+                      pathutils.DEFAULT_SHARED_FILE_STORAGE_DIR))
+  if constants.DT_GLUSTER in enabled_disk_templates:
+    cmd.append(
+        "--gluster-storage-dir=%s" %
+        qa_config.get("default-gluster-storage-dir",
+                      pathutils.DEFAULT_GLUSTER_STORAGE_DIR))
 
   for spec_type in ("mem-size", "disk-size", "disk-count", "cpu-count",
                     "nic-count"):
@@ -1236,7 +1266,8 @@ def TestClusterRenewCrypto():
     AssertCommand(["gnt-cluster", "renew-crypto", "--force",
                    "--new-cluster-certificate", "--new-confd-hmac-key",
                    "--new-rapi-certificate", "--new-cluster-domain-secret",
-                   "--new-node-certificates"])
+                   "--new-node-certificates", "--new-ssh-keys",
+                   "--no-ssh-key-check"])
     _AssertSsconfCertFiles()
     AssertCommand(["gnt-cluster", "verify"])
 
@@ -1252,6 +1283,10 @@ def TestClusterRenewCrypto():
     _AssertSsconfCertFiles()
     AssertCommand(["gnt-cluster", "verify"])
 
+    # Only renew SSH keys
+    AssertCommand(["gnt-cluster", "renew-crypto", "--force",
+                   "--new-ssh-keys", "--no-ssh-key-check"])
+
     # Restore RAPI certificate
     AssertCommand(["gnt-cluster", "renew-crypto", "--force",
                    "--rapi-certificate=%s" % rapi_cert_backup])
@@ -1286,7 +1321,9 @@ def TestClusterBurnin():
     if len(instances) < 1:
       raise qa_error.Error("Burnin needs at least one instance")
 
-    script = qa_utils.UploadFile(master.primary, "../tools/burnin")
+    burnin_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
+                               "../tools/burnin")
+    script = qa_utils.UploadFile(master.primary, burnin_file)
     try:
       disks = qa_config.GetDiskOptions()
       # Run burnin
@@ -1370,8 +1407,8 @@ def TestUpgrade():
     nodes = qa_config.AcquireManyNodes(n)
     live_instances.append(cf(nodes))
 
-  AssertCommand(["gnt-cluster", "upgrade", "--to", other_version])
-  AssertCommand(["gnt-cluster", "verify"])
+  AssertRedirectedCommand(["gnt-cluster", "upgrade", "--to", other_version])
+  AssertRedirectedCommand(["gnt-cluster", "verify"])
 
   for instance in live_instances:
     qa_instance.TestInstanceRemove(instance)
@@ -1381,8 +1418,8 @@ def TestUpgrade():
     nodes = qa_config.AcquireManyNodes(n)
     live_instances.append(cf(nodes))
 
-  AssertCommand(["gnt-cluster", "upgrade", "--to", this_version])
-  AssertCommand(["gnt-cluster", "verify"])
+  AssertRedirectedCommand(["gnt-cluster", "upgrade", "--to", this_version])
+  AssertRedirectedCommand(["gnt-cluster", "verify"])
 
   for instance in live_instances:
     qa_instance.TestInstanceRemove(instance)
