@@ -647,6 +647,41 @@ class _QaConfig(object):
     else:
       return None
 
+  def GetModifySshSetup(self):
+    """Return whether to modify the SSH setup or not.
+
+    This specified whether Ganeti should create and distribute SSH
+    keys for the nodes or not. The default is 'True'.
+
+    """
+    if self.get("modify_ssh_setup") is None:
+      return True
+    else:
+      return self.get("modify_ssh_setup")
+
+  def GetSshConfig(self):
+    """Returns the SSH configuration necessary to create keys etc.
+
+    @rtype: tuple of (string, string, string, string, string)
+    @return: tuple containing key type ('dsa' or 'rsa'), ssh directory
+      (default '/root/.ssh/'), file path of the private key, file path
+      of the public key, file path of the 'authorized_keys' file
+
+    """
+    key_type = "dsa"
+    if self.get("ssh_key_type") is not None:
+      key_type = self.get("ssh_key_type")
+
+    ssh_dir = "/root/.ssh/"
+    if self.get("ssh_dir") is not None:
+      ssh_dir = self.get("ssh_dir")
+
+    priv_key_file = os.path.join(ssh_dir, "id_%s" % (key_type))
+    pub_key_file = "%s.pub" % priv_key_file
+    auth_key_file = os.path.join(ssh_dir, "authorized_keys")
+
+    return (key_type, ssh_dir, priv_key_file, pub_key_file, auth_key_file)
+
 
 def Load(path):
   """Loads the passed configuration file.
@@ -795,6 +830,20 @@ def GetDefaultDiskTemplate(*args):
   return GetConfig().GetDefaultDiskTemplate(*args)
 
 
+def GetModifySshSetup(*args):
+  """Wrapper for L{_QaConfig.GetModifySshSetup}.
+
+  """
+  return GetConfig().GetModifySshSetup(*args)
+
+
+def GetSshConfig(*args):
+  """Wrapper for L{_QaConfig.GetSshConfig}.
+
+  """
+  return GetConfig().GetSshConfig(*args)
+
+
 def GetMasterNode():
   """Wrapper for L{_QaConfig.GetMasterNode}.
 
