@@ -41,7 +41,8 @@ import qa_config
 import qa_error
 import qa_utils
 
-from qa_utils import AssertCommand, AssertEqual, AssertIn, GetCommandOutput
+from qa_utils import AssertCommand, AssertRedirectedCommand, AssertEqual, \
+  AssertIn, GetCommandOutput
 
 
 def NodeAdd(node, readd=False, group=None):
@@ -57,10 +58,17 @@ def NodeAdd(node, readd=False, group=None):
     cmd.append("--readd")
   if group is not None:
     cmd.extend(["--node-group", group])
+
+  if not qa_config.GetModifySshSetup():
+    cmd.append("--no-node-setup")
+
   cmd.append(node.primary)
 
   AssertCommand(cmd)
 
+  if readd:
+    AssertRedirectedCommand(["gnt-cluster", "verify"])
+
   if readd:
     assert node.added
   else:
@@ -261,6 +269,8 @@ def TestNodeModify(node):
   AssertCommand(["gnt-node", "modify", "--secondary-ip=%s" % node.secondary,
                  node.primary])
 
+  AssertRedirectedCommand(["gnt-cluster", "verify"])
+
 
 def _CreateOobScriptStructure():
   """Create a simple OOB handling script and its structure."""
@@ -491,6 +501,8 @@ def _BuildSetESCmd(action, value, node_name):
   cmd = ["gnt-node"]
   if action == "add":
     cmd.extend(["add", "--readd"])
+    if not qa_config.GetModifySshSetup():
+      cmd.append("--no-node-setup")
   else:
     cmd.append("modify")
   cmd.extend(["--node-parameters", "exclusive_storage=%s" % value, node_name])
