@@ -1,3 +1,5 @@
+{-# LANGUAGE ViewPatterns #-}
+
 {-| Implementation of the Ganeti configuration database.
 
 -}
@@ -51,6 +53,7 @@ module Ganeti.Config
     , getNode
     , getInstance
     , getDisk
+    , getFilterRule
     , getGroup
     , getGroupNdParams
     , getGroupIpolicy
@@ -63,6 +66,7 @@ module Ganeti.Config
     , getInstAllNodes
     , getInstDisks
     , getInstDisksFromObj
+    , getDrbdMinorsForDisk
     , getDrbdMinorsForInstance
     , getFilledInstHvParams
     , getFilledInstBeParams
@@ -79,10 +83,12 @@ module Ganeti.Config
     ) where
 
 import Control.Applicative
+import Control.Arrow ((&&&))
 import Control.Monad
 import Control.Monad.State
 import qualified Data.Foldable as F
 import Data.List (foldl', nub)
+import Data.Maybe (fromMaybe)
 import Data.Monoid
 import qualified Data.Map as M
 import qualified Data.Set as S
@@ -124,11 +130,16 @@ saveConfig fh = hPutStr fh . encodeConfig
 
 -- * Query functions
 
+-- | Annotate Nothing as missing parameter and apply the given
+-- transformation otherwise
+withMissingParam :: String -> (a -> ErrorResult b) -> Maybe a -> ErrorResult b
+withMissingParam = maybe . Bad . ParameterError
+
 -- | Computes the nodes covered by a disk.
 computeDiskNodes :: Disk -> S.Set String
 computeDiskNodes dsk =
   case diskLogicalId dsk of
-    LIDDrbd8 nodeA nodeB _ _ _ _ -> S.fromList [nodeA, nodeB]
+    Just (LIDDrbd8 nodeA nodeB _ _ _ _) -> S.fromList [nodeA, nodeB]
     _ -> S.empty
 
 -- | Computes all disk-related nodes of an instance. For non-DRBD,
@@ -142,21 +153,22 @@ instDiskNodes cfg inst =
 
 -- | Computes all nodes of an instance.
 instNodes :: ConfigData -> Instance -> S.Set String
-instNodes cfg inst = instPrimaryNode inst `S.insert` instDiskNodes cfg inst
+instNodes cfg inst = maybe id S.insert (instPrimaryNode inst)
+                      $ instDiskNodes cfg inst
 
 -- | Computes the secondary nodes of an instance. Since this is valid
 -- only for DRBD, we call directly 'instDiskNodes', skipping over the
 -- extra primary insert.
 instSecondaryNodes :: ConfigData -> Instance -> S.Set String
 instSecondaryNodes cfg inst =
-  instPrimaryNode inst `S.delete` instDiskNodes cfg inst
+  maybe id S.delete (instPrimaryNode inst) $ instDiskNodes cfg inst
 
 -- | Get instances of a given node.
 -- The node is specified through its UUID.
 getNodeInstances :: ConfigData -> String -> ([Instance], [Instance])
 getNodeInstances cfg nname =
     let all_inst = M.elems . fromContainer . configInstances $ cfg
-        pri_inst = filter ((== nname) . instPrimaryNode) all_inst
+        pri_inst = filter ((== Just nname) . instPrimaryNode) all_inst
         sec_inst = filter ((nname `S.member`) . instSecondaryNodes cfg) all_inst
     in (pri_inst, sec_inst)
 
@@ -255,8 +267,10 @@ getInstance cfg name =
   in case getItem "Instance" name instances of
        -- if not found by uuid, we need to look it up by name
        Ok inst -> Ok inst
-       Bad _ -> let by_name = M.mapKeys
-                              (instName . (M.!) instances) instances
+       Bad _ -> let by_name =
+                      M.delete ""
+                      . M.mapKeys (fromMaybe "" . instName . (M.!) instances)
+                      $ instances
                 in getItem "Instance" name by_name
 
 -- | Looks up a disk by uuid.
@@ -265,6 +279,12 @@ getDisk cfg name =
   let disks = fromContainer (configDisks cfg)
   in getItem "Disk" name disks
 
+-- | Looks up a filter by uuid.
+getFilterRule :: ConfigData -> String -> ErrorResult FilterRule
+getFilterRule cfg name =
+  let filters = fromContainer (configFilters cfg)
+  in getItem "Filter" name filters
+
 -- | Looks up a node group by name or uuid.
 getGroup :: ConfigData -> String -> ErrorResult NodeGroup
 getGroup cfg name =
@@ -279,12 +299,12 @@ getGroup cfg name =
 -- | Computes a node group's node params.
 getGroupNdParams :: ConfigData -> NodeGroup -> FilledNDParams
 getGroupNdParams cfg ng =
-  fillNDParams (clusterNdparams $ configCluster cfg) (groupNdparams ng)
+  fillParams (clusterNdparams $ configCluster cfg) (groupNdparams ng)
 
 -- | Computes a node group's ipolicy.
 getGroupIpolicy :: ConfigData -> NodeGroup -> FilledIPolicy
 getGroupIpolicy cfg ng =
-  fillIPolicy (clusterIpolicy $ configCluster cfg) (groupIpolicy ng)
+  fillParams (clusterIpolicy $ configCluster cfg) (groupIpolicy ng)
 
 -- | Computes a group\'s (merged) disk params.
 getGroupDiskParams :: ConfigData -> NodeGroup -> GroupDiskParams
@@ -311,14 +331,19 @@ getGroupInstances cfg gname =
 getFilledInstHvParams :: [String] -> ConfigData -> Instance -> HvParams
 getFilledInstHvParams globals cfg inst =
   -- First get the defaults of the parent
-  let hvName = instHypervisor inst
+  let maybeHvName = instHypervisor inst
       hvParamMap = fromContainer . clusterHvparams $ configCluster cfg
-      parentHvParams = maybe M.empty fromContainer $ M.lookup hvName hvParamMap
+      parentHvParams =
+        maybe M.empty fromContainer (maybeHvName >>= flip M.lookup hvParamMap)
   -- Then the os defaults for the given hypervisor
-      osName = instOs inst
+      maybeOsName = instOs inst
       osParamMap = fromContainer . clusterOsHvp $ configCluster cfg
-      osHvParamMap = maybe M.empty fromContainer $ M.lookup osName osParamMap
-      osHvParams = maybe M.empty fromContainer $ M.lookup hvName osHvParamMap
+      osHvParamMap =
+        maybe M.empty (maybe M.empty fromContainer . flip M.lookup osParamMap)
+          maybeOsName
+      osHvParams =
+        maybe M.empty (maybe M.empty fromContainer . flip M.lookup osHvParamMap)
+          maybeHvName
   -- Then the child
       childHvParams = fromContainer . instHvparams $ inst
   -- Helper function
@@ -331,16 +356,18 @@ getFilledInstBeParams :: ConfigData -> Instance -> ErrorResult FilledBeParams
 getFilledInstBeParams cfg inst = do
   let beParamMap = fromContainer . clusterBeparams . configCluster $ cfg
   parentParams <- getItem "FilledBeParams" C.ppDefault beParamMap
-  return $ fillBeParams parentParams (instBeparams inst)
+  return $ fillParams parentParams (instBeparams inst)
 
 -- | Retrieves the instance os params, missing values filled with cluster
 -- defaults. This does NOT include private and secret parameters.
 getFilledInstOsParams :: ConfigData -> Instance -> OsParams
 getFilledInstOsParams cfg inst =
-  let osLookupName = takeWhile (/= '+') (instOs inst)
+  let maybeOsLookupName = liftM (takeWhile (/= '+')) (instOs inst)
       osParamMap = fromContainer . clusterOsparams $ configCluster cfg
       childOsParams = instOsparams inst
-  in case getItem "OsParams" osLookupName osParamMap of
+  in case withMissingParam "Instance without OS"
+            (flip (getItem "OsParams") osParamMap)
+            maybeOsLookupName of
        Ok parentOsParams -> GenericContainer $
                               fillDict (fromContainer parentOsParams)
                                        (fromContainer childOsParams) []
@@ -349,13 +376,15 @@ getFilledInstOsParams cfg inst =
 -- | Looks up an instance's primary node.
 getInstPrimaryNode :: ConfigData -> String -> ErrorResult Node
 getInstPrimaryNode cfg name =
-  liftM instPrimaryNode (getInstance cfg name) >>= getNode cfg
+  getInstance cfg name
+  >>= withMissingParam "Instance without primary node" return . instPrimaryNode
+  >>= getNode cfg
 
 -- | Retrieves all nodes hosting a DRBD disk
 getDrbdDiskNodes :: ConfigData -> Disk -> [Node]
 getDrbdDiskNodes cfg disk =
   let retrieved = case diskLogicalId disk of
-                    LIDDrbd8 nodeA nodeB _ _ _ _ ->
+                    Just (LIDDrbd8 nodeA nodeB _ _ _ _) ->
                       justOk [getNode cfg nodeA, getNode cfg nodeB]
                     _                            -> []
   in retrieved ++ concatMap (getDrbdDiskNodes cfg) (diskChildren disk)
@@ -367,9 +396,9 @@ getDrbdDiskNodes cfg disk =
 getInstAllNodes :: ConfigData -> String -> ErrorResult [Node]
 getInstAllNodes cfg name = do
   inst_disks <- getInstDisks cfg name
-  let diskNodes = concatMap (getDrbdDiskNodes cfg) inst_disks
+  let disk_nodes = concatMap (getDrbdDiskNodes cfg) inst_disks
   pNode <- getInstPrimaryNode cfg name
-  return . nub $ pNode:diskNodes
+  return . nub $ pNode:disk_nodes
 
 -- | Get disks for a given instance.
 -- The instance is specified by name or uuid.
@@ -390,9 +419,9 @@ collectFromDrbdDisks
   -> Disk -> a
 collectFromDrbdDisks f = col
   where
-    col Disk { diskLogicalId = (LIDDrbd8 nA nB port mA mB secret)
-             , diskChildren = ch
-             } = f nA nB port mA mB secret <> F.foldMap col ch
+    col (diskLogicalId &&& diskChildren ->
+           (Just (LIDDrbd8 nA nB port mA mB secret), ch)) =
+             f nA nB port mA mB secret <> F.foldMap col ch
     col d = F.foldMap col (diskChildren d)
 
 -- | Returns the DRBD secrets of a given 'Disk'
@@ -410,7 +439,7 @@ getDrbdMinorsForNode node disk =
   let child_minors = concatMap (getDrbdMinorsForNode node) (diskChildren disk)
       this_minors =
         case diskLogicalId disk of
-          LIDDrbd8 nodeA nodeB _ minorA minorB _
+          Just (LIDDrbd8 nodeA nodeB _ minorA minorB _)
             | nodeA == node -> [(minorA, nodeB)]
             | nodeB == node -> [(minorB, nodeA)]
           _ -> []
@@ -437,10 +466,10 @@ getInstMinorsForNode :: ConfigData
                      -> Instance
                      -> [(String, Int, String, String, String, String)]
 getInstMinorsForNode cfg node inst =
-  let role = if node == instPrimaryNode inst
+  let role = if Just node == instPrimaryNode inst
                then rolePrimary
                else roleSecondary
-      iname = instName inst
+      iname = fromMaybe "" $ instName inst
       inst_disks = case getInstDisksFromObj cfg inst of
                      Ok disks -> disks
                      Bad _ -> []
@@ -453,6 +482,7 @@ getInstMinorsForNode cfg node inst =
      zip [(0::Int)..] $ inst_disks
 
 -- | Builds link -> ip -> instname map.
+-- For instances without a name, we insert the uuid instead.
 --
 -- TODO: improve this by splitting it into multiple independent functions:
 --
@@ -466,11 +496,12 @@ buildLinkIpInstnameMap cfg =
   let cluster = configCluster cfg
       instances = M.elems . fromContainer . configInstances $ cfg
       defparams = (M.!) (fromContainer $ clusterNicparams cluster) C.ppDefault
-      nics = concatMap (\i -> [(instName i, nic) | nic <- instNics i])
+      nics = concatMap (\i -> [(fromMaybe (instUuid i) $ instName i, nic)
+                                | nic <- instNics i])
              instances
   in foldl' (\accum (iname, nic) ->
                let pparams = nicNicparams nic
-                   fparams = fillNicParams defparams pparams
+                   fparams = fillParams defparams pparams
                    link = nicpLink fparams
                in case nicIp nic of
                     Nothing -> accum
@@ -492,7 +523,7 @@ getNodeNdParams :: ConfigData -> Node -> Maybe FilledNDParams
 getNodeNdParams cfg node = do
   group <- getGroupOfNode cfg node
   let gparams = getGroupNdParams cfg group
-  return $ fillNDParams gparams (nodeNdparams node)
+  return $ fillParams gparams (nodeNdparams node)
 
 -- * Network
 
@@ -531,19 +562,20 @@ type NodeLVsMap = MM.MultiMap String LogicalVolume
 
 getInstanceLVsByNode :: ConfigData -> Instance -> ErrorResult NodeLVsMap
 getInstanceLVsByNode cd inst =
-    (MM.fromList . lvsByNode (instPrimaryNode inst))
-    <$> getInstDisksFromObj cd inst
+    withMissingParam "Instance without Primary Node"
+      (\i -> return $ MM.fromList . lvsByNode i)
+      (instPrimaryNode inst)
+    <*> getInstDisksFromObj cd inst
   where
     lvsByNode :: String -> [Disk] -> [(String, LogicalVolume)]
     lvsByNode node = concatMap (lvsByNode1 node)
     lvsByNode1 :: String -> Disk -> [(String, LogicalVolume)]
-    lvsByNode1 _    Disk { diskLogicalId = (LIDDrbd8 nA nB _ _ _ _)
-                         , diskChildren = ch
-                         } = lvsByNode nA ch ++ lvsByNode nB ch
-    lvsByNode1 node Disk { diskLogicalId = (LIDPlain lv) }
-                           = [(node, lv)]
-    lvsByNode1 node Disk { diskChildren = ch }
-                           = lvsByNode node ch
+    lvsByNode1 _ (diskLogicalId &&& diskChildren
+                   -> (Just (LIDDrbd8 nA nB _ _ _ _), ch)) =
+                         lvsByNode nA ch ++ lvsByNode nB ch
+    lvsByNode1 node (diskLogicalId -> (Just (LIDPlain lv))) =
+                         [(node, lv)]
+    lvsByNode1 node (diskChildren -> ch) = lvsByNode node ch
 
 getAllLVs :: ConfigData -> ErrorResult (S.Set LogicalVolume)
 getAllLVs cd = mconcat <$> mapM (liftM MM.values . getInstanceLVsByNode cd)
