@@ -37,6 +37,9 @@ module Ganeti.HTools.Backend.IAlloc
   , runIAllocator
   , processRelocate
   , loadData
+  , formatAllocate
+  , formatIAllocResult
+  , formatMultiAlloc
   ) where
 
 import Data.Either ()
@@ -49,12 +52,14 @@ import Text.JSON (JSObject, JSValue(JSArray),
 
 import Ganeti.BasicTypes
 import qualified Ganeti.HTools.Cluster as Cluster
+import qualified Ganeti.HTools.Cluster.Evacuate as Evacuate
 import qualified Ganeti.HTools.Container as Container
 import qualified Ganeti.HTools.Group as Group
 import qualified Ganeti.HTools.Node as Node
 import qualified Ganeti.HTools.Instance as Instance
 import qualified Ganeti.HTools.Nic as Nic
 import qualified Ganeti.Constants as C
+import Ganeti.HTools.AlgorithmParams (AlgorithmOptions)
 import Ganeti.HTools.CLI
 import Ganeti.HTools.Loader
 import Ganeti.HTools.Types
@@ -107,9 +112,12 @@ parseBaseInstance n a = do
   su    <- extract "spindle_use"
   nics  <- extract "nics" >>= toArray >>= asObjectList >>=
            mapM (parseNic n . fromJSObject)
+  -- Not forthcoming by default.
+  forthcoming <- extract "forthcoming" `orElse` Ok False
   return
     (n,
-     Instance.create n mem disk disks vcpus Running tags True 0 0 dt su nics)
+     Instance.create n mem disk disks vcpus Running tags True 0 0 dt su nics
+                     forthcoming)
 
 -- | Parses an instance as found in the cluster instance list.
 parseInstance :: NameAssoc -- ^ The node name-to-index association list
@@ -220,10 +228,11 @@ parseData now body = do
       _ | optype == C.iallocatorModeAlloc ->
             do
               rname     <- extrReq "name"
+              rgn       <- maybeFromObj request "group_name"
               req_nodes <- extrReq "required_nodes"
               inew      <- parseBaseInstance rname request
               let io = updateExclTags (extractExTags ctags) $ snd inew
-              return $ Allocate io req_nodes
+              return $ Allocate io (Cluster.AllocDetails req_nodes rgn)
         | optype == C.iallocatorModeReloc ->
             do
               rname     <- extrReq "name"
@@ -252,14 +261,16 @@ parseData now body = do
             do
               arry <- extrReq "instances" :: Result [JSObject JSValue]
               let inst_reqs = map fromJSObject arry
-              prqs <- mapM (\r ->
+              prqs <- forM inst_reqs (\r ->
                                do
                                  rname     <- extrFromReq r "name"
+                                 rgn       <- maybeFromObj request "group_name"
                                  req_nodes <- extrFromReq r "required_nodes"
                                  inew      <- parseBaseInstance rname r
                                  let io = updateExclTags (extractExTags ctags)
                                             $ snd inew
-                                 return (io, req_nodes)) inst_reqs
+                                 return (io, Cluster.AllocDetails
+                                               req_nodes rgn))
               return $ MultiAllocate prqs
         | otherwise -> fail ("Invalid request type '" ++ optype ++ "'")
   return (msgs, Request rqtype cdata)
@@ -276,11 +287,13 @@ formatResponse success info result =
   in encodeStrict $ makeObj [e_success, e_info, e_result]
 
 -- | Flatten the log of a solution into a string.
-describeSolution :: Cluster.AllocSolution -> String
+describeSolution :: Cluster.GenericAllocSolution a -> String
 describeSolution = intercalate ", " . Cluster.asLog
 
 -- | Convert allocation/relocation results into the result format.
-formatAllocate :: Instance.List -> Cluster.AllocSolution -> Result IAllocResult
+formatAllocate :: Instance.List
+               -> Cluster.GenericAllocSolution a
+               -> Result IAllocResult
 formatAllocate il as = do
   let info = describeSolution as
   case Cluster.asSolution as of
@@ -291,7 +304,8 @@ formatAllocate il as = do
         return (info, showJSON $ map Node.name nodes, nl, il')
 
 -- | Convert multi allocation results into the result format.
-formatMultiAlloc :: (Node.List, Instance.List, Cluster.AllocSolutionList)
+formatMultiAlloc :: ( Node.List, Instance.List
+                    , Cluster.GenericAllocSolutionList a)
                  -> Result IAllocResult
 formatMultiAlloc (fin_nl, fin_il, ars) =
   let rars = reverse ars
@@ -312,20 +326,20 @@ formatMultiAlloc (fin_nl, fin_il, ars) =
 formatNodeEvac :: Group.List
                -> Node.List
                -> Instance.List
-               -> (Node.List, Instance.List, Cluster.EvacSolution)
+               -> (Node.List, Instance.List, Evacuate.EvacSolution)
                -> Result IAllocResult
 formatNodeEvac gl nl il (fin_nl, fin_il, es) =
   let iname = Instance.name . flip Container.find il
       nname = Node.name . flip Container.find nl
       gname = Group.name . flip Container.find gl
-      fes = map (\(idx, msg) -> (iname idx, msg)) $ Cluster.esFailed es
+      fes = map (\(idx, msg) -> (iname idx, msg)) $ Evacuate.esFailed es
       mes = map (\(idx, gdx, ndxs) -> (iname idx, gname gdx, map nname ndxs))
-            $ Cluster.esMoved es
+            $ Evacuate.esMoved es
       failed = length fes
       moved  = length mes
       info = show failed ++ " instances failed to move and " ++ show moved ++
              " were moved successfully"
-  in Ok (info, showJSON (mes, fes, Cluster.esOpCodes es), fin_nl, fin_il)
+  in Ok (info, showJSON (mes, fes, Evacuate.esOpCodes es), fin_nl, fin_il)
 
 -- | Runs relocate for a single instance.
 --
@@ -334,14 +348,15 @@ formatNodeEvac gl nl il (fin_nl, fin_il, es) =
 -- result it got (in the nodes field) is actually consistent, as
 -- tryNodeEvac is designed to output primarily an opcode list, not a
 -- node list.
-processRelocate :: Group.List      -- ^ The group list
+processRelocate :: AlgorithmOptions
+                -> Group.List      -- ^ The group list
                 -> Node.List       -- ^ The node list
                 -> Instance.List   -- ^ The instance list
                 -> Idx             -- ^ The index of the instance to move
                 -> Int             -- ^ The number of nodes required
                 -> [Ndx]           -- ^ Nodes which should not be used
                 -> Result (Node.List, Instance.List, [Ndx]) -- ^ Solution list
-processRelocate gl nl il idx 1 exndx = do
+processRelocate opts gl nl il idx 1 exndx = do
   let orig = Container.find idx il
       sorig = Instance.sNode orig
       porig = Instance.pNode orig
@@ -357,12 +372,12 @@ processRelocate gl nl il idx 1 exndx = do
        fail $ "Unsupported request: excluded nodes not equal to\
               \ instance's " ++  node_type ++ "(" ++ show exp_node
               ++ " versus " ++ show exndx ++ ")"
-  (nl', il', esol) <- Cluster.tryNodeEvac gl nl il reloc_type [idx]
-  nodes <- case lookup idx (Cluster.esFailed esol) of
+  (nl', il', esol) <- Evacuate.tryNodeEvac opts gl nl il reloc_type [idx]
+  nodes <- case lookup idx (Evacuate.esFailed esol) of
              Just msg -> fail msg
              Nothing ->
                  case lookup idx (map (\(a, _, b) -> (a, b))
-                                  (Cluster.esMoved esol)) of
+                                  (Evacuate.esMoved esol)) of
                    Nothing ->
                        fail "Internal error: lost instance idx during move"
                    Just n -> return n
@@ -393,7 +408,7 @@ processRelocate gl nl il idx 1 exndx = do
                  show nodes ++ ") versus instance node (" ++ show pnode ++ ")"
   return (nl', il', nodes')
 
-processRelocate _ _ _ _ reqn _ =
+processRelocate _ _ _ _ _ reqn _ =
   fail $ "Exchange " ++ show reqn ++ " nodes mode is not implemented"
 
 formatRelocate :: (Node.List, Instance.List, [Ndx])
@@ -404,22 +419,24 @@ formatRelocate (nl, il, ndxs) =
   in Ok ("success", showJSON names, nl, il)
 
 -- | Process a request and return new node lists.
-processRequest :: Request -> Result IAllocResult
-processRequest request =
+processRequest :: AlgorithmOptions -> Request -> Result IAllocResult
+processRequest opts request =
   let Request rqtype (ClusterData gl nl il _ _) = request
   in case rqtype of
-       Allocate xi reqn ->
-         Cluster.tryMGAlloc gl nl il xi reqn >>= formatAllocate il
+       Allocate xi (Cluster.AllocDetails reqn Nothing) ->
+         Cluster.tryMGAlloc opts gl nl il xi reqn >>= formatAllocate il
+       Allocate xi (Cluster.AllocDetails reqn (Just gn)) ->
+         Cluster.tryGroupAlloc opts gl nl il gn xi reqn >>= formatAllocate il
        Relocate idx reqn exnodes ->
-         processRelocate gl nl il idx reqn exnodes >>= formatRelocate
+         processRelocate opts gl nl il idx reqn exnodes >>= formatRelocate
        ChangeGroup gdxs idxs ->
-         Cluster.tryChangeGroup gl nl il idxs gdxs >>=
+         Cluster.tryChangeGroup opts gl nl il idxs gdxs >>=
                 formatNodeEvac gl nl il
        NodeEvacuate xi mode ->
-         Cluster.tryNodeEvac gl nl il mode xi >>=
+         Evacuate.tryNodeEvac opts gl nl il mode xi >>=
                 formatNodeEvac gl nl il
        MultiAllocate xies ->
-         Cluster.allocList gl nl il xies [] >>= formatMultiAlloc
+         Cluster.allocList opts gl nl il xies [] >>= formatMultiAlloc
 
 -- | Reads the request from the data file(s).
 readRequest :: FilePath -> IO Request
@@ -432,17 +449,23 @@ readRequest fp = do
     Bad err -> exitErr err
     Ok (fix_msgs, rq) -> maybeShowWarnings fix_msgs >> return rq
 
--- | Main iallocator pipeline.
-runIAllocator :: Request -> (Maybe (Node.List, Instance.List), String)
-runIAllocator request =
+-- | Format an IAlloc result to maybe the new cluster and a response.
+formatIAllocResult :: Result IAllocResult
+                   -> (Maybe (Node.List, Instance.List), String)
+formatIAllocResult iallocResult =
   let (ok, info, result, cdata) =
-        case processRequest request of
+        case iallocResult of
           Ok (msg, r, nl, il) -> (True, "Request successful: " ++ msg, r,
                                   Just (nl, il))
           Bad msg -> (False, "Request failed: " ++ msg, JSArray [], Nothing)
       rstring = formatResponse ok info result
   in (cdata, rstring)
 
+-- | Main iallocator pipeline.
+runIAllocator :: AlgorithmOptions
+              -> Request -> (Maybe (Node.List, Instance.List), String)
+runIAllocator opts request = formatIAllocResult $ processRequest opts request
+
 -- | Load the data from an iallocation request file
 loadData :: FilePath -- ^ The path to the file
          -> IO (Result ClusterData)
