@@ -149,9 +149,11 @@ parseInstance ktn a = do
   auto_balance <- extract "auto_balance" beparams
   dt <- extract "disk_template" a
   su <- extract "spindle_use" beparams
+  -- Not forthcoming by default.
+  forthcoming <- extract "forthcoming" a `orElse` Ok False
   let disks = zipWith Instance.Disk dsizes dspindles
   let inst = Instance.create name mem disk disks vcpus running tags
-             auto_balance pnode snode dt su []
+             auto_balance pnode snode dt su [] forthcoming
   return (name, inst)
 
 -- | Construct a node from a JSON object.
