@@ -1,4 +1,4 @@
-{-# LANGUAGE TemplateHaskell, RankNTypes #-}
+{-# LANGUAGE RankNTypes #-}
 {-| Implementation of a reader for the job queue.
 
 -}
@@ -38,13 +38,17 @@ module Ganeti.JQScheduler
   , jqLivelock
   , jqForkLock
   , emptyJQStatus
+  , selectJobsToRun
+  , scheduleSomeJobs
   , initJQScheduler
   , enqueueNewJobs
   , dequeueJob
   , setJobPriority
+  , cleanupIfDead
+  , configChangeNeedsRescheduling
   ) where
 
-import Control.Applicative (liftA2)
+import Control.Applicative (liftA2, (<$>))
 import Control.Arrow
 import Control.Concurrent
 import Control.Exception
@@ -55,13 +59,20 @@ import Data.Functor ((<$))
 import Data.IORef
 import Data.List
 import Data.Maybe
+import qualified Data.Map as Map
+import Data.Ord (comparing)
+import Data.Set (Set)
 import qualified Data.Set as S
 import System.INotify
 
 import Ganeti.BasicTypes
 import Ganeti.Constants as C
 import Ganeti.Errors
+import Ganeti.JQScheduler.Filtering (applyingFilter, jobFiltering)
+import Ganeti.JQScheduler.Types
+import Ganeti.JQScheduler.ReasonRateLimiting (reasonRateLimit)
 import Ganeti.JQueue as JQ
+import Ganeti.JSON (fromContainer)
 import Ganeti.Lens hiding (chosen)
 import Ganeti.Logging
 import Ganeti.Objects
@@ -71,21 +82,6 @@ import Ganeti.Utils
 import Ganeti.Utils.Livelock
 import Ganeti.Utils.MVarLock
 
-data JobWithStat = JobWithStat { jINotify :: Maybe INotify
-                               , jStat :: FStat
-                               , jJob :: QueuedJob
-                               }
-
-$(makeCustomLenses' ''JobWithStat ['jJob])
-
-data Queue = Queue { qEnqueued :: [JobWithStat]
-                   , qRunning :: [JobWithStat]
-                   , qManipulated :: [JobWithStat] -- ^ running jobs that are
-                                                   -- being manipulated by
-                                                   -- some thread
-                   }
-
-$(makeCustomLenses ''Queue)
 
 {-| Representation of the job queue
 
@@ -307,10 +303,19 @@ jobEligible queue jWS =
 
 -- | Decide on which jobs to schedule next for execution. This is the
 -- pure function doing the scheduling.
-selectJobsToRun :: Int -> Queue -> (Queue, [JobWithStat])
-selectJobsToRun count queue =
+selectJobsToRun :: Int  -- ^ How many jobs are allowed to run at the
+                        -- same time.
+                -> Set FilterRule -- ^ Filter rules to respect for scheduling
+                -> Queue
+                -> (Queue, [JobWithStat])
+selectJobsToRun count filters queue =
   let n = count - length (qRunning queue) - length (qManipulated queue)
-      chosen = take n . filter (jobEligible queue) $ qEnqueued queue
+      chosen = take n
+               . jobFiltering queue filters
+               . reasonRateLimit queue
+               . sortBy (comparing (calcJobPriority . jJob))
+               . filter (jobEligible queue)
+               $ qEnqueued queue
       remain = deleteFirstsBy ((==) `on` (qjId . jJob)) (qEnqueued queue) chosen
   in (queue {qEnqueued=remain, qRunning=qRunning queue ++ chosen}, chosen)
 
@@ -325,16 +330,6 @@ logFailedJobs jobs = do
                ++ show (map snd jobs)
   return jids
 
--- | Requeue jobs that were previously selected for execution
--- but couldn't be started.
-requeueJobs :: JQStatus -> [(JobWithStat, GanetiException)] -> IO ()
-requeueJobs qstate jobs = do
-  jids <- logFailedJobs jobs
-  let rmJobs = filter ((`S.notMember` jids) . qjId . jJob)
-  logWarning "Rescheduling jobs"
-  modifyJobs qstate $ onQueuedJobs (map fst jobs ++)
-                      . onRunningJobs rmJobs
-
 -- | Fail jobs that were previously selected for execution
 -- but couldn't be started.
 failJobs :: ConfigData -> JQStatus -> [(JobWithStat, GanetiException)]
@@ -359,24 +354,72 @@ failJobs cfg qstate jobs = do
   mapM_ (runResultT . failAndSaveJobWithStat) jobs
   logDebug $ "Failed jobs " ++ sjobs
 
+
+-- | Checks if any jobs match a REJECT filter rule, and cancels them.
+cancelRejectedJobs :: JQStatus -> ConfigData -> Set FilterRule -> IO ()
+cancelRejectedJobs qstate cfg filters = do
+
+  enqueuedJobs <- map jJob . qEnqueued <$> readIORef (jqJobs qstate)
+
+  -- Determine which jobs are rejected.
+  let jobsToCancel =
+        [ (job, fr) | job <- enqueuedJobs
+                    , Just fr <- [applyingFilter filters job]
+                    , frAction fr == Reject ]
+
+  -- Cancel them.
+  qDir <- queueDir
+  forM_ jobsToCancel $ \(job, fr) -> do
+    let jid = qjId job
+    logDebug $ "Cancelling job " ++ show (fromJobId jid)
+               ++ " because it was REJECTed by filter rule " ++ frUuid fr
+    -- First dequeue, then cancel.
+    dequeueResult <- dequeueJob qstate jid
+    case dequeueResult of
+      Ok True -> do
+        now <- currentTimestamp
+        r <- runResultT
+               $ writeAndReplicateJob cfg qDir (cancelQueuedJob now job)
+        case r of
+          Ok _ -> return ()
+          Bad err -> logError $
+            "Failed to write config when cancelling job: " ++ err
+      Ok False -> do
+        logDebug $ "Job " ++ show (fromJobId jid)
+                   ++ " not queued; trying to cancel directly"
+        _ <- cancelJob False (jqLivelock qstate) jid  -- sigTERM-kill only
+        return ()
+      Bad s -> logError s -- passing a nonexistent job ID is an error here
+
+
 -- | Schedule jobs to be run. This is the IO wrapper around the
 -- pure `selectJobsToRun`.
 scheduleSomeJobs :: JQStatus -> IO ()
 scheduleSomeJobs qstate = do
-  count <- getMaxRunningJobs qstate
-  chosen <- atomicModifyIORef (jqJobs qstate) (selectJobsToRun count)
-  let jobs = map jJob chosen
-  unless (null chosen) . logInfo . (++) "Starting jobs: " . commaJoin
-    $ map (show . fromJobId . qjId) jobs
-  mapM_ (attachWatcher qstate) chosen
   cfgR <- readIORef (jqConfig qstate)
   case cfgR of
     Bad err -> do
       let msg = "Configuration unavailable: " ++ err
       logError msg
-      requeueJobs qstate . map (\x -> (x, strMsg msg)) $ chosen
     Ok cfg -> do
-      result <- JQ.startJobs cfg (jqLivelock qstate) (jqForkLock qstate) jobs
+      let filters = S.fromList . Map.elems . fromContainer $ configFilters cfg
+
+      -- Check if jobs are rejected by a REJECT filter, and cancel them.
+      cancelRejectedJobs qstate cfg filters
+
+      -- Select the jobs to run.
+      count <- getMaxRunningJobs qstate
+      chosen <- atomicModifyIORef (jqJobs qstate)
+                                  (selectJobsToRun count filters)
+      let jobs = map jJob chosen
+      unless (null chosen) . logInfo . (++) "Starting jobs: " . commaJoin
+        $ map (show . fromJobId . qjId) jobs
+
+      -- Attach the watcher.
+      mapM_ (attachWatcher qstate) chosen
+
+      -- Start the jobs.
+      result <- JQ.startJobs (jqLivelock qstate) (jqForkLock qstate) jobs
       let badWith (x, Bad y) = Just (x, y)
           badWith _          = Nothing
       let failed = mapMaybe badWith $ zip chosen result
@@ -389,8 +432,9 @@ showQueue (Queue {qEnqueued=waiting, qRunning=running}) =
   in "Waiting jobs: " ++ showids waiting 
        ++ "; running jobs: " ++ showids running
 
--- | Check if a job died, and clean up if so.
-checkForDeath :: JQStatus -> JobWithStat -> IO ()
+-- | Check if a job died, and clean up if so. Return True, if
+-- the job was found dead.
+checkForDeath :: JQStatus -> JobWithStat -> IO Bool
 checkForDeath state jobWS = do
   let job = jJob jobWS
       jid = qjId job
@@ -400,6 +444,7 @@ checkForDeath state jobWS = do
   died <- maybe (return False) isDead
           . mfilter (/= jqLivelock state)
           $ livelock
+  logDebug $ "Death of " ++ sjid ++ ": " ++ show died
   when died $ do
     logInfo $ "Detected death of job " ++ sjid
     -- if we manage to remove the job from the queue, we own the job file
@@ -417,6 +462,16 @@ checkForDeath state jobWS = do
             failedJob = failQueuedJob reason now $ jJob jobWS'
         cfg <- mkResultT . readIORef $ jqConfig state
         writeAndReplicateJob cfg qDir failedJob
+  return died
+
+-- | Trigger job detection for the job with the given job id.
+-- Return True, if the job is dead.
+cleanupIfDead :: JQStatus -> JobId -> IO Bool
+cleanupIfDead state jid = do
+  logDebug $ "Extra job-death detection for " ++ show (fromJobId jid)
+  jobs <- readIORef (jqJobs state)
+  let jobWS = find ((==) jid . qjId . jJob) $ qRunning jobs
+  maybe (return True) (checkForDeath state) jobWS
 
 -- | Time-based watcher for updating the job queue.
 onTimeWatcher :: JQStatus -> IO ()
@@ -529,3 +584,12 @@ setJobPriority state jid prio = runResultT $ do
       mkResultT $ writeJobToDisk qDir job'
       liftIO $ enqueueNewJobs state [job']
       return $ Just job'
+
+
+-- | Given old and new configs, determines if the changes between them should
+-- trigger the scheduler to run.
+configChangeNeedsRescheduling :: ConfigData -> ConfigData -> Bool
+configChangeNeedsRescheduling oldConfig newConfig =
+  or -- Trigger rescheduling if:
+    [ configFilters oldConfig /= configFilters newConfig -- filters changed
+    ]
