@@ -1,4 +1,5 @@
-{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances, FlexibleInstances, TupleSections,
+             GeneralizedNewtypeDeriving, DeriveTraversable #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 {-| JSON utility functions. -}
@@ -57,6 +58,7 @@ module Ganeti.JSON
   , toArray
   , optionalJSField
   , optFieldsToObj
+  , containerFromList
   , lookupContainer
   , alterContainerL
   , readContainer
@@ -72,6 +74,12 @@ module Ganeti.JSON
   , Container
   , MaybeForJSON(..)
   , TimeAsDoubleJSON(..)
+  , Tuple5(..)
+  , nestedAccessByKey
+  , nestedAccessByKeyDotted
+  , branchOnField
+  , addField
+  , maybeParseMap
   )
   where
 
@@ -80,6 +88,7 @@ import Control.DeepSeq
 import Control.Monad.Error.Class
 import Control.Monad.Writer
 import qualified Data.Foldable as F
+import qualified Data.Text as T
 import qualified Data.Traversable as F
 import Data.Maybe (fromMaybe, catMaybes)
 import qualified Data.Map as Map
@@ -88,6 +97,7 @@ import System.Time (ClockTime(..))
 import Text.Printf (printf)
 
 import qualified Text.JSON as J
+import qualified Text.JSON.Types as JT
 import Text.JSON.Pretty (pp_value)
 
 -- Note: this module should not import any Ganeti-specific modules
@@ -96,6 +106,10 @@ import Text.JSON.Pretty (pp_value)
 
 import Ganeti.BasicTypes
 
+-- Remove after we require >= 1.8.58
+-- See: https://github.com/ndmitchell/hlint/issues/24
+{-# ANN module "HLint: ignore Unused LANGUAGE pragma" #-}
+
 -- * JSON-related functions
 
 instance NFData J.JSValue where
@@ -123,12 +137,12 @@ readJSONWithDesc :: (J.JSON a)
                                               --   error messages
                  -> J.JSValue                 -- ^ input value
                  -> J.Result a
-readJSONWithDesc typ incInput input =
+readJSONWithDesc name incInput input =
   case J.readJSON input of
     J.Ok r    -> J.Ok r
     J.Error e -> J.Error $ if incInput then msg ++ " from " ++ show input
                                        else msg
-      where msg = "Can't parse value for type " ++ typ ++ ": " ++ e
+      where msg = "Can't parse value for '" ++ name ++ "': " ++ e
 
 -- | Converts a JSON Result into a monadic value.
 fromJResult :: Monad m => String -> J.Result a -> m a
@@ -314,20 +328,11 @@ instance HasStringRepr String where
 -- | The container type, a wrapper over Data.Map
 newtype GenericContainer a b =
   GenericContainer { fromContainer :: Map.Map a b }
-  deriving (Show, Eq)
+  deriving (Show, Eq, Ord, Functor, F.Foldable, F.Traversable)
 
 instance (NFData a, NFData b) => NFData (GenericContainer a b) where
   rnf = rnf . Map.toList . fromContainer
 
-instance Functor (GenericContainer a) where
-  fmap f = GenericContainer . fmap f . fromContainer
-
-instance F.Foldable (GenericContainer a) where
-  foldMap f = F.foldMap f . fromContainer
-
-instance F.Traversable (GenericContainer a) where
-  traverse f = fmap GenericContainer . F.traverse f . fromContainer
-
 -- | The empty container.
 emptyContainer :: GenericContainer a b
 emptyContainer = GenericContainer Map.empty
@@ -335,6 +340,10 @@ emptyContainer = GenericContainer Map.empty
 -- | Type alias for string keys.
 type Container = GenericContainer String
 
+-- | Creates a GenericContainer from a list of key-value pairs.
+containerFromList :: Ord a => [(a,b)] -> GenericContainer a b
+containerFromList = GenericContainer . Map.fromList
+
 -- | Looks up a value in a container with a default value.
 -- If a key has no value, a given monadic default is returned.
 -- This allows simple error handling, as the default can be
@@ -438,6 +447,10 @@ class ArrayObject a where
 -- | A Maybe newtype that allows for serialization more appropriate to the
 -- semantics of Maybe and JSON in our calls. Does not produce needless
 -- and confusing dictionaries.
+--
+-- In particular, `J.JSNull` corresponds to `Nothing`.
+-- This also means that this `Maybe a` newtype should not be used with `a`
+-- values that themselves can serialize to `null`.
 newtype MaybeForJSON a = MaybeForJSON { unMaybeForJSON :: Maybe a }
   deriving (Show, Eq, Ord)
 instance (J.JSON a) => J.JSON (MaybeForJSON a) where
@@ -459,3 +472,73 @@ instance J.JSON TimeAsDoubleJSON where
       pico = 10^(12 :: Int)
   showJSON (TimeAsDoubleJSON (TOD ss ps)) = J.showJSON
       (fromIntegral ss + fromIntegral ps / 10^(12 :: Int) :: Double)
+
+-- Text.JSON from the JSON package only has instances for tuples up to size 4.
+-- We use these newtypes so that we don't get a breakage once the 'json'
+-- package adds instances for larger tuples (or have to resort to CPP).
+
+newtype Tuple5 a b c d e = Tuple5 { unTuple5 :: (a, b, c, d, e) }
+
+instance (J.JSON a, J.JSON b, J.JSON c, J.JSON d, J.JSON e)
+         => J.JSON (Tuple5 a b c d e) where
+  readJSON (J.JSArray [a,b,c,d,e]) =
+    Tuple5 <$> ((,,,,) <$> J.readJSON a
+                       <*> J.readJSON b
+                       <*> J.readJSON c
+                       <*> J.readJSON d
+                       <*> J.readJSON e)
+  readJSON _ = fail "Unable to read Tuple5"
+  showJSON (Tuple5 (a, b, c, d, e)) =
+    J.JSArray
+      [ J.showJSON a
+      , J.showJSON b
+      , J.showJSON c
+      , J.showJSON d
+      , J.showJSON e
+      ]
+
+
+-- | Look up a value in a JSON object. Accessing @["a", "b", "c"]@ on an
+-- object is equivalent as accessing @myobject.a.b.c@ on a JavaScript object.
+--
+-- An error is returned if the object doesn't have such an accessor or if
+-- any value during the nested access is not an object at all.
+nestedAccessByKey :: [String] -> J.JSValue -> J.Result J.JSValue
+nestedAccessByKey keys json = case keys of
+  []   -> return json
+  k:ks -> case json of
+            J.JSObject obj -> J.valFromObj k obj >>= nestedAccessByKey ks
+            _ -> J.Error $ "Cannot access non-object with key '" ++ k ++ "'"
+
+
+-- | Same as `nestedAccessByKey`, but accessing with a dotted string instead
+-- (like @nestedAccessByKeyDotted "a.b.c"@).
+nestedAccessByKeyDotted :: String -> J.JSValue -> J.Result J.JSValue
+nestedAccessByKeyDotted s =
+  nestedAccessByKey (map T.unpack . T.splitOn (T.pack ".") . T.pack $ s)
+
+
+-- | Branch decoding on a field in a JSON object.
+branchOnField :: String -- ^ fieldname to branch on
+              -> (J.JSValue -> J.Result a)
+                  -- ^ decoding function if field is present and @true@; field
+                  -- will already be removed in the input
+              -> (J.JSValue -> J.Result a)
+                 -- ^ decoding function otherwise
+              -> J.JSValue -> J.Result a
+branchOnField k ifTrue ifFalse (J.JSObject jobj) =
+  let fields = J.fromJSObject jobj
+      jobj' = J.JSObject . J.toJSObject $ filter ((/=) k . fst) fields
+  in if lookup k fields == Just (J.JSBool True)
+       then ifTrue jobj'
+       else ifFalse jobj'
+branchOnField k _ _ _ = J.Error $ "Need an object to branch on key " ++ k
+
+-- | Add a field to a JSON object; to nothing, if the argument is not an object.
+addField :: (String, J.JSValue) -> J.JSValue -> J.JSValue
+addField (n,v) (J.JSObject obj) = J.JSObject $ JT.set_field obj n v
+addField _ jsval = jsval
+
+-- | Maybe obtain a map from a JSON object.
+maybeParseMap :: J.JSON a => J.JSValue -> Maybe (Map.Map String a)
+maybeParseMap = liftM fromContainer . readContainer <=< asJSObject
