@@ -1,4 +1,4 @@
-{-# LANGUAGE ExistentialQuantification, TemplateHaskell #-}
+{-# LANGUAGE ExistentialQuantification, TemplateHaskell, StandaloneDeriving #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 {-| Implementation of the opcodes.
@@ -159,6 +159,7 @@ $(genOpCode "OpCode"
      , pIgnoreErrors
      , pVerbose
      , pOptGroupName
+     , pVerifyClutter
      ],
      [])
   , ("OpClusterVerifyConfig",
@@ -179,6 +180,7 @@ $(genOpCode "OpCode"
      , pSkipChecks
      , pIgnoreErrors
      , pVerbose
+     , pVerifyClutter
      ],
      "group_name")
   , ("OpClusterVerifyDisks",
@@ -255,6 +257,8 @@ $(genOpCode "OpCode"
      , pZeroingImage
      , pCompressionTools
      , pEnabledUserShutdown
+     , pEnabledDataCollectors
+     , pDataCollectorInterval
      ],
      [])
   , ("OpClusterRedistConf",
@@ -275,7 +279,9 @@ $(genOpCode "OpCode"
   , ("OpClusterRenewCrypto",
      [t| () |],
      OpDoc.opClusterRenewCrypto,
-     [ pVerbose
+     [ pNodeSslCerts
+     , pSshKeys
+     , pVerbose
      , pDebug
      ],
      [])
@@ -339,6 +345,7 @@ $(genOpCode "OpCode"
      , pMasterCapable
      , pVmCapable
      , pNdParams
+     , pNodeSetup
      ],
      "node_name")
   , ("OpNodeQueryvols",
@@ -430,6 +437,7 @@ $(genOpCode "OpCode"
      , pRemoteNodeUuid
      , pIallocator
      , pEvacMode
+     , pIgnoreSoftErrors
      ],
      "node_name")
   , ("OpInstanceCreate",
@@ -444,6 +452,7 @@ $(genOpCode "OpCode"
      , pInstBeParams
      , pInstDisks
      , pOptDiskTemplate
+     , pOptGroupName
      , pFileDriver
      , pFileStorageDir
      , pInstHvParams
@@ -472,6 +481,8 @@ $(genOpCode "OpCode"
      , pSrcPath
      , pBackupCompress
      , pStartInstance
+     , pForthcoming
+     , pCommit
      , pInstTags
      , pInstanceCommunication
      , pHelperStartupTimeout
@@ -596,6 +607,7 @@ $(genOpCode "OpCode"
      , pMigrationCleanup
      , pIallocator
      , pAllowFailover
+     , pIgnoreHVVersions
      ],
      "instance_name")
   , ("OpInstanceMove",
@@ -619,7 +631,7 @@ $(genOpCode "OpCode"
      ],
      "instance_name")
   , ("OpInstanceActivateDisks",
-     [t| [(NonEmptyString, NonEmptyString, NonEmptyString)] |],
+     [t| [(NonEmptyString, NonEmptyString, Maybe NonEmptyString)] |],
      OpDoc.opInstanceActivateDisks,
      [ pInstanceName
      , pInstanceUuid
@@ -668,6 +680,9 @@ $(genOpCode "OpCode"
      , pRuntimeMem
      , pInstHvParams
      , pOptDiskTemplate
+     , pExtParams
+     , pFileDriver
+     , pFileStorageDir
      , pPrimaryNode
      , pPrimaryNodeUuid
      , withDoc "Secondary node (used when changing disk template)" pRemoteNode
@@ -804,6 +819,7 @@ $(genOpCode "OpCode"
      , pZeroFreeSpace
      , pZeroingTimeoutFixed
      , pZeroingTimeoutPerMiB
+     , pLongSleep
      ],
      "instance_name")
   , ("OpBackupRemove",
@@ -869,12 +885,13 @@ $(genOpCode "OpCode"
      , pIAllocatorMemory
      , pIAllocatorVCpus
      , pIAllocatorOs
-     , pDiskTemplate
+     , pOptDiskTemplate
      , pIAllocatorInstances
      , pIAllocatorEvacMode
      , pTargetGroups
      , pIAllocatorSpindleUse
      , pIAllocatorCount
+     , pOptGroupName
      ],
      "iallocator")
   , ("OpTestJqueue",
@@ -948,6 +965,8 @@ $(genOpCode "OpCode"
      "network_name")
   ])
 
+deriving instance Ord OpCode
+
 -- | Returns the OP_ID for a given opcode value.
 $(genOpID ''OpCode "opID")
 
@@ -1035,6 +1054,8 @@ $(buildObject "CommonOpParams" "op"
   , pReason
   ])
 
+deriving instance Ord CommonOpParams
+
 -- | Default common parameter values.
 defOpParams :: CommonOpParams
 defOpParams =
@@ -1056,7 +1077,7 @@ resolveDependsCommon p _ = return p
 -- | The top-level opcode type.
 data MetaOpCode = MetaOpCode { metaParams :: CommonOpParams
                              , metaOpCode :: OpCode
-                             } deriving (Show, Eq)
+                             } deriving (Show, Eq, Ord)
 
 -- | Resolve relative dependencies to absolute ones, given the job Id.
 resolveDependencies :: (Monad m) => MetaOpCode -> JobId -> m MetaOpCode
