@@ -42,6 +42,7 @@ module Ganeti.Query.Instance
   ) where
 
 import Control.Applicative
+import Control.Monad (liftM, (>=>))
 import Data.Either
 import Data.List
 import Data.Maybe
@@ -108,32 +109,32 @@ instanceFields =
   -- Simple fields
   [ (FieldDefinition "admin_state" "InstanceState" QFTText
      "Desired state of instance",
-     FieldSimple (rsNormal . adminStateToRaw . instAdminState), QffNormal)
+     FieldSimple (rsMaybeNoData . liftM adminStateToRaw . instAdminState),
+     QffNormal)
   , (FieldDefinition "admin_state_source" "InstanceStateSource" QFTText
      "Who last changed the desired state of the instance",
-     FieldSimple (rsNormal . adminStateSourceToRaw . instAdminStateSource),
+     FieldSimple (rsMaybeNoData . liftM adminStateSourceToRaw
+                  . instAdminStateSource),
      QffNormal)
   , (FieldDefinition "admin_up" "Autostart" QFTBool
      "Desired state of instance",
-     FieldSimple (rsNormal . (== AdminUp) . instAdminState), QffNormal)
-  , (FieldDefinition "disk_template" "Disk_template" QFTText
-     "Instance disk template",
-     FieldSimple (rsNormal . instDiskTemplate), QffNormal)
+     FieldSimple (rsMaybeNoData . liftM (== AdminUp) . instAdminState),
+     QffNormal)
   , (FieldDefinition "disks_active" "DisksActive" QFTBool
      "Desired state of instance disks",
-     FieldSimple (rsNormal . instDisksActive), QffNormal)
+     FieldSimple (rsMaybeNoData . instDisksActive), QffNormal)
   , (FieldDefinition "name" "Instance" QFTText
      "Instance name",
-     FieldSimple (rsNormal . instName), QffHostname)
+     FieldSimple (rsMaybeNoData . instName), QffHostname)
   , (FieldDefinition "hypervisor" "Hypervisor" QFTText
      "Hypervisor name",
-     FieldSimple (rsNormal . instHypervisor), QffNormal)
+     FieldSimple (rsMaybeNoData . instHypervisor), QffNormal)
   , (FieldDefinition "network_port" "Network_port" QFTOther
      "Instance network port if available (e.g. for VNC console)",
      FieldSimple (rsMaybeUnavail . instNetworkPort), QffNormal)
   , (FieldDefinition "os" "OS" QFTText
      "Operating system",
-     FieldSimple (rsNormal . instOs), QffNormal)
+     FieldSimple (rsMaybeNoData . instOs), QffNormal)
   , (FieldDefinition "pnode" "Primary_node" QFTText
      "Primary node",
      FieldConfig getPrimaryNodeName, QffHostname)
@@ -186,7 +187,7 @@ instanceFields =
   map (buildHvParamField hvParamGetter)
       (C.toList C.hvsParameters \\ C.toList C.hvcGlobals) ++
 
-  -- Aggregate disk parameter fields
+  -- disk parameter fields
   [ (FieldDefinition "disk_usage" "DiskUsage" QFTUnit
      "Total disk space used by instance on each of its nodes; this is not the\
      \ disk size visible to the instance, but the usage on the node",
@@ -195,8 +196,7 @@ instanceFields =
      "Number of disks",
      FieldSimple (rsNormal . length . instDisks), QffNormal)
   , (FieldDefinition "disk.sizes" "Disk_sizes" QFTOther
-     "List of disk sizes",
-     FieldConfig getDiskSizes, QffNormal)
+     "List of disk sizes", FieldConfig getDiskSizes, QffNormal)
   , (FieldDefinition "disk.spindles" "Disk_spindles" QFTOther
      "List of disk spindles",
      FieldConfig getDiskSpindles, QffNormal)
@@ -206,16 +206,22 @@ instanceFields =
   , (FieldDefinition "disk.uuids" "Disk_UUIDs" QFTOther
      "List of disk UUIDs",
      FieldConfig getDiskUuids, QffNormal)
+    -- For pre-2.14 backwards compatibility
+  , (FieldDefinition "disk_template" "Disk_template" QFTText
+     "Instance disk template",
+     FieldConfig getDiskTemplate, QffNormal)
   ] ++
 
   -- Per-disk parameter fields
   instantiateIndexedFields C.maxDisks
   [ (fieldDefinitionCompleter "disk.size/%d" "Disk/%d" QFTUnit
     "Disk size of %s disk",
-    getIndexedConfField getInstDisksFromObj diskSize, QffNormal)
+    getIndexedOptionalConfField getInstDisksFromObj diskSize,
+    QffNormal)
   , (fieldDefinitionCompleter "disk.spindles/%d" "DiskSpindles/%d" QFTNumber
     "Spindles of %s disk",
-    getIndexedOptionalConfField getInstDisksFromObj diskSpindles, QffNormal)
+    getIndexedOptionalConfField getInstDisksFromObj diskSpindles,
+    QffNormal)
   , (fieldDefinitionCompleter "disk.name/%d" "DiskName/%d" QFTText
     "Name of %s disk",
     getIndexedOptionalConfField getInstDisksFromObj diskName, QffNormal)
@@ -262,8 +268,7 @@ instanceFields =
      (nicAggDescPrefix ++ "link"),
      FieldConfig (\cfg -> rsNormal . map
        (nicpLink . fillNicParamsFromConfig cfg . nicNicparams)
-       . instNics),
-     QffNormal)
+       . instNics), QffNormal)
   , (FieldDefinition "nic.networks" "NIC_networks" QFTOther
      "List containing each interface's network",
      FieldSimple (rsNormal . map (MaybeForJSON . nicNetwork) . instNics),
@@ -330,6 +335,7 @@ instanceFields =
   timeStampFields ++
   serialFields "Instance" ++
   uuidFields "Instance" ++
+  forthcomingFields "Instance" ++
   tagsFields
 
 -- * Helper functions for node property retrieval
@@ -360,7 +366,7 @@ getNicVlan params
 
 -- | Fill partial NIC params by using the defaults from the configuration.
 fillNicParamsFromConfig :: ConfigData -> PartialNicParams -> FilledNicParams
-fillNicParamsFromConfig cfg = fillNicParams (getDefaultNicParams cfg)
+fillNicParamsFromConfig cfg = fillParams (getDefaultNicParams cfg)
 
 -- | Retrieves the default network interface parameters.
 getDefaultNicParams :: ConfigData -> FilledNicParams
@@ -372,20 +378,23 @@ getDefaultNicParams cfg =
 -- visible to the instance.
 getDiskSizeRequirements :: ConfigData -> Instance -> ResultEntry
 getDiskSizeRequirements cfg inst =
-  rsErrorNoData . liftA (sum . map getSizes) . getInstDisksFromObj cfg $ inst
+  rsErrorNoData . liftA (sum . map getSize) . getInstDisksFromObj cfg $ inst
  where
-  getSizes :: Disk -> Int
-  getSizes disk =
-    case instDiskTemplate inst of
-      DTDrbd8 -> diskSize disk + C.drbdMetaSize
-      DTDiskless -> 0
-      DTBlock    -> 0
-      _          -> diskSize disk
+  diskType x = lidDiskType <$> diskLogicalId x
+  getSize :: Disk -> Int
+  getSize disk =
+    let dt = diskType disk
+    in case dt of
+         Just DTDrbd8    -> fromMaybe 0 (diskSize disk) + C.drbdMetaSize
+         Just DTDiskless -> 0
+         Just DTBlock    -> 0
+         _               -> fromMaybe 0 (diskSize disk)
 
 -- | Get a list of disk sizes for an instance
 getDiskSizes :: ConfigData -> Instance -> ResultEntry
 getDiskSizes cfg =
-  rsErrorNoData . liftA (map diskSize) . getInstDisksFromObj cfg
+  rsErrorNoData . liftA (map $ MaybeForJSON . diskSize)
+  . getInstDisksFromObj cfg
 
 -- | Get a list of disk spindles
 getDiskSpindles :: ConfigData -> Instance -> ResultEntry
@@ -454,7 +463,7 @@ getOptionalIndexedNicField :: (J.JSON a)
                            -> FieldGetter Instance Runtime
 getOptionalIndexedNicField =
   getIndexedFieldWithDefault
-    (map nicNicparams . instNics) (\x _ -> getDefaultNicParams x) fillNicParams
+    (map nicNicparams . instNics) (\x _ -> getDefaultNicParams x) fillParams
 
 -- | Creates a function which produces a 'FieldGetter' when fed an index. Works
 -- for fields that should be filled out through the use of a default.
@@ -550,7 +559,9 @@ instantiateIndexedFields listSize fields = do
 
 -- | Helper function for primary node retrieval
 getPrimaryNode :: ConfigData -> Instance -> ErrorResult Node
-getPrimaryNode cfg = getInstPrimaryNode cfg . instName
+getPrimaryNode cfg = maybe (Bad $ ParameterError "no primary node") return
+                       . instName
+                     >=> getInstPrimaryNode cfg
 
 -- | Get primary node hostname
 getPrimaryNodeName :: ConfigData -> Instance -> ResultEntry
@@ -577,7 +588,8 @@ getPrimaryNodeGroupUuid cfg inst =
 getSecondaryNodes :: ConfigData -> Instance -> ErrorResult [Node]
 getSecondaryNodes cfg inst = do
   pNode <- getPrimaryNode cfg inst
-  allNodes <- getInstAllNodes cfg $ instName inst
+  iname <- maybe (Bad $ ParameterError "no name") return $ instName inst
+  allNodes <- getInstAllNodes cfg iname
   return $ delete pNode allNodes
 
 -- | Get attributes of the secondary nodes
@@ -688,11 +700,17 @@ statusDocText =
 -- | Checks if the primary node of an instance is offline
 isPrimaryOffline :: ConfigData -> Instance -> Bool
 isPrimaryOffline cfg inst =
-  let pNodeResult = getNode cfg $ instPrimaryNode inst
+  let pNodeResult = maybe (Bad $ ParameterError "no primary node") return
+                          (instPrimaryNode inst)
+                    >>= getNode cfg
   in case pNodeResult of
      Ok pNode -> nodeOffline pNode
      Bad    _ -> error "Programmer error - result assumed to be OK is Bad!"
 
+-- | Determines if user shutdown reporting is enabled
+userShutdownEnabled :: ConfigData -> Bool
+userShutdownEnabled = clusterEnabledUserShutdown . configCluster
+
 -- | Determines the status of a live instance
 liveInstanceStatus :: ConfigData
                    -> (InstanceInfo, Bool)
@@ -703,11 +721,11 @@ liveInstanceStatus cfg (instInfo, foundOnPrimary) inst
   | otherwise =
     case instanceState of
       InstanceStateRunning
-        | adminState == AdminUp -> Running
+        | adminState == Just AdminUp -> Running
         | otherwise -> ErrorUp
       InstanceStateShutdown
-        | adminState == AdminUp && allowDown -> UserDown
-        | adminState == AdminUp -> ErrorDown
+        | adminState == Just AdminUp && allowDown -> UserDown
+        | adminState == Just AdminUp -> ErrorDown
         | otherwise -> StatusDown
   where adminState = instAdminState inst
         instanceState = instInfoState instInfo
@@ -716,18 +734,21 @@ liveInstanceStatus cfg (instInfo, foundOnPrimary) inst
           fromContainer $ getFilledInstHvParams (C.toList C.hvcGlobals) cfg inst
 
         allowDown =
-          instHypervisor inst /= Kvm ||
-          (Map.member C.hvKvmUserShutdown hvparams &&
-           hvparams Map.! C.hvKvmUserShutdown == J.JSBool True)
+          userShutdownEnabled cfg &&
+          (instHypervisor inst /= Just Kvm ||
+           (Map.member C.hvKvmUserShutdown hvparams &&
+            hvparams Map.! C.hvKvmUserShutdown == J.JSBool True))
 
 -- | Determines the status of a dead instance.
-deadInstanceStatus :: Instance -> InstanceStatus
-deadInstanceStatus inst =
+deadInstanceStatus :: ConfigData -> Instance -> InstanceStatus
+deadInstanceStatus cfg inst =
   case instAdminState inst of
-    AdminUp -> ErrorDown
-    AdminDown | instAdminStateSource inst == UserSource -> UserDown
-              | otherwise -> StatusDown
-    AdminOffline -> StatusOffline
+    Just AdminUp -> ErrorDown
+    Just AdminDown | wasCleanedUp && userShutdownEnabled cfg -> UserDown
+                   | otherwise -> StatusDown
+    Just AdminOffline -> StatusOffline
+    Nothing -> StatusDown
+  where wasCleanedUp = instAdminStateSource inst == Just UserSource
 
 -- | Determines the status of the instance, depending on whether it is possible
 -- to communicate with its primary node, on which node it is, and its
@@ -741,7 +762,7 @@ determineInstanceStatus cfg res inst
   | otherwise = case res of
       Left _                   -> NodeDown
       Right (Just liveData, _) -> liveInstanceStatus cfg liveData inst
-      Right (Nothing, _)       -> deadInstanceStatus inst
+      Right (Nothing, _)       -> deadInstanceStatus cfg inst
 
 -- | Extracts the instance status, retrieving it using the functions above and
 -- transforming it into a 'ResultEntry'.
@@ -787,7 +808,7 @@ findInfoInNodeResult inst nodeResponse =
     Left  _err    -> Nothing
     Right allInfo ->
       let instances = rpcResAllInstInfoInstances allInfo
-          maybeMatch = pickPairUnique (instName inst) instances
+          maybeMatch = instName inst >>= (`pickPairUnique` instances)
       in snd <$> maybeMatch
 
 -- | Retrieves the instance information if it is present anywhere in the all
@@ -797,21 +818,24 @@ getInstanceInfo :: [(String, ERpcError RpcResultAllInstancesInfo)]
                 -> Instance
                 -> ERpcError (Maybe (InstanceInfo, Bool))
 getInstanceInfo uuidList inst =
-  let pNodeUuid = instPrimaryNode inst
-      primarySearchResult =
-        pickPairUnique pNodeUuid uuidList >>= findInfoInNodeResult inst . snd
-  in case primarySearchResult of
-       Just instInfo -> Right . Just $ (instInfo, True)
-       Nothing       ->
-         let allSearchResult =
-               getFirst . mconcat $ map
-               (First . findInfoInNodeResult inst . snd) uuidList
-         in case allSearchResult of
-              Just instInfo -> Right . Just $ (instInfo, False)
-              Nothing       ->
-                case checkForNodeError uuidList pNodeUuid of
-                  Just err -> Left err
-                  Nothing  -> Right Nothing
+  case instPrimaryNode inst of
+    Nothing -> Right Nothing
+    Just pNodeUuid ->
+      let primarySearchResult =
+            pickPairUnique pNodeUuid uuidList >>= findInfoInNodeResult inst
+                                                    . snd
+      in case primarySearchResult of
+           Just instInfo -> Right . Just $ (instInfo, True)
+           Nothing       ->
+             let allSearchResult =
+                   getFirst . mconcat $ map
+                   (First . findInfoInNodeResult inst . snd) uuidList
+             in case allSearchResult of
+                  Just instInfo -> Right . Just $ (instInfo, False)
+                  Nothing       ->
+                    case checkForNodeError uuidList pNodeUuid of
+                      Just err -> Left err
+                      Nothing  -> Right Nothing
 
 -- | Retrieves the console information if present anywhere in the given results
 getConsoleInfo :: [(String, ERpcError RpcResultInstanceConsoleInfo)]
@@ -820,7 +844,7 @@ getConsoleInfo :: [(String, ERpcError RpcResultInstanceConsoleInfo)]
 getConsoleInfo uuidList inst =
   let allValidResults = concatMap rpcResInstConsInfoInstancesInfo .
                         rights . map snd $ uuidList
-  in snd <$> pickPairUnique (instName inst) allValidResults
+  in snd <$> (instName inst >>= flip pickPairUnique allValidResults)
 
 -- | Extracts all the live information that can be extracted.
 extractLiveInfo :: [(Node, ERpcError RpcResultAllInstancesInfo)]
@@ -865,15 +889,18 @@ consoleParamsToCalls params =
             [] -> error "Programmer error: group must have one or more members"
             paramGroup@(y:_) ->
               let node = instConsInfoParamsNode y
-                  packer z = (instName $ instConsInfoParamsInstance z, z)
-              in (node, RpcCallInstanceConsoleInfo . map packer $ paramGroup)
+                  packer z = do
+                              name <- instName $ instConsInfoParamsInstance z
+                              return (name, z)
+              in (node, RpcCallInstanceConsoleInfo . mapMaybe packer
+                          $ paramGroup)
          ) groupedParams
 
 -- | Retrieves a list of all the hypervisors and params used by the given
 -- instances.
 getHypervisorSpecs :: ConfigData -> [Instance] -> [(Hypervisor, HvParams)]
 getHypervisorSpecs cfg instances =
-  let hvs = nub . map instHypervisor $ instances
+  let hvs = nub . mapMaybe instHypervisor $ instances
       hvParamMap = (fromContainer . clusterHvparams . configCluster $ cfg)
   in zip hvs . map ((Map.!) hvParamMap) $ hvs
 
@@ -888,8 +915,11 @@ collectLiveData liveDataEnabled cfg fields instances
                             RpcResultError $ "Live data disabled"
   | otherwise = do
       let hvSpecs = getHypervisorSpecs cfg instances
-          instanceNodes = nub . justOk $
-                            map (getNode cfg . instPrimaryNode) instances
+          instanceNodes =
+            nub . justOk
+                $ map ( maybe (Bad $ ParameterError "no primary node") return
+                       . instPrimaryNode
+                       >=> getNode cfg) instances
           goodNodes = nodesWithValidConfig cfg instanceNodes
       instInfoRes <- executeRpcCall goodNodes (RpcCallAllInstancesInfo hvSpecs)
       consInfoRes <-
@@ -901,3 +931,18 @@ collectLiveData liveDataEnabled cfg fields instances
           else return [] -- The information is not necessary
       return . zip instances .
         map (extractLiveInfo instInfoRes consInfoRes) $ instances
+
+-- | An aggregate disk attribute for backward compatibility.
+getDiskTemplate :: ConfigData -> Instance -> ResultEntry
+getDiskTemplate cfg inst =
+  let disks = getInstDisksFromObj cfg inst
+      getDt x = lidDiskType <$> diskLogicalId x
+      disk_types :: ErrorResult [DiskTemplate]
+      disk_types = nub <$> catMaybes <$> map getDt <$> disks
+      mix :: [DiskTemplate] -> J.JSValue
+      mix []  = J.showJSON C.dtDiskless
+      mix [t] = J.showJSON t
+      mix _   = J.showJSON C.dtMixed
+  in case mix <$> disk_types of
+       Ok t -> rsNormal t
+       Bad _ -> rsNoData
