@@ -88,6 +88,7 @@ import Ganeti.Logging
 import Ganeti.Objects
 import Ganeti.Query.Common
 import qualified Ganeti.Query.Export as Export
+import qualified Ganeti.Query.FilterRules as FilterRules
 import Ganeti.Query.Filter
 import qualified Ganeti.Query.Instance as Instance
 import qualified Ganeti.Query.Job as Query.Job
@@ -218,7 +219,10 @@ genericQuery fieldsMap collector nameFn configFn getFn cfg
              live fields qfilter wanted =
   runResultT $ do
   cfilter <- toError $ compileFilter fieldsMap qfilter
-  let selected = getSelectedFields fieldsMap fields
+  let allfields = (++) fields . filter (not . (`elem` fields))
+                  . ordNub $ filterArguments qfilter
+      count = length fields
+      selected = getSelectedFields fieldsMap allfields
       (fdefs, fgetters, _) = unzip3 selected
       live' = live && needsLiveData fgetters
   objects <- toError $ case wanted of
@@ -228,18 +232,19 @@ genericQuery fieldsMap collector nameFn configFn getFn cfg
   -- Run the first pass of the filter, without a runtime context; this will
   -- limit the objects that we'll contact for exports
   fobjects <- toError $
-    filterM (\n -> evaluateFilter cfg Nothing n cfilter) objects
+    filterM (\n -> evaluateQueryFilter cfg Nothing n cfilter) objects
   -- Gather the runtime data and filter the results again,
   -- based on the gathered data
   runtimes <- (case collector of
     CollectorSimple     collFn -> lift $ collFn live' cfg fobjects
-    CollectorFieldAware collFn -> lift $ collFn live' cfg fields fobjects) >>=
-    (toError . filterM (\(obj, runtime) ->
-      evaluateFilter cfg (Just runtime) obj cfilter))
+    CollectorFieldAware collFn -> lift $ collFn live' cfg allfields fobjects)
+    >>= (toError . filterM (\(obj, runtime) ->
+      evaluateQueryFilter cfg (Just runtime) obj cfilter))
   let fdata = map (\(obj, runtime) ->
                      map (execGetter cfg runtime obj) fgetters)
               runtimes
-  return QueryResult { qresFields = fdefs, qresData = fdata }
+  return QueryResult { qresFields = take count fdefs
+                     , qresData = map (take count) fdata }
 
 -- | Dummy recollection of the data for a lock from the prefected
 -- data for all locks.
@@ -312,7 +317,8 @@ queryInner cfg live (Query (ItemTypeOpCode QRNode) fields qfilter) wanted =
 
 queryInner cfg live (Query (ItemTypeOpCode QRInstance) fields qfilter) wanted =
   genericQuery Instance.fieldsMap (CollectorFieldAware Instance.collectLiveData)
-               instName configInstances getInstance cfg live fields qfilter
+               (fromMaybe "" . instName) configInstances getInstance cfg live
+               fields qfilter
                wanted
 
 queryInner cfg live (Query (ItemTypeOpCode QRGroup) fields qfilter) wanted =
@@ -328,6 +334,10 @@ queryInner cfg live (Query (ItemTypeOpCode QRExport) fields qfilter) wanted =
   genericQuery Export.fieldsMap (CollectorSimple Export.collectLiveData)
                nodeName configNodes getNode cfg live fields qfilter wanted
 
+queryInner cfg live (Query (ItemTypeLuxi QRFilter) fields qfilter) wanted =
+  genericQuery FilterRules.fieldsMap (CollectorSimple dummyCollectLiveData)
+               frUuid configFilters getFilterRule cfg live fields qfilter wanted
+
 queryInner _ _ (Query qkind _ _) _ =
   return . Bad . GenericError $ "Query '" ++ show qkind ++ "' not supported"
 
@@ -362,7 +372,7 @@ queryJobs cfg live fields qfilter = runResultT $ do
   -- runs first pass of the filter, without a runtime context; this
   -- will limit the jobs that we'll load from disk
   jids <- toError $
-          filterM (\jid -> evaluateFilter cfg Nothing jid cfilter) rjids
+    filterM (\jid -> evaluateQueryFilter cfg Nothing jid cfilter) rjids
   -- here we run the runtime data gathering, filtering and evaluation,
   -- all in the same step, so that we don't keep jobs in memory longer
   -- than we need; we can't be fully lazy due to the multiple monad
@@ -374,7 +384,7 @@ queryJobs cfg live fields qfilter = runResultT $ do
               job <- lift $ if live'
                               then loadJobFromDisk qdir True jid
                               else return disabled_data
-              pass <- toError $ evaluateFilter cfg (Just job) jid cfilter
+              pass <- toError $ evaluateQueryFilter cfg (Just job) jid cfilter
               let nlst = if pass
                            then let row = map (execGetter cfg job jid) fgetters
                                 in rnf row `seq` row:lst
@@ -415,6 +425,9 @@ queryFields (QueryFields (ItemTypeOpCode QRInstance) fields) =
 queryFields (QueryFields (ItemTypeLuxi QRLock) fields) =
   Ok $ fieldsExtractor Locks.fieldsMap fields
 
+queryFields (QueryFields (ItemTypeLuxi QRFilter) fields) =
+  Ok $ fieldsExtractor FilterRules.fieldsMap fields
+
 queryFields (QueryFields qkind _) =
   Bad . GenericError $ "QueryFields '" ++ show qkind ++ "' not supported"
 
