@@ -1,3 +1,5 @@
+{-# LANGUAGE FlexibleContexts #-}
+
 {-| Implementation of the Ganeti Query2 server.
 
 -}
@@ -41,18 +43,22 @@ module Ganeti.Query.Server
 import Control.Applicative
 import Control.Concurrent
 import Control.Exception
+import Control.Lens ((.~))
 import Control.Monad (forever, when, mzero, guard, zipWithM, liftM, void)
+import Control.Monad.Base (MonadBase, liftBase)
+import Control.Monad.Error (MonadError)
 import Control.Monad.IO.Class
 import Control.Monad.Trans (lift)
 import Control.Monad.Trans.Maybe
-import Data.Bits (bitSize)
 import qualified Data.Set as Set (toList)
 import Data.IORef
+import Data.List (intersperse)
 import Data.Maybe (fromMaybe)
 import qualified Text.JSON as J
 import Text.JSON (encode, showJSON, JSValue(..))
 import System.Info (arch)
 import System.Directory
+import System.Posix.Process (getProcessID)
 import System.Posix.Signals as P
 
 import qualified Ganeti.Constants as C
@@ -62,12 +68,15 @@ import qualified Ganeti.Path as Path
 import Ganeti.Daemon
 import Ganeti.Daemon.Utils (handleMasterVerificationOptions)
 import Ganeti.Objects
+import Ganeti.Objects.Lens (configFiltersL)
 import qualified Ganeti.Config as Config
+import qualified Ganeti.Compat as Compat
 import Ganeti.ConfigReader
 import Ganeti.BasicTypes
 import Ganeti.JQueue
 import Ganeti.JQScheduler
-import Ganeti.JSON (TimeAsDoubleJSON(..))
+import Ganeti.JSON (TimeAsDoubleJSON(..), alterContainerL, lookupContainer)
+import Ganeti.Locking.Locks (ClientId(..), ClientType(ClientOther))
 import Ganeti.Logging
 import Ganeti.Luxi
 import qualified Ganeti.Query.Language as Qlang
@@ -77,14 +86,73 @@ import Ganeti.Rpc
 import qualified Ganeti.Query.Exec as Exec
 import Ganeti.Query.Query
 import Ganeti.Query.Filter (makeSimpleFilter)
+import Ganeti.THH.HsRPC (runRpcClient, RpcClientMonad)
 import Ganeti.Types
 import qualified Ganeti.UDSServer as U (Handler(..), listener)
 import Ganeti.Utils ( lockFile, exitIfBad, exitUnless, watchFile
-                    , safeRenameFile )
+                    , safeRenameFile, newUUID, isUUID )
+import Ganeti.Utils.Monad (orM)
 import Ganeti.Utils.MVarLock
 import qualified Ganeti.Version as Version
+import Ganeti.WConfd.Client ( getWConfdClient, withLockedConfig, writeConfig
+                            , cleanupLocks)
+
+
+-- | Creates a `ClientId` that identifies the current luxi
+-- (process, thread).
+--
+-- This means that this `ClientId` will be different for each request
+-- handled by luxid.
+makeLuxidClientId :: JQStatus -> IO ClientId
+makeLuxidClientId status = do
+  pid <- getProcessID
+  tid <- myThreadId
+  return ClientId
+           { ciIdentifier = ClientOther $ "luxid-" ++ show tid
+           , ciLockFile = jqLivelock status
+           , ciPid = pid
+           }
+
+-- | Creates a connection to WConfd and locks the config, allowing
+-- to run some WConfd RPC commands given the locked config.
+--
+-- This is needed when luxid wants to change the config.
+--
+-- Example:
+--
+-- > cid <- makeLuxidClientId ...
+-- > withLockedWconfdConfig cid $ \lockedCfg -> do
+-- >   -- some (IO) action that needs to be run inside having the lock
+-- >   writeConfig cid (updateConfig lockedCfg)
+withLockedWconfdConfig
+  :: (MonadBase IO m, MonadError GanetiException m)
+  => ClientId
+  -> (ConfigData -> RpcClientMonad a)
+  -> m a
+withLockedWconfdConfig cid f = do
+  wconfdClient <- liftBase $ getWConfdClient =<< Path.defaultWConfdSocket
+  runRpcClient (withLockedConfig cid False f) wconfdClient
 
 -- | Helper for classic queries.
+handleQuery :: [Qlang.ItemType -> Qlang.FilterField] -- ^ Fields to put into
+                                                     -- the query
+            -> ConfigData      -- ^ Cluster config
+            -> Qlang.ItemType  -- ^ Query type
+            -> [Either String Integer] -- ^ Requested names
+                                       -- (empty means all)
+            -> [String]        -- ^ Requested fields
+            -> Bool            -- ^ Whether to do sync queries or not
+            -> IO (GenericResult GanetiException JSValue)
+handleQuery _ _ _ _ _ True =
+  return . Bad $ OpPrereqError "Sync queries are not allowed" ECodeInval
+handleQuery filterFields cfg qkind names fields _ = do
+  let simpleNameFilter field = makeSimpleFilter (field qkind) names
+      flt = Qlang.OrFilter $ map simpleNameFilter filterFields
+  qr <- query cfg True (Qlang.Query qkind fields flt)
+  return $ showJSON <$> (qr >>= queryCompat)
+
+-- | Helper for classic queries.
+-- Queries `name` and `uuid` fields.
 handleClassicQuery :: ConfigData      -- ^ Cluster config
                    -> Qlang.ItemType  -- ^ Query type
                    -> [Either String Integer] -- ^ Requested names
@@ -92,13 +160,17 @@ handleClassicQuery :: ConfigData      -- ^ Cluster config
                    -> [String]        -- ^ Requested fields
                    -> Bool            -- ^ Whether to do sync queries or not
                    -> IO (GenericResult GanetiException JSValue)
-handleClassicQuery _ _ _ _ True =
-  return . Bad $ OpPrereqError "Sync queries are not allowed" ECodeInval
-handleClassicQuery cfg qkind names fields _ = do
-  let simpleNameFilter field = makeSimpleFilter (field qkind) names
-      flt = Qlang.OrFilter $ map simpleNameFilter [nameField, uuidField]
-  qr <- query cfg True (Qlang.Query qkind fields flt)
-  return $ showJSON <$> (qr >>= queryCompat)
+handleClassicQuery = handleQuery [nameField, uuidField]
+
+-- | Like `handleClassicQuery`, but filters only by UUID.
+handleUuidQuery :: ConfigData      -- ^ Cluster config
+                -> Qlang.ItemType  -- ^ Query type
+                -> [Either String Integer] -- ^ Requested names
+                                           -- (empty means all)
+                -> [String]        -- ^ Requested fields
+                -> Bool            -- ^ Whether to do sync queries or not
+                -> IO (GenericResult GanetiException JSValue)
+handleUuidQuery = handleQuery [uuidField]
 
 -- | Minimal wrapper to handle the missing config case.
 handleCallWrapper :: Lock -> JQStatus ->  Result ConfigData
@@ -120,7 +192,7 @@ handleCall _ _ cdata QueryClusterInfo =
       def_hv = case hypervisors of
                  x:_ -> showJSON x
                  [] -> JSNull
-      bits = show (bitSize (0::Int)) ++ "bits"
+      bits = show (Compat.finiteBitSize (0::Int)) ++ "bits"
       arch_tuple = [bits, arch]
       obj = [ ("software_version", showJSON C.releaseVersion)
             , ("protocol_version", showJSON C.protocolVersion)
@@ -192,6 +264,12 @@ handleCall _ _ cdata QueryClusterInfo =
                showJSON $ clusterCompressionTools cluster)
             , ("enabled_user_shutdown",
                showJSON $ clusterEnabledUserShutdown cluster)
+            , ("enabled_data_collectors",
+               showJSON . fmap dataCollectorActive
+                        $ clusterDataCollectors cluster)
+            , ("data_collector_interval",
+               showJSON . fmap dataCollectorInterval
+                        $ clusterDataCollectors cluster)
             ]
 
   in case master of
@@ -231,6 +309,64 @@ handleCall _ _ cfg (QueryJobs names fields) =
   handleClassicQuery cfg (Qlang.ItemTypeLuxi Qlang.QRJob)
     (map (Right . fromIntegral . fromJobId) names)  fields False
 
+handleCall _ _ cfg (QueryFilters uuids fields) =
+  handleUuidQuery cfg (Qlang.ItemTypeLuxi Qlang.QRFilter)
+    (map Left uuids) fields False
+
+handleCall _ status _ (ReplaceFilter mUuid priority predicates action
+                                     reason) =
+  -- Handles both adding new filter and changing existing ones.
+  runResultT $ do
+
+    -- Check that uuid `String` is actually a UUID.
+    uuid <- case mUuid of
+      Nothing -> liftIO newUUID  -- Request to add a new filter
+      Just u                     -- Request to edit an existing filter
+        | isUUID u  -> return u
+        | otherwise -> fail "Unable to parse UUID"
+
+    timestamp <- liftIO $ reasonTrailTimestamp <$> currentTimestamp
+    let luxidReason = ("luxid", "", timestamp)
+
+    -- Ask WConfd to change the config for us.
+    cid <- liftIO $ makeLuxidClientId status
+    withLockedWconfdConfig cid $ \lockedCfg -> do
+      -- Reading the latest JobID inside the Wconfd lock to really get the
+      -- most recent one (locking may block us for some time).
+      serial <- liftIO readSerialFromDisk
+      case serial of
+        Bad err -> fail $ "AddFilter: reading current JobId failed: " ++ err
+        Ok watermark -> do
+          let rule = FilterRule { frWatermark = watermark
+                                , frPriority = priority
+                                , frPredicates = predicates
+                                , frAction = action
+                                , frReasonTrail = reason ++ [luxidReason]
+                                , frUuid = uuid
+                                }
+          writeConfig cid
+            . (configFiltersL . alterContainerL uuid .~ Just rule)
+            $ lockedCfg
+
+    -- Return UUID of added/replaced filter.
+    return $ showJSON uuid
+
+handleCall _ status cfg (DeleteFilter uuid) = runResultT $ do
+  -- Check if filter exists.
+  _ <- lookupContainer
+    (failError $ "Filter rule with UUID " ++ uuid ++ " does not exist")
+    uuid
+    (configFilters cfg)
+
+  -- Ask WConfd to change the config for us.
+  cid <- liftIO $ makeLuxidClientId status
+  withLockedWconfdConfig cid $ \lockedCfg ->
+    writeConfig cid
+      . (configFiltersL . alterContainerL uuid .~ Nothing)
+      $ lockedCfg
+
+  return JSNull
+
 handleCall _ _ cfg (QueryNetworks names fields lock) =
   handleClassicQuery cfg (Qlang.ItemTypeOpCode Qlang.QRNetwork)
     (map Left names) fields lock
@@ -342,7 +478,7 @@ handleCall _ qstat cfg (ChangeJobPriority jid prio) = do
       logDebug $ jName ++ " started, will signal"
       fmap showJSON <$> tellJobPriority (jqLivelock qstat) jid prio
 
-handleCall _ qstat  cfg (CancelJob jid) = do
+handleCall _ qstat  cfg (CancelJob jid kill) = do
   let jName = (++) "job " . show $ fromJobId jid
   dequeueResult <- dequeueJob qstat jid
   case dequeueResult of
@@ -361,7 +497,16 @@ handleCall _ qstat  cfg (CancelJob jid) = do
             writeAndReplicateJob cfg qDir job'
     Ok False -> do
       logDebug $ jName ++ " not queued; trying to cancel directly"
-      fmap showJSON <$> cancelJob (jqLivelock qstat) jid
+      result <- fmap showJSON <$> cancelJob kill (jqLivelock qstat) jid
+      when kill . void . forkIO $ do
+        _ <- orM
+             . intersperse (threadDelay C.luxidJobDeathDelay >> return False)
+             . replicate C.luxidJobDeathDetectionRetries
+             $ cleanupIfDead qstat jid
+        wconfdsocket <- Path.defaultWConfdSocket
+        wconfdclient <- getWConfdClient wconfdsocket
+        void . runResultT $ runRpcClient cleanupLocks wconfdclient
+      return result
     Bad s -> return . Ok . showJSON $ (False, s)
 
 handleCall qlock _ cfg (ArchiveJob jid) =
@@ -480,7 +625,17 @@ prepMain _ _ = do
 -- | Main function.
 main :: MainFn () PrepResult
 main _ _ (server, cref, jq) = do
-  initConfigReader id cref
+  -- Subscribe to config udpates. If the config changes, write new config and
+  -- check if the changes should trigger the scheduler.
+  initConfigReader $ \newConfig -> do
+
+    runScheduler <- atomicModifyIORef cref $ \oldConfig ->
+      case (oldConfig, newConfig) of
+        (Ok old, Ok new) -> (newConfig, configChangeNeedsRescheduling old new)
+        _                -> (newConfig, True) -- no old or new config, schedule
+
+    when runScheduler (scheduleSomeJobs jq)
+
   let creader = readIORef cref
 
   qlockFile <- jobQueueLockFile
