@@ -61,19 +61,24 @@ module Ganeti.THH ( declareSADT
                   , andRestArguments
                   , withDoc
                   , defaultField
+                  , presentInForthcoming
                   , optionalField
                   , optionalNullSerField
+                  , makeOptional
                   , renameField
                   , customField
                   , buildObject
+                  , buildObjectWithForthcoming
                   , buildObjectSerialisation
                   , buildParam
                   , genException
                   , excErrMsg
                   ) where
 
-import Control.Arrow ((&&&))
+import Control.Arrow ((&&&), second)
 import Control.Applicative
+import Control.Lens.Type (Lens')
+import Control.Lens (lens, set, element)
 import Control.Monad
 import Control.Monad.Base () -- Needed to prevent spurious GHC linking errors.
 import Control.Monad.Writer (tell)
@@ -86,6 +91,7 @@ import Data.Function (on)
 import Data.List
 import Data.Maybe
 import qualified Data.Map as M
+import Data.Monoid
 import qualified Data.Set as S
 import Language.Haskell.TH
 import Language.Haskell.TH.Syntax (lift)
@@ -94,6 +100,7 @@ import qualified Text.JSON as JSON
 import Text.JSON.Pretty (pp_value)
 
 import Ganeti.JSON
+import Ganeti.PartialParams
 import Ganeti.PyValue
 import Ganeti.THH.PyType
 
@@ -181,6 +188,7 @@ data Field = Field { fieldName        :: String
                      -- ^ determines if a field is optional, and if yes,
                      -- how
                    , fieldDoc         :: String
+                   , fieldPresentInForthcoming :: Bool
                    }
 
 -- | Generates a simple field.
@@ -195,6 +203,7 @@ simpleField fname ftype =
         , fieldConstr      = Nothing
         , fieldIsOptional  = NotOptional
         , fieldDoc         = ""
+        , fieldPresentInForthcoming = False
         }
 
 -- | Generate an AndRestArguments catch-all field.
@@ -209,6 +218,7 @@ andRestArguments fname =
         , fieldConstr      = Nothing
         , fieldIsOptional  = AndRestArguments
         , fieldDoc         = ""
+        , fieldPresentInForthcoming = True
         }
 
 withDoc :: String -> Field -> Field
@@ -224,6 +234,10 @@ renameField constrName field = field { fieldConstr = Just constrName }
 defaultField :: Q Exp -> Field -> Field
 defaultField defval field = field { fieldDefault = Just defval }
 
+-- | Mark a field as present in the forthcoming variant.
+presentInForthcoming :: Field -> Field
+presentInForthcoming field = field { fieldPresentInForthcoming = True }
+
 -- | Marks a field optional (turning its base type into a Maybe).
 optionalField :: Field -> Field
 optionalField field = field { fieldIsOptional = OptionalOmitNull }
@@ -233,6 +247,15 @@ optionalField field = field { fieldIsOptional = OptionalOmitNull }
 optionalNullSerField :: Field -> Field
 optionalNullSerField field = field { fieldIsOptional = OptionalSerializeNull }
 
+-- | Make a field optional, if it isn't already.
+makeOptional :: Field -> Field
+makeOptional field = if  and [ fieldIsOptional field == NotOptional
+                             , isNothing $ fieldDefault field
+                             , not $ fieldPresentInForthcoming field
+                             ]
+                        then optionalField field
+                        else field
+
 -- | Sets custom functions on a field.
 customField :: Name      -- ^ The name of the read function
             -> Name      -- ^ The name of the show function
@@ -351,18 +374,6 @@ ensureUpper :: String -> String
 ensureUpper [] = []
 ensureUpper (x:xs) = toUpper x:xs
 
--- | Helper for quoted expressions.
-varNameE :: String -> Q Exp
-varNameE = varE . mkName
-
--- | showJSON as an expression, for reuse.
-showJSONE :: Q Exp
-showJSONE = varE 'JSON.showJSON
-
--- | makeObj as an expression, for reuse.
-makeObjE :: Q Exp
-makeObjE = varE 'JSON.makeObj
-
 -- | fromObj (Ganeti specific) as an expression, for reuse.
 fromObjE :: Q Exp
 fromObjE = varE 'fromObj
@@ -379,14 +390,6 @@ fromRawName = mkName . (++ "FromRaw") . ensureLower
 reprE :: Either String Name -> Q Exp
 reprE = either stringE varE
 
--- | Smarter function application.
---
--- This does simply f x, except that if is 'id', it will skip it, in
--- order to generate more readable code when using -ddump-splices.
-appFn :: Exp -> Exp -> Exp
-appFn f x | f == VarE 'id = x
-          | otherwise = AppE f x
-
 -- | Apply a constructor to a list of expressions
 appCons :: Name -> [Exp] -> Exp
 appCons cname = foldl AppE (ConE cname)
@@ -395,7 +398,7 @@ appCons cname = foldl AppE (ConE cname)
 appConsApp :: Name -> [Exp] -> Exp
 appConsApp cname =
   foldl (\accu e -> InfixE (Just accu) (VarE '(<*>)) (Just e))
-          (AppE (VarE 'pure) (ConE cname))
+        (AppE (VarE 'pure) (ConE cname))
 
 -- | Builds a field for a normal constructor.
 buildConsField :: Q Type -> StrictTypeQ
@@ -508,7 +511,7 @@ declareADT fn traw sname cons = do
   let name = mkName sname
       ddecl = strADTDecl name (map fst cons)
       -- process cons in the format expected by genToRaw
-      cons' = map (\(a, b) -> (a, fn b)) cons
+      cons' = map (second fn) cons
   toraw <- genToRaw traw (toRawName sname) name cons'
   fromraw <- genFromRaw traw (fromRawName sname) name cons'
   return $ ddecl:toraw ++ fromraw
@@ -660,9 +663,6 @@ genAllConstr trans_fun name vstr = do
 genAllOpIDs :: Name -> String -> Q [Dec]
 genAllOpIDs = genAllConstr deCamelCase
 
--- | OpCode parameter (field) type.
-type OpParam = (String, Q Type, Q Exp)
-
 -- * Python code generation
 
 data OpCodeField = OpCodeField { ocfName :: String
@@ -719,8 +719,7 @@ pyField f = genPyType f >>= \t ->
 -- necessary information on how to build the Python class string.
 pyClass :: OpCodeConstructor -> Q Exp
 pyClass (consName, consType, consDoc, consFields, consDscField) =
-  do let pyClassVar = varNameE "showPyClass"
-         consName' = stringE consName
+  do let consName' = stringE consName
      consType' <- genPyType' NotOptional consType
      let consDoc' = stringE consDoc
      [| OpCodeDescriptor $consName'
@@ -953,6 +952,158 @@ buildObject sname field_pfx fields = do
   ser_decls <- buildObjectSerialisation sname fields
   return $ declD:ser_decls
 
+-- | Build an accessor function for a field of an object
+-- that can have a forthcoming variant.
+buildAccessor :: Name -- ^ name of the forthcoming constructor
+              -> String -- ^ prefix for the forthcoming field
+              -> Name -- ^ name of the real constructor
+              -> String -- ^ prefix for the real field
+              -> Name -- ^ name of the generated accessor
+              -> String -- ^ prefix of the generated accessor
+              -> Field -- ^ field description
+              -> Q [Dec]
+buildAccessor fnm fpfx rnm rpfx nm pfx field = do
+  let optField = makeOptional field
+  x <- newName "x"
+  (rpfx_name, _, _) <- fieldTypeInfo rpfx field
+  (fpfx_name, _, ftype) <- fieldTypeInfo fpfx optField
+  (pfx_name, _, _) <- fieldTypeInfo pfx field
+  let r_body_core = AppE (VarE rpfx_name) $ VarE x
+      r_body = if fieldIsOptional field == fieldIsOptional optField
+                 then r_body_core
+                 else AppE (VarE 'return) r_body_core
+      f_body = AppE (VarE fpfx_name) $ VarE x
+  return $ [ SigD pfx_name $ ArrowT `AppT` ConT nm `AppT` ftype
+           , FunD pfx_name
+             [ Clause [ConP rnm [VarP x]] (NormalB r_body) []
+             , Clause [ConP fnm [VarP x]] (NormalB f_body) []
+             ]]
+
+-- | Build lense declartions for a field, if the type of the field
+-- is the same in the forthcoming and the real variant.
+buildLens :: (Name, Name) -- ^ names of the forthcoming constructors
+          -> (Name, Name) -- ^ names of the real constructors
+          -> Name -- ^ name of the type
+          -> String -- ^ the field prefix
+          -> Int -- ^ arity
+          -> (Field, Int) -- ^ the Field to generate the lens for, and its
+                          -- position
+          -> Q [Dec]
+buildLens (fnm, fdnm) (rnm, rdnm) nm pfx ar (field, i) = do
+  let optField = makeOptional field
+  if fieldIsOptional field /= fieldIsOptional optField
+     then return []
+     else do
+       let lensnm = mkName $ pfx ++ fieldRecordName  field ++ "L"
+       (accnm, _, ftype) <- fieldTypeInfo pfx field
+       vars <- replicateM ar (newName "x")
+       var <- newName "val"
+       context <- newName "val"
+       let body cn cdn = NormalB
+                           . (ConE cn `AppE`)
+                           . foldl (\e (j, x) -> AppE e . VarE
+                                                   $ if i == j then var else x)
+                             (ConE cdn)
+                          $ zip [0..] vars
+       let setterE = LamE [VarP context, VarP var] $ CaseE (VarE context)
+                        [ Match (ConP fnm [ConP fdnm . set (element i) WildP
+                                             $ map VarP vars])
+                                (body fnm fdnm) []
+                        , Match (ConP rnm [ConP rdnm . set (element i) WildP
+                                             $ map VarP vars])
+                                (body rnm rdnm) []
+                        ]
+       return [ SigD lensnm $ ConT ''Lens' `AppT` ConT nm `AppT` ftype
+              , ValD (VarP lensnm)
+                     (NormalB  $ VarE 'lens `AppE` VarE accnm `AppE` setterE) []
+              ]
+
+-- | Build an object that can have a forthcoming variant.
+-- This will create 3 data types: two objects, prefixed by
+-- "Real" and "Forthcoming", respectively, and a sum type
+-- of those. The JSON representation of the latter will
+-- be a JSON object, dispatching on the "forthcoming" key.
+buildObjectWithForthcoming ::
+  String -- ^ Name of the newly defined type
+  -> String -- ^ base prefix for field names; for the real and fortcoming
+            -- variant, with base prefix will be prefixed with "real"
+            -- and forthcoming, respectively.
+  -> [Field] -- ^ List of fields in the real version
+  -> Q [Dec]
+buildObjectWithForthcoming sname field_pfx fields = do
+  let capitalPrefix = ensureUpper field_pfx
+      forth_nm = "Forthcoming" ++ sname
+      forth_data_nm = forth_nm ++ "Data"
+      forth_pfx = "forthcoming" ++ capitalPrefix
+      real_nm =  "Real" ++ sname
+      real_data_nm = real_nm ++ "Data"
+      real_pfx = "real" ++ capitalPrefix
+  concreteDecls <- buildObject real_data_nm real_pfx fields
+  forthcomingDecls <- buildObject forth_data_nm forth_pfx
+                      (map makeOptional fields)
+  let name = mkName sname
+      real_d = NormalC (mkName real_nm)
+                 [(NotStrict, ConT (mkName real_data_nm))]
+      forth_d = NormalC (mkName forth_nm)
+                  [(NotStrict, ConT (mkName forth_data_nm))]
+      declD = DataD [] name [] [real_d, forth_d] [''Show, ''Eq]
+
+  read_body <- [| branchOnField "forthcoming"
+                  (liftM $(conE $ mkName forth_nm) . JSON.readJSON)
+                  (liftM $(conE $ mkName real_nm) . JSON.readJSON) |]
+  x <- newName "x"
+  show_real_body <- [| JSON.showJSON $(varE x) |]
+  show_forth_body <- [| addField ("forthcoming", JSON.JSBool True)
+                          $ JSON.showJSON $(varE x) |]
+  let rdjson = FunD 'JSON.readJSON [Clause [] (NormalB read_body) []]
+      shjson = FunD 'JSON.showJSON
+                 [ Clause [ConP (mkName real_nm) [VarP x]]
+                    (NormalB show_real_body) []
+                 , Clause [ConP (mkName forth_nm) [VarP x]]
+                    (NormalB show_forth_body) []
+                 ]
+      instJSONdecl = InstanceD [] (AppT (ConT ''JSON.JSON) (ConT name))
+                     [rdjson, shjson]
+  accessors <- liftM concat . flip mapM fields
+                 $ buildAccessor (mkName forth_nm) forth_pfx
+                                 (mkName real_nm) real_pfx
+                                 name field_pfx
+  lenses <- liftM concat . flip mapM (zip fields [0..])
+              $ buildLens (mkName forth_nm, mkName forth_data_nm)
+                          (mkName real_nm, mkName real_data_nm)
+                          name field_pfx (length fields)
+  xs <- newName "xs"
+  fromDictWKeysbody <- [| if ("forthcoming", JSON.JSBool True) `elem` $(varE xs)
+                            then liftM $(conE $ mkName forth_nm)
+                                   (fromDictWKeys $(varE xs))
+                            else liftM $(conE $ mkName real_nm)
+                                   (fromDictWKeys $(varE xs)) |]
+  todictx_r <- [| toDict $(varE x) |]
+  todictx_f <- [| ("forthcoming", JSON.JSBool True) : toDict $(varE x) |]
+  let todict = FunD 'toDict [ Clause [ConP (mkName real_nm) [VarP x]]
+                               (NormalB todictx_r) []
+                            , Clause [ConP (mkName forth_nm) [VarP x]]
+                               (NormalB todictx_f) []
+                            ]
+      fromdict = FunD 'fromDictWKeys [ Clause [VarP xs]
+                                       (NormalB fromDictWKeysbody) [] ]
+      instDict = InstanceD [] (AppT (ConT ''DictObject) (ConT name))
+                 [todict, fromdict]
+  instArray <- genArrayObjectInstance name
+                 (simpleField "forthcoming" [t| Bool |] : fields)
+  let forthPredName = mkName $ field_pfx ++ "Forthcoming"
+  let forthPredDecls = [ SigD forthPredName
+                           $ ArrowT `AppT` ConT name `AppT` ConT ''Bool
+                       , FunD forthPredName
+                         [ Clause [ConP (mkName real_nm) [WildP]]
+                                   (NormalB $ ConE 'False) []
+                         , Clause [ConP (mkName forth_nm) [WildP]]
+                                   (NormalB $ ConE 'True) []
+                         ]
+                       ]
+  return $ concreteDecls ++ forthcomingDecls ++ [declD, instJSONdecl]
+           ++ forthPredDecls ++ accessors ++ lenses ++ [instDict, instArray]
+
 -- | Generates an object definition: data type and its JSON instance.
 buildObjectSerialisation :: String -> [Field] -> Q [Dec]
 buildObjectSerialisation sname fields = do
@@ -1138,13 +1289,17 @@ paramTypeNames :: String -> (String, String)
 paramTypeNames root = ("Filled"  ++ root ++ "Params",
                        "Partial" ++ root ++ "Params")
 
+-- | Compute the name of a full and a partial parameter field.
+paramFieldNames :: String -> Field -> (Name, Name)
+paramFieldNames field_pfx fd =
+  let base = field_pfx ++ fieldRecordName fd
+   in (mkName base, mkName (base ++ "P"))
+
 -- | Compute information about the type of a parameter field.
-paramFieldTypeInfo :: String -> Field -> Q (Name, Strict, Type)
+paramFieldTypeInfo :: String -> Field -> VarStrictTypeQ
 paramFieldTypeInfo field_pfx fd = do
   t <- actualFieldType fd
-  let n = mkName . (++ "P") . (field_pfx ++) .
-          fieldRecordName $ fd
-  return (n, NotStrict, AppT (ConT ''Maybe) t)
+  return (snd $ paramFieldNames field_pfx fd, NotStrict, AppT (ConT ''Maybe) t)
 
 -- | Build a parameter declaration.
 --
@@ -1153,6 +1308,7 @@ paramFieldTypeInfo field_pfx fd = do
 -- fields are optional. Due to the current record syntax issues, the
 -- fields need to be named differrently for the two structures, so the
 -- partial ones get a /P/ suffix.
+-- Also generate a default value for the partial parameters.
 buildParam :: String -> String -> [Field] -> Q [Dec]
 buildParam sname field_pfx fields = do
   let (sname_f, sname_p) = paramTypeNames sname
@@ -1216,37 +1372,56 @@ loadPParamField field = do
       loadexp = [| $(varE 'maybeFromObj) $objvar $objfield |]
   loadFnOpt field loadexp objvar
 
--- | Builds a simple declaration of type @n_x = fromMaybe f_x p_x@.
-buildFromMaybe :: String -> Q Dec
-buildFromMaybe fname =
-  valD (varP (mkName $ "n_" ++ fname))
-         (normalB [| $(varE 'fromMaybe)
-                        $(varNameE $ "f_" ++ fname)
-                        $(varNameE $ "p_" ++ fname) |]) []
-
 -- | Builds a function that executes the filling of partial parameter
 -- from a full copy (similar to Python's fillDict).
 fillParam :: String -> String -> [Field] -> Q [Dec]
 fillParam sname field_pfx fields = do
-  let fnames = map (\fd -> field_pfx ++ fieldRecordName fd) fields
-      (sname_f, sname_p) = paramTypeNames sname
-      oname_f = "fobj"
-      oname_p = "pobj"
+  let (sname_f, sname_p) = paramTypeNames sname
       name_f = mkName sname_f
       name_p = mkName sname_p
-      fun_name = mkName $ "fill" ++ sname ++ "Params"
-      le_full = ValD (ConP name_f (map (VarP . mkName . ("f_" ++)) fnames))
-                (NormalB . VarE . mkName $ oname_f) []
-      le_part = ValD (ConP name_p (map (VarP . mkName . ("p_" ++)) fnames))
-                (NormalB . VarE . mkName $ oname_p) []
-      obj_new = appCons name_f $ map (VarE . mkName . ("n_" ++)) fnames
-  le_new <- mapM buildFromMaybe fnames
-  funt <- [t| $(conT name_f) -> $(conT name_p) -> $(conT name_f) |]
-  let sig = SigD fun_name funt
-      fclause = Clause [VarP (mkName oname_f), VarP (mkName oname_p)]
-                (NormalB $ LetE (le_full:le_part:le_new) obj_new) []
-      fun = FunD fun_name [fclause]
-  return [sig, fun]
+  let (fnames, pnames) = unzip $ map (paramFieldNames field_pfx) fields
+  -- due to apparent bugs in some older GHC versions, we need to add these
+  -- prefixes to avoid "binding shadows ..." errors
+  fbinds <- mapM (newName . ("f_" ++) . nameBase) fnames
+  let fConP = ConP name_f (map VarP fbinds)
+  pbinds <- mapM (newName . ("p_" ++) . nameBase) pnames
+  let pConP = ConP name_p (map VarP pbinds)
+  -- PartialParams instance --------
+  -- fillParams
+  let fromMaybeExp fn pn = AppE (AppE (VarE 'fromMaybe) (VarE fn)) (VarE pn)
+      fupdates = appCons name_f $ zipWith fromMaybeExp fbinds pbinds
+      fclause = Clause [fConP, pConP] (NormalB fupdates) []
+  -- toPartial
+  let tpupdates = appCons name_p $ map (AppE (ConE 'Just) . VarE) fbinds
+      tpclause = Clause [fConP] (NormalB tpupdates) []
+  -- toFilled
+  let tfupdates = appConsApp name_f $ map VarE pbinds
+      tfclause = Clause [pConP] (NormalB tfupdates) []
+  -- the instance
+  let instType = AppT (AppT (ConT ''PartialParams) (ConT name_f)) (ConT name_p)
+  -- Monoid instance for the partial part ----
+  -- mempty
+  let memptyExp = appCons name_p $ map (const $ VarE 'empty) fields
+      memptyClause = Clause [] (NormalB memptyExp) []
+  -- mappend
+  pbinds2 <- mapM (newName . ("p2_" ++) . nameBase) pnames
+  let pConP2 = ConP name_p (map VarP pbinds2)
+  -- note the reversal of 'l' and 'r' in the call to <|>
+  -- as we want the result to be the rightmost value
+  let altExp = zipWith (\l r -> AppE (AppE (VarE '(<|>)) (VarE r)) (VarE l))
+      mappendExp = appCons name_p $ altExp pbinds pbinds2
+      mappendClause = Clause [pConP, pConP2] (NormalB mappendExp) []
+  let monoidType = AppT (ConT ''Monoid) (ConT name_p)
+  -- the instances combined
+  return [ InstanceD [] instType
+                     [ FunD 'fillParams [fclause]
+                     , FunD 'toPartial [tpclause]
+                     , FunD 'toFilled [tfclause]
+                     ]
+         , InstanceD [] monoidType
+                     [ FunD 'mempty [memptyClause]
+                     , FunD 'mappend [mappendClause]
+                     ]]
 
 -- * Template code for exceptions
 
