@@ -64,11 +64,14 @@ import GHC.Exts (IsString(..))
 import System.Time (ClockTime(..))
 import qualified Text.JSON as J
 
+import Test.Ganeti.Query.Language ()
+import Test.Ganeti.SlotMap (genSlotLimit)
 import Test.Ganeti.TestHelper
 import Test.Ganeti.TestCommon
 import Test.Ganeti.Types ()
 
 import qualified Ganeti.Constants as C
+import qualified Ganeti.ConstantUtils as CU
 import Ganeti.Network
 import Ganeti.Objects as Objects
 import qualified Ganeti.Objects.BitArray as BA
@@ -77,6 +80,15 @@ import Ganeti.Types
 
 -- * Arbitrary instances
 
+instance Arbitrary (Container DataCollectorConfig) where
+  arbitrary = do
+    let names = CU.toList C.dataCollectorNames
+    activations <- vector $ length names
+    timeouts <- vector $ length names
+    let configs = zipWith DataCollectorConfig activations timeouts
+    return GenericContainer {
+      fromContainer = Map.fromList $ zip names configs }
+
 $(genArbitrary ''PartialNDParams)
 
 instance Arbitrary Node where
@@ -109,10 +121,20 @@ instance Arbitrary DiskLogicalId where
 -- properties, we only generate disks with no children (FIXME), as
 -- generating recursive datastructures is a bit more work.
 instance Arbitrary Disk where
-  arbitrary = Disk <$> arbitrary <*> pure [] <*> arbitrary
+  arbitrary =
+   frequency [ (2, liftM RealDisk $ RealDiskData <$> arbitrary
+                   <*> pure [] <*> arbitrary
                    <*> arbitrary <*> arbitrary <*> arbitrary
                    <*> arbitrary <*> arbitrary <*> arbitrary
                    <*> arbitrary <*> arbitrary <*> arbitrary
+                   <*> arbitrary)
+             , (1, liftM ForthcomingDisk $ ForthcomingDiskData <$> arbitrary
+                   <*> pure [] <*> arbitrary
+                   <*> arbitrary <*> arbitrary <*> arbitrary
+                   <*> arbitrary <*> arbitrary <*> arbitrary
+                   <*> arbitrary <*> arbitrary <*> arbitrary
+                   <*> arbitrary)
+             ]
 
 -- FIXME: we should generate proper values, >=0, etc., but this is
 -- hard for partial ones, where all must be wrapped in a 'Maybe'
@@ -126,9 +148,50 @@ $(genArbitrary ''PartialNicParams)
 
 $(genArbitrary ''PartialNic)
 
-instance Arbitrary Instance where
+instance Arbitrary ForthcomingInstanceData where
   arbitrary =
-    Instance
+    ForthcomingInstanceData
+      -- name
+      <$> genMaybe genFQDN
+      -- primary node
+      <*> genMaybe genFQDN
+      -- OS
+      <*> genMaybe genFQDN
+      -- hypervisor
+      <*> arbitrary
+      -- hvparams
+      -- FIXME: add non-empty hvparams when they're a proper type
+      <*> pure (GenericContainer Map.empty)
+      -- beparams
+      <*> arbitrary
+      -- osparams
+      <*> pure (GenericContainer Map.empty)
+      -- osparams_private
+      <*> pure (GenericContainer Map.empty)
+      -- admin_state
+      <*> genMaybe arbitrary
+      -- admin_state_source
+      <*> genMaybe arbitrary
+      -- nics
+      <*> arbitrary
+      -- disks
+      <*> vectorOf 5 arbitrary
+      -- disks active
+      <*> genMaybe arbitrary
+      -- network port
+      <*> arbitrary
+      -- ts
+      <*> arbitrary <*> arbitrary
+      -- uuid
+      <*> arbitrary
+      -- serial
+      <*> arbitrary
+      -- tags
+      <*> (Set.fromList <$> genTags)
+
+instance Arbitrary RealInstanceData where
+  arbitrary =
+    RealInstanceData
       -- name
       <$> genFQDN
       -- primary node
@@ -154,8 +217,6 @@ instance Arbitrary Instance where
       <*> arbitrary
       -- disks
       <*> vectorOf 5 arbitrary
-      -- disk template
-      <*> arbitrary
       -- disks active
       <*> arbitrary
       -- network port
@@ -169,11 +230,16 @@ instance Arbitrary Instance where
       -- tags
       <*> (Set.fromList <$> genTags)
 
+instance Arbitrary Instance where
+  arbitrary = frequency [ (1, ForthcomingInstance <$> arbitrary)
+                        , (3, RealInstance <$> arbitrary)
+                        ]
+
 -- | Generates an instance that is connected to the given networks
 -- and possibly some other networks
 genInstWithNets :: [String] -> Gen Instance
 genInstWithNets nets = do
-  plain_inst <- arbitrary
+  plain_inst <- RealInstance <$> arbitrary
   enhanceInstWithNets plain_inst nets
 
 -- | Generates an instance that is connected to some networks
@@ -195,13 +261,17 @@ enhanceInstWithNets inst nets = do
   let genNic net = PartialNic mac ip nicparams net name uuid
       partial_nics = map (genNic . Just)
                          (List.nub (nets ++ more_nets))
-      new_inst = inst { instNics = partial_nics }
+      new_inst = case inst of
+                   RealInstance rinst ->
+                     RealInstance rinst { realInstNics = partial_nics }
+                   ForthcomingInstance _ -> inst
   return new_inst
 
 genDiskWithChildren :: Int -> Gen Disk
 genDiskWithChildren num_children = do
   logicalid <- arbitrary
   children <- vectorOf num_children (genDiskWithChildren 0)
+  nodes <- arbitrary
   ivname <- genName
   size <- arbitrary
   mode <- arbitrary
@@ -211,8 +281,8 @@ genDiskWithChildren num_children = do
   uuid <- genName
   serial <- arbitrary
   time <- arbitrary
-  return $
-    Disk logicalid children ivname size mode name
+  return . RealDisk $
+    RealDiskData logicalid children nodes ivname size mode name
       spindles params uuid serial time time
 
 genDisk :: Gen Disk
@@ -278,6 +348,30 @@ instance Arbitrary AddressPool where
 instance Arbitrary Network where
   arbitrary = genValidNetwork
 
+instance Arbitrary FilterAction where
+  arbitrary = oneof
+    [ pure Accept
+    , pure Pause
+    , pure Reject
+    , pure Continue
+    , RateLimit <$> genSlotLimit
+    ]
+
+instance Arbitrary FilterPredicate where
+  arbitrary = oneof
+    [ FPJobId <$> arbitrary
+    , FPOpCode <$> arbitrary
+    , FPReason <$> arbitrary
+    ]
+
+instance Arbitrary FilterRule where
+  arbitrary = FilterRule <$> arbitrary
+                         <*> arbitrary
+                         <*> arbitrary
+                         <*> arbitrary
+                         <*> arbitrary
+                         <*> genUUID
+
 -- | Generates a network instance with minimum netmasks of /24. Generating
 -- bigger networks slows down the tests, because long bit strings are generated
 -- for the reservations.
@@ -333,6 +427,7 @@ genEmptyCluster ncount = do
       continsts = GenericContainer Map.empty
       networks = GenericContainer Map.empty
       disks = GenericContainer Map.empty
+      filters = GenericContainer Map.empty
   let contgroups = GenericContainer $ Map.singleton guuid grp
   serial <- arbitrary
   -- timestamp fields
@@ -340,7 +435,7 @@ genEmptyCluster ncount = do
   mtime <- arbitrary
   cluster <- resize 8 arbitrary
   let c = ConfigData version cluster contnodes contgroups continsts networks
-            disks ctime mtime serial
+            disks filters ctime mtime serial
   return c
 
 -- | FIXME: make an even simpler base version of creating a cluster.
@@ -421,6 +516,14 @@ prop_AddressPool_serialisation = testSerialisation
 prop_Network_serialisation :: Network -> Property
 prop_Network_serialisation = testSerialisation
 
+-- | Check that filter action serialisation is idempotent.
+prop_FilterAction_serialisation :: FilterAction -> Property
+prop_FilterAction_serialisation = testSerialisation
+
+-- | Check that filter predicate serialisation is idempotent.
+prop_FilterPredicate_serialisation :: FilterPredicate -> Property
+prop_FilterPredicate_serialisation = testSerialisation
+
 -- | Check config serialisation.
 prop_Config_serialisation :: Property
 prop_Config_serialisation =
@@ -620,8 +723,9 @@ caseIncludeLogicalIdPlain =
       lv_name = "1234sdf-qwef-2134-asff-asd2-23145d.data" :: String
       lv = LogicalVolume vg_name lv_name
       time = TOD 0 0
-      d =
-        Disk (LIDPlain lv) [] "diskname" 1000 DiskRdWr
+      d = RealDisk $
+        RealDiskData (LIDPlain lv) [] ["node1.example.com"] "diskname"
+          1000 DiskRdWr
           Nothing Nothing Nothing "asdfgr-1234-5123-daf3-sdfw-134f43"
           0 time time
   in
@@ -634,14 +738,19 @@ caseIncludeLogicalIdDrbd =
   let vg_name = "xenvg" :: String
       lv_name = "1234sdf-qwef-2134-asff-asd2-23145d.data" :: String
       time = TOD 0 0
-      d =
-        Disk
+      d = RealDisk $
+        RealDiskData
           (LIDDrbd8 "node1.example.com" "node2.example.com" 2000 1 5 "secret")
-          [ Disk (mkLIDPlain "onevg" "onelv") [] "disk1" 1000 DiskRdWr Nothing
-              Nothing Nothing "145145-asdf-sdf2-2134-asfd-534g2x" 0 time time
-          , Disk (mkLIDPlain vg_name lv_name) [] "disk2" 1000 DiskRdWr Nothing
-              Nothing Nothing "6gd3sd-423f-ag2j-563b-dg34-gj3fse" 0 time time
-          ] "diskname" 1000 DiskRdWr Nothing Nothing Nothing
+          [ RealDisk $ RealDiskData (mkLIDPlain "onevg" "onelv") []
+              ["node1.example.com", "node2.example.com"] "disk1" 1000 DiskRdWr
+              Nothing Nothing Nothing "145145-asdf-sdf2-2134-asfd-534g2x"
+              0 time time
+          , RealDisk $ RealDiskData (mkLIDPlain vg_name lv_name) []
+              ["node1.example.com", "node2.example.com"] "disk2" 1000 DiskRdWr
+              Nothing Nothing Nothing "6gd3sd-423f-ag2j-563b-dg34-gj3fse"
+              0 time time
+          ] ["node1.example.com", "node2.example.com"] "diskname" 1000 DiskRdWr
+          Nothing Nothing Nothing
           "asdfgr-1234-5123-daf3-sdfw-134f43" 0 time time
   in
     HUnit.assertBool "Unable to detect that plain Disk includes logical ID" $
@@ -653,9 +762,10 @@ caseNotIncludeLogicalIdPlain =
   let vg_name = "xenvg" :: String
       lv_name = "1234sdf-qwef-2134-asff-asd2-23145d.data" :: String
       time = TOD 0 0
-      d =
-        Disk (mkLIDPlain "othervg" "otherlv") [] "diskname" 1000 DiskRdWr
-          Nothing Nothing Nothing "asdfgr-1234-5123-daf3-sdfw-134f43"
+      d = RealDisk $
+        RealDiskData (mkLIDPlain "othervg" "otherlv") [] ["node1.example.com"]
+          "diskname" 1000 DiskRdWr Nothing Nothing Nothing
+          "asdfgr-1234-5123-daf3-sdfw-134f43"
           0 time time
   in
     HUnit.assertBool "Unable to detect that plain Disk includes logical ID" $
@@ -672,6 +782,8 @@ testSuite "Objects"
   , 'prop_Network_serialisation
   , 'prop_Node_serialisation
   , 'prop_Config_serialisation
+  , 'prop_FilterAction_serialisation
+  , 'prop_FilterPredicate_serialisation
   , 'casePyCompatNetworks
   , 'casePyCompatNodegroups
   , 'casePyCompatInstances
