@@ -55,10 +55,11 @@ import Test.Ganeti.Objects ()
 import Test.Ganeti.Query.Language ()
 import Test.Ganeti.TestHelper
 import Test.Ganeti.TestCommon
-import Test.Ganeti.Types ()
+import Test.Ganeti.Types (genReasonTrail)
 
 import Ganeti.BasicTypes
 import qualified Ganeti.Constants as C
+import qualified Ganeti.ConstantUtils as CU
 import qualified Ganeti.OpCodes as OpCodes
 import Ganeti.Types
 import Ganeti.OpParams
@@ -68,13 +69,10 @@ import Ganeti.JSON
 
 -- * Arbitrary instances
 
-instance (Ord k, Arbitrary k, Arbitrary a) => Arbitrary (Map.Map k a) where
-  arbitrary = Map.fromList <$> arbitrary
-
 arbitraryOpTagsGet :: Gen OpCodes.OpCode
 arbitraryOpTagsGet = do
   kind <- arbitrary
-  OpCodes.OpTagsSet kind <$> arbitrary <*> genOpCodesTagName kind
+  OpCodes.OpTagsSet kind <$> genTags <*> genOpCodesTagName kind
 
 arbitraryOpTagsSet :: Gen OpCodes.OpCode
 arbitraryOpTagsSet = do
@@ -103,7 +101,8 @@ instance Arbitrary IDiskParams where
   arbitrary = IDiskParams <$> arbitrary <*> arbitrary <*>
               genMaybe genNameNE <*> genMaybe genNameNE <*>
               genMaybe genNameNE <*> genMaybe genNameNE <*>
-              genMaybe genNameNE <*> arbitrary <*> genAndRestArguments
+              genMaybe genNameNE <*> arbitrary <*>
+              genMaybe genNameNE <*> genAndRestArguments
 
 instance Arbitrary RecreateDisksInfo where
   arbitrary = oneof [ pure RecreateDisksAll
@@ -127,6 +126,18 @@ instance Arbitrary ExportTarget where
                     , ExportTargetRemote <$> pure []
                     ]
 
+arbitraryDataCollector :: Gen (Container Bool)
+arbitraryDataCollector = do
+  els <-  listOf . elements $ CU.toList C.dataCollectorNames
+  activation <- vector $ length els
+  return . GenericContainer . Map.fromList $ zip els activation
+
+arbitraryDataCollectorInterval :: Gen (Maybe (Container Int))
+arbitraryDataCollectorInterval = do
+  els <-  listOf . elements $ CU.toList C.dataCollectorNames
+  intervals <- vector $ length els
+  genMaybe . return . containerFromList $ zip els intervals
+
 instance Arbitrary OpCodes.OpCode where
   arbitrary = do
     op_id <- elements OpCodes.allOpIDs
@@ -147,7 +158,7 @@ instance Arbitrary OpCodes.OpCode where
         OpCodes.OpInstanceMigrate <$> genFQDN <*> return Nothing <*>
           arbitrary <*> arbitrary <*> genMaybe genNodeNameNE <*>
           return Nothing <*> arbitrary <*> arbitrary <*> arbitrary <*>
-          genMaybe genNameNE <*> arbitrary
+          genMaybe genNameNE <*> arbitrary <*> arbitrary
       "OP_TAGS_GET" ->
         arbitraryOpTagsGet
       "OP_TAGS_SEARCH" ->
@@ -158,19 +169,20 @@ instance Arbitrary OpCodes.OpCode where
         arbitraryOpTagsDel
       "OP_CLUSTER_POST_INIT" -> pure OpCodes.OpClusterPostInit
       "OP_CLUSTER_RENEW_CRYPTO" -> OpCodes.OpClusterRenewCrypto <$>
-         arbitrary <*> arbitrary
+         arbitrary <*> arbitrary <*> arbitrary <*> arbitrary
       "OP_CLUSTER_DESTROY" -> pure OpCodes.OpClusterDestroy
       "OP_CLUSTER_QUERY" -> pure OpCodes.OpClusterQuery
       "OP_CLUSTER_VERIFY" ->
         OpCodes.OpClusterVerify <$> arbitrary <*> arbitrary <*>
           genListSet Nothing <*> genListSet Nothing <*> arbitrary <*>
-          genMaybe genNameNE
+          genMaybe genNameNE <*> arbitrary
       "OP_CLUSTER_VERIFY_CONFIG" ->
         OpCodes.OpClusterVerifyConfig <$> arbitrary <*> arbitrary <*>
           genListSet Nothing <*> arbitrary
       "OP_CLUSTER_VERIFY_GROUP" ->
         OpCodes.OpClusterVerifyGroup <$> genNameNE <*> arbitrary <*>
-          arbitrary <*> genListSet Nothing <*> genListSet Nothing <*> arbitrary
+          arbitrary <*> genListSet Nothing <*> genListSet Nothing <*>
+          arbitrary <*> arbitrary
       "OP_CLUSTER_VERIFY_DISKS" -> pure OpCodes.OpClusterVerifyDisks
       "OP_GROUP_VERIFY_DISKS" ->
         OpCodes.OpGroupVerifyDisks <$> genNameNE
@@ -185,7 +197,7 @@ instance Arbitrary OpCodes.OpCode where
           <$> arbitrary                    -- force
           <*> emptyMUD                     -- hv_state
           <*> emptyMUD                     -- disk_state
-          <*> arbitrary                    -- vg_name
+          <*> genMaybe genName             -- vg_name
           <*> genMaybe arbitrary           -- enabled_hypervisors
           <*> genMaybe genEmptyContainer   -- hvparams
           <*> emptyMUD                     -- beparams
@@ -204,36 +216,50 @@ instance Arbitrary OpCodes.OpCode where
           <*> arbitrary                    -- nicparams
           <*> emptyMUD                     -- ndparams
           <*> emptyMUD                     -- ipolicy
-          <*> arbitrary                    -- drbd_helper
-          <*> arbitrary                    -- default_iallocator
+          <*> genMaybe genPrintableAsciiString
+                                           -- drbd_helper
+          <*> genMaybe genPrintableAsciiString
+                                           -- default_iallocator
           <*> emptyMUD                     -- default_iallocator_params
           <*> genMaybe genMacPrefix        -- mac_prefix
-          <*> arbitrary                    -- master_netdev
+          <*> genMaybe genPrintableAsciiString
+                                           -- master_netdev
           <*> arbitrary                    -- master_netmask
-          <*> arbitrary                    -- reserved_lvs
-          <*> arbitrary                    -- hidden_os
-          <*> arbitrary                    -- blacklisted_os
+          <*> genMaybe (listOf genPrintableAsciiStringNE)
+                                           -- reserved_lvs
+          <*> genMaybe (listOf ((,) <$> arbitrary
+                                    <*> genPrintableAsciiStringNE))
+                                           -- hidden_os
+          <*> genMaybe (listOf ((,) <$> arbitrary
+                                    <*> genPrintableAsciiStringNE))
+                                           -- blacklisted_os
           <*> arbitrary                    -- use_external_mip_script
           <*> arbitrary                    -- enabled_disk_templates
           <*> arbitrary                    -- modify_etc_hosts
           <*> genMaybe genName             -- file_storage_dir
           <*> genMaybe genName             -- shared_file_storage_dir
           <*> genMaybe genName             -- gluster_file_storage_dir
-          <*> arbitrary                    -- install_image
-          <*> arbitrary                    -- instance_communication_network
-          <*> arbitrary                    -- zeroing_image
-          <*> arbitrary                    -- compression_tools
+          <*> genMaybe genPrintableAsciiString
+                                           -- install_image
+          <*> genMaybe genPrintableAsciiString
+                                           -- instance_communication_network
+          <*> genMaybe genPrintableAsciiString
+                                           -- zeroing_image
+          <*> genMaybe (listOf genPrintableAsciiStringNE)
+                                           -- compression_tools
           <*> arbitrary                    -- enabled_user_shutdown
+          <*> genMaybe arbitraryDataCollector   -- enabled_data_collectors
+          <*> arbitraryDataCollectorInterval   -- data_collector_interval
       "OP_CLUSTER_REDIST_CONF" -> pure OpCodes.OpClusterRedistConf
       "OP_CLUSTER_ACTIVATE_MASTER_IP" ->
         pure OpCodes.OpClusterActivateMasterIp
       "OP_CLUSTER_DEACTIVATE_MASTER_IP" ->
         pure OpCodes.OpClusterDeactivateMasterIp
       "OP_QUERY" ->
-        OpCodes.OpQuery <$> arbitrary <*> arbitrary <*> arbitrary <*>
+        OpCodes.OpQuery <$> arbitrary <*> arbitrary <*> genNamesNE <*>
         pure Nothing
       "OP_QUERY_FIELDS" ->
-        OpCodes.OpQueryFields <$> arbitrary <*> arbitrary
+        OpCodes.OpQueryFields <$> arbitrary <*> genMaybe genNamesNE
       "OP_OOB_COMMAND" ->
         OpCodes.OpOobCommand <$> genNodeNamesNE <*> return Nothing <*>
           arbitrary <*> arbitrary <*> arbitrary <*>
@@ -243,11 +269,12 @@ instance Arbitrary OpCodes.OpCode where
       "OP_NODE_ADD" ->
         OpCodes.OpNodeAdd <$> genNodeNameNE <*> emptyMUD <*> emptyMUD <*>
           genMaybe genNameNE <*> genMaybe genNameNE <*> arbitrary <*>
-          genMaybe genNameNE <*> arbitrary <*> arbitrary <*> emptyMUD
+          genMaybe genNameNE <*> arbitrary <*> arbitrary <*> emptyMUD <*>
+          arbitrary
       "OP_NODE_QUERYVOLS" ->
-        OpCodes.OpNodeQueryvols <$> arbitrary <*> genNodeNamesNE
+        OpCodes.OpNodeQueryvols <$> genNamesNE <*> genNodeNamesNE
       "OP_NODE_QUERY_STORAGE" ->
-        OpCodes.OpNodeQueryStorage <$> arbitrary <*> arbitrary <*>
+        OpCodes.OpNodeQueryStorage <$> genNamesNE <*> arbitrary <*>
           genNodeNamesNE <*> genMaybe genNameNE
       "OP_NODE_MODIFY_STORAGE" ->
         OpCodes.OpNodeModifyStorage <$> genNodeNameNE <*> return Nothing <*>
@@ -270,7 +297,7 @@ instance Arbitrary OpCodes.OpCode where
       "OP_NODE_EVACUATE" ->
         OpCodes.OpNodeEvacuate <$> arbitrary <*> genNodeNameNE <*>
           return Nothing <*> genMaybe genNodeNameNE <*> return Nothing <*>
-          genMaybe genNameNE <*> arbitrary
+          genMaybe genNameNE <*> arbitrary <*> arbitrary
       "OP_INSTANCE_CREATE" ->
         OpCodes.OpInstanceCreate
           <$> genFQDN                         -- instance_name
@@ -282,6 +309,7 @@ instance Arbitrary OpCodes.OpCode where
           <*> pure emptyJSObject              -- beparams
           <*> arbitrary                       -- disks
           <*> arbitrary                       -- disk_template
+          <*> genMaybe genNameNE              -- group_name
           <*> arbitrary                       -- file_driver
           <*> genMaybe genNameNE              -- file_storage_dir
           <*> pure emptyJSObject              -- hvparams
@@ -308,8 +336,10 @@ instance Arbitrary OpCodes.OpCode where
           <*> return Nothing                  -- src_node
           <*> genMaybe genNodeNameNE          -- src_node_uuid
           <*> genMaybe genNameNE              -- src_path
-          <*> arbitrary                       -- compress
+          <*> genPrintableAsciiString         -- compress
           <*> arbitrary                       -- start
+          <*> arbitrary                       -- forthcoming
+          <*> arbitrary                       -- commit
           <*> (genTags >>= mapM mkNonEmpty)   -- tags
           <*> arbitrary                       -- instance_communication
           <*> arbitrary                       -- helper_startup_timeout
@@ -347,7 +377,7 @@ instance Arbitrary OpCodes.OpCode where
       "OP_INSTANCE_MOVE" ->
         OpCodes.OpInstanceMove <$> genFQDN <*> return Nothing <*>
           arbitrary <*> arbitrary <*> genNodeNameNE <*> return Nothing <*>
-          arbitrary <*> arbitrary
+          genPrintableAsciiString <*> arbitrary
       "OP_INSTANCE_CONSOLE" -> OpCodes.OpInstanceConsole <$> genFQDN <*>
           return Nothing
       "OP_INSTANCE_ACTIVATE_DISKS" ->
@@ -376,6 +406,9 @@ instance Arbitrary OpCodes.OpCode where
           <*> arbitrary                       -- runtime_mem
           <*> pure emptyJSObject              -- hvparams
           <*> arbitrary                       -- disk_template
+          <*> pure emptyJSObject              -- ext_params
+          <*> arbitrary                       -- file_driver
+          <*> genMaybe genNameNE              -- file_storage_dir
           <*> genMaybe genNodeNameNE          -- pnode
           <*> return Nothing                  -- pnode_uuid
           <*> genMaybe genNodeNameNE          -- remote_node
@@ -421,11 +454,23 @@ instance Arbitrary OpCodes.OpCode where
       "OP_BACKUP_PREPARE" ->
         OpCodes.OpBackupPrepare <$> genFQDN <*> return Nothing <*> arbitrary
       "OP_BACKUP_EXPORT" ->
-        OpCodes.OpBackupExport <$> genFQDN <*> return Nothing <*>
-          arbitrary <*> arbitrary <*> arbitrary <*> return Nothing <*>
-          arbitrary <*> arbitrary <*> arbitrary <*> arbitrary <*>
-          genMaybe (pure []) <*> genMaybe genNameNE <*> arbitrary <*>
-          arbitrary <*> arbitrary
+        OpCodes.OpBackupExport
+          <$> genFQDN                  -- instance_name
+          <*> return Nothing           -- instance_uuid
+          <*> genPrintableAsciiString  -- compress
+          <*> arbitrary                -- shutdown_timeout
+          <*> arbitrary                -- target_node
+          <*> return Nothing           -- target_node_uuid
+          <*> arbitrary                -- shutdown
+          <*> arbitrary                -- remove_instance
+          <*> arbitrary                -- ignore_remove_failures
+          <*> arbitrary                -- mode
+          <*> genMaybe (pure [])       -- x509_key_name
+          <*> genMaybe genNameNE       -- destination_x509_ca
+          <*> arbitrary                -- zero_free_space
+          <*> arbitrary                -- zeroing_timeout_fixed
+          <*> arbitrary                -- zeroing_timeout_per_mib
+          <*> arbitrary                -- long_sleep
       "OP_BACKUP_REMOVE" ->
         OpCodes.OpBackupRemove <$> genFQDN <*> return Nothing
       "OP_TEST_ALLOCATOR" ->
@@ -435,7 +480,8 @@ instance Arbitrary OpCodes.OpCode where
           (genTags >>= mapM mkNonEmpty) <*>
           arbitrary <*> arbitrary <*> genMaybe genNameNE <*>
           arbitrary <*> genMaybe genNodeNamesNE <*> arbitrary <*>
-          genMaybe genNamesNE <*> arbitrary <*> arbitrary
+          genMaybe genNamesNE <*> arbitrary <*> arbitrary <*>
+          genMaybe genNameNE
       "OP_TEST_JQUEUE" ->
         OpCodes.OpTestJqueue <$> arbitrary <*> arbitrary <*>
           resize 20 (listOf genFQDN) <*> arbitrary
@@ -456,7 +502,7 @@ instance Arbitrary OpCodes.OpCode where
           genMaybe (listOf genIPv4Address)
       "OP_NETWORK_CONNECT" ->
         OpCodes.OpNetworkConnect <$> genNameNE <*> genNameNE <*>
-          arbitrary <*> genNameNE <*> arbitrary <*> arbitrary
+          arbitrary <*> genNameNE <*> genPrintableAsciiString <*> arbitrary
       "OP_NETWORK_DISCONNECT" ->
         OpCodes.OpNetworkDisconnect <$> genNameNE <*> genNameNE
       "OP_RESTRICTED_COMMAND" ->
@@ -464,16 +510,6 @@ instance Arbitrary OpCodes.OpCode where
           return Nothing <*> genNameNE
       _ -> fail $ "Undefined arbitrary for opcode " ++ op_id
 
--- | Generates one element of a reason trail
-genReasonElem :: Gen ReasonElem
-genReasonElem = (,,) <$> genFQDN <*> genFQDN <*> arbitrary
-
--- | Generates a reason trail
-genReasonTrail :: Gen ReasonTrail
-genReasonTrail = do
-  size <- choose (0, 10)
-  vectorOf size genReasonElem
-
 instance Arbitrary OpCodes.CommonOpParams where
   arbitrary = OpCodes.CommonOpParams <$> arbitrary <*> arbitrary <*>
                 arbitrary <*> resize 5 arbitrary <*> genMaybe genName <*>
