@@ -48,10 +48,14 @@ module Test.Ganeti.TestCommon
   , (/=?)
   , failTest
   , passTest
+  , stableCover
   , pythonCmd
   , runPython
   , checkPythonResult
   , DNSChar(..)
+  , genPrintableAsciiChar
+  , genPrintableAsciiString
+  , genPrintableAsciiStringNE
   , genName
   , genFQDN
   , genUUID
@@ -84,17 +88,10 @@ module Test.Ganeti.TestCommon
   , genNonNegative
   , relativeError
   , getTempFileName
+  , listOfUniqueBy
   , counterexample
   ) where
 
--- The following macro is just a temporary solution for 2.12 and 2.13.
--- Since 2.14 cabal creates proper macros for all dependencies.
-#define MIN_VERSION_QuickCheck(maj,min,rev) \
-  (((maj)<QUICKCHECK_MAJOR)|| \
-   (((maj)==QUICKCHECK_MAJOR)&&((min)<=QUICKCHECK_MINOR))|| \
-   (((maj)==QUICKCHECK_MAJOR)&&((min)==QUICKCHECK_MINOR)&& \
-    ((rev)<=QUICKCHECK_REV)))
-
 import Control.Applicative
 import Control.Exception (catchJust)
 import Control.Monad
@@ -122,6 +119,14 @@ import Numeric
 import qualified Ganeti.BasicTypes as BasicTypes
 import Ganeti.JSON (ArrayObject(..))
 import Ganeti.Types
+import Ganeti.Utils.Monad (unfoldrM)
+
+-- * Arbitrary orphan instances
+
+instance (Ord k, Arbitrary k, Arbitrary a) => Arbitrary (M.Map k a) where
+  arbitrary = M.fromList <$> arbitrary
+  shrink m = M.fromList <$> shrink (M.toList m)
+
 
 -- * Constants
 
@@ -185,6 +190,18 @@ failTest msg = counterexample msg False
 passTest :: Property
 passTest = property True
 
+-- | A stable version of QuickCheck's `cover`. In its current implementation,
+-- cover will not detect insufficient coverage if the actual coverage in the
+-- sample is 0. Work around this by lifting the probability to at least
+-- 10 percent.
+-- The underlying issue is tracked at
+-- https://github.com/nick8325/quickcheck/issues/26
+stableCover :: Testable prop => Bool -> Int -> String -> prop -> Property
+stableCover p percent s prop =
+  let newlabel = "(stabilized to at least 10%) " ++ s
+  in forAll (frequency [(1, return True), (9, return False)]) $ \ basechance ->
+     cover (basechance || p) (10 + (percent * 9 `div` 10)) newlabel prop
+
 -- | Return the python binary to use. If the PYTHON environment
 -- variable is defined, use its value, otherwise use just \"python\".
 pythonCmd :: IO String
@@ -217,6 +234,24 @@ instance Arbitrary DNSChar where
 instance Show DNSChar where
   show = show . dnsGetChar
 
+-- * Generators
+
+-- | Generates printable ASCII characters (from ' ' to '~').
+genPrintableAsciiChar :: Gen Char
+genPrintableAsciiChar = choose ('\x20', '\x7e')
+
+-- | Generates a short string (0 <= n <= 40 chars) from printable ASCII.
+genPrintableAsciiString :: Gen String
+genPrintableAsciiString = do
+  n <- choose (0, 40)
+  vectorOf n genPrintableAsciiChar
+
+-- | Generates a short string (1 <= n <= 40 chars) from printable ASCII.
+genPrintableAsciiStringNE :: Gen NonEmptyString
+genPrintableAsciiStringNE = do
+  n <- choose (1, 40)
+  vectorOf n genPrintableAsciiChar >>= mkNonEmpty
+
 -- | Generates a single name component.
 genName :: Gen String
 genName = do
@@ -517,6 +552,40 @@ getTempFileName filename = do
   return fpath
 
 
+-- | @listOfUniqueBy gen keyFun forbidden@: Generates a list of random length,
+-- where all generated elements will be unique by the keying function
+-- @keyFun@. They will also be distinct from all elements in @forbidden@ by
+-- the keying function.
+--
+-- As for 'listOf', the maximum output length depends on the size parameter.
+--
+-- Example:
+--
+-- > listOfUniqueBy (arbitrary :: Gen String) (length) ["hey"]
+-- > -- Generates a list of strings of different length, but not of length 3.
+--
+-- The passed @gen@ should not make key collisions too likely, since the
+-- implementation uses `suchThat`, looping until enough unique elements
+-- have been generated. If the @gen@ makes collisions likely, this function
+-- will consequently be slow, or not terminate if it is not possible to
+-- generate enough elements, like in:
+--
+-- > listOfUniqueBy (arbitrary :: Gen Int) (`mod` 2) []
+-- > -- May not terminate depending on the size parameter of the Gen,
+-- > -- since there are only 2 unique keys (0 and 1).
+listOfUniqueBy :: (Ord b) => Gen a -> (a -> b) -> [a] -> Gen [a]
+listOfUniqueBy gen keyFun forbidden = do
+  let keysOf = Set.fromList . map keyFun
+
+  k <- sized $ \n -> choose (0, n)
+  flip unfoldrM (0, keysOf forbidden) $ \(i, usedKeys) ->
+    if i == k
+      then return Nothing
+      else do
+        x <- gen `suchThat` ((`Set.notMember` usedKeys) . keyFun)
+        return $ Just (x, (i + 1, Set.insert (keyFun x) usedKeys))
+
+
 #if !MIN_VERSION_QuickCheck(2,7,0)
 counterexample :: Testable prop => String -> prop -> Property
 counterexample = QC.printTestCase
