@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 
-# Copyright (C) 2010, 2012, 2013 Google Inc.
+# Copyright (C) 2010, 2012, 2013, 2014, 2015 Google Inc.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -38,14 +38,17 @@ import tempfile
 import operator
 import json
 
+from ganeti import cli
 from ganeti import constants
 from ganeti import utils
 from ganeti import serializer
 from ganeti import netutils
 
 from ganeti.utils import version
+from ganeti.tools.cfgupgrade import CfgUpgrade, ParseOptions, Error
 
 import testutils
+import mock
 
 
 def GetMinimalConfig():
@@ -55,6 +58,7 @@ def GetMinimalConfig():
       "master_node": "node1-uuid",
       "ipolicy": None,
       "default_iallocator_params": {},
+      "diskparams": {},
       "ndparams": {},
       "candidate_certs": {},
       "install_image": "",
@@ -62,10 +66,19 @@ def GetMinimalConfig():
       "zeroing_image": "",
       "compression_tools": constants.IEC_DEFAULT_TOOLS,
       "enabled_user_shutdown": False,
+      "data_collectors": {
+        "diskstats": { "active": True, "interval": 5000000 },
+        "drbd": { "active": True, "interval": 5000000 },
+        "lv": { "active": True, "interval": 5000000 },
+        "inst-status-xen": { "active": True, "interval": 5000000 },
+        "cpu-avg-load": { "active": True, "interval": 5000000 },
+        "xen-cpu-avg-load": { "active": True, "interval": 5000000 },
+      },
     },
     "instances": {},
     "disks": {},
     "networks": {},
+    "filters": {},
     "nodegroups": {},
     "nodes": {
       "node1-uuid": {
@@ -77,23 +90,28 @@ def GetMinimalConfig():
 
 
 def _RunUpgrade(path, dry_run, no_verify, ignore_hostname=True,
-                downgrade=False):
-  cmd = [sys.executable, "%s/tools/cfgupgrade" % testutils.GetSourceDir(),
-         "--debug", "--force", "--path=%s" % path, "--confdir=%s" % path]
+                downgrade=False, fail=False):
+  args = ["--debug", "--force", "--path=%s" % path, "--confdir=%s" % path]
 
   if ignore_hostname:
-    cmd.append("--ignore-hostname")
+    args.append("--ignore-hostname")
   if dry_run:
-    cmd.append("--dry-run")
+    args.append("--dry-run")
   if no_verify:
-    cmd.append("--no-verify")
+    args.append("--no-verify")
   if downgrade:
-    cmd.append("--downgrade")
+    args.append("--downgrade")
+
+  opts, args = ParseOptions(args=args)
+  upgrade = CfgUpgrade(opts, args)
+
+  with mock.patch('sys.exit'):
+    with mock.patch.object(upgrade, 'SetupLogging'):
+      with mock.patch.object(cli, 'ToStderr'):
+        upgrade.Run()
+    if sys.exit.called:
+      raise Error("upgrade failed")
 
-  result = utils.RunCmd(cmd, cwd=os.getcwd())
-  if result.failed:
-    raise Exception("cfgupgrade failed: %s, output %r" %
-                    (result.fail_reason, result.output))
 
 
 class TestCfgupgrade(unittest.TestCase):
@@ -396,6 +414,18 @@ class TestCfgupgrade(unittest.TestCase):
   def testUpgradeFullConfigFrom_2_10(self):
     self._TestUpgradeFromFile("cluster_config_2.10.json", False)
 
+  def testUpgradeFullConfigFrom_2_11(self):
+    self._TestUpgradeFromFile("cluster_config_2.11.json", False)
+
+  def testUpgradeFullConfigFrom_2_12(self):
+    self._TestUpgradeFromFile("cluster_config_2.12.json", False)
+
+  def testUpgradeFullConfigFrom_2_13(self):
+    self._TestUpgradeFromFile("cluster_config_2.13.json", False)
+
+  def testUpgradeFullConfigFrom_2_14(self):
+    self._TestUpgradeFromFile("cluster_config_2.14.json", False)
+
   def testUpgradeCurrent(self):
     self._TestSimpleUpgrade(constants.CONFIG_VERSION, False)
 
@@ -413,11 +443,12 @@ class TestCfgupgrade(unittest.TestCase):
   def testDowngradeFullConfig(self):
     """Test for upgrade + downgrade combination."""
     # This test can work only with the previous version of a configuration!
-    oldconfname = "cluster_config_2.11.json"
+    oldconfname = "cluster_config_2.14.json"
     self._TestUpgradeFromFile(oldconfname, False)
     _RunUpgrade(self.tmpdir, False, True, downgrade=True)
     oldconf = self._LoadTestDataConfig(oldconfname)
     newconf = self._LoadConfig()
+    self.maxDiff = None
     self.assertEqual(oldconf, newconf)
 
   def testDowngradeFullConfigBackwardFrom_2_7(self):
@@ -472,5 +503,6 @@ class TestCfgupgrade(unittest.TestCase):
     newconf = self._LoadConfig()
     self.assertEqual(oldconf["version"], newconf["version"])
 
+
 if __name__ == "__main__":
   testutils.GanetiTestProgram()
