@@ -42,6 +42,7 @@ import shutil
 import os
 
 from ganeti.cmdlib import cluster
+from ganeti.cmdlib.cluster import verify
 from ganeti import constants
 from ganeti import errors
 from ganeti import netutils
@@ -59,31 +60,48 @@ import testutils
 
 class TestClusterVerifySsh(unittest.TestCase):
   def testMultipleGroups(self):
-    fn = cluster.LUClusterVerifyGroup._SelectSshCheckNodes
+    fn = verify.LUClusterVerifyGroup._SelectSshCheckNodes
     mygroupnodes = [
-      objects.Node(name="node20", group="my", offline=False),
-      objects.Node(name="node21", group="my", offline=False),
-      objects.Node(name="node22", group="my", offline=False),
-      objects.Node(name="node23", group="my", offline=False),
-      objects.Node(name="node24", group="my", offline=False),
-      objects.Node(name="node25", group="my", offline=False),
-      objects.Node(name="node26", group="my", offline=True),
+      objects.Node(name="node20", group="my", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node21", group="my", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node22", group="my", offline=False,
+                   master_candidate=False),
+      objects.Node(name="node23", group="my", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node24", group="my", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node25", group="my", offline=False,
+                   master_candidate=False),
+      objects.Node(name="node26", group="my", offline=True,
+                   master_candidate=True),
       ]
     nodes = [
-      objects.Node(name="node1", group="g1", offline=True),
-      objects.Node(name="node2", group="g1", offline=False),
-      objects.Node(name="node3", group="g1", offline=False),
-      objects.Node(name="node4", group="g1", offline=True),
-      objects.Node(name="node5", group="g1", offline=False),
-      objects.Node(name="node10", group="xyz", offline=False),
-      objects.Node(name="node11", group="xyz", offline=False),
-      objects.Node(name="node40", group="alloff", offline=True),
-      objects.Node(name="node41", group="alloff", offline=True),
-      objects.Node(name="node50", group="aaa", offline=False),
+      objects.Node(name="node1", group="g1", offline=True,
+                   master_candidate=True),
+      objects.Node(name="node2", group="g1", offline=False,
+                   master_candidate=False),
+      objects.Node(name="node3", group="g1", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node4", group="g1", offline=True,
+                   master_candidate=True),
+      objects.Node(name="node5", group="g1", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node10", group="xyz", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node11", group="xyz", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node40", group="alloff", offline=True,
+                   master_candidate=True),
+      objects.Node(name="node41", group="alloff", offline=True,
+                   master_candidate=True),
+      objects.Node(name="node50", group="aaa", offline=False,
+                   master_candidate=True),
       ] + mygroupnodes
     assert not utils.FindDuplicates(map(operator.attrgetter("name"), nodes))
 
-    (online, perhost) = fn(mygroupnodes, "my", nodes)
+    (online, perhost, _) = fn(mygroupnodes, "my", nodes)
     self.assertEqual(online, ["node%s" % i for i in range(20, 26)])
     self.assertEqual(set(perhost.keys()), set(online))
 
@@ -97,16 +115,20 @@ class TestClusterVerifySsh(unittest.TestCase):
       })
 
   def testSingleGroup(self):
-    fn = cluster.LUClusterVerifyGroup._SelectSshCheckNodes
+    fn = verify.LUClusterVerifyGroup._SelectSshCheckNodes
     nodes = [
-      objects.Node(name="node1", group="default", offline=True),
-      objects.Node(name="node2", group="default", offline=False),
-      objects.Node(name="node3", group="default", offline=False),
-      objects.Node(name="node4", group="default", offline=True),
+      objects.Node(name="node1", group="default", offline=True,
+                   master_candidate=True),
+      objects.Node(name="node2", group="default", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node3", group="default", offline=False,
+                   master_candidate=True),
+      objects.Node(name="node4", group="default", offline=True,
+                   master_candidate=True),
       ]
     assert not utils.FindDuplicates(map(operator.attrgetter("name"), nodes))
 
-    (online, perhost) = fn(nodes, "default", nodes)
+    (online, perhost, _) = fn(nodes, "default", nodes)
     self.assertEqual(online, ["node2", "node3"])
     self.assertEqual(set(perhost.keys()), set(online))
 
@@ -868,7 +890,7 @@ class TestLUClusterSetParams(CmdlibTestCase):
     op = opcodes.OpClusterSetParams(
            enabled_disk_templates=new_disk_templates,
            ipolicy={constants.IPOLICY_DTS: new_disk_templates})
-    self.ExecOpCodeExpectOpPrereqError(op, "least one instance using it")
+    self.ExecOpCodeExpectOpPrereqError(op, "least one disk using it")
 
   def testEnabledDiskTemplatesWithoutVgName(self):
     enabled_disk_templates = [constants.DT_PLAIN]
@@ -886,6 +908,14 @@ class TestLUClusterSetParams(CmdlibTestCase):
            ipolicy={constants.IPOLICY_DTS: enabled_disk_templates})
     self.ExecOpCodeExpectOpPrereqError(op, "Cannot disable disk template")
 
+  def testDisableDiskTemplateWithExistingInstanceDiskless(self):
+    self.cfg.AddNewInstance(disks=[])
+    enabled_disk_templates = [constants.DT_PLAIN]
+    op = opcodes.OpClusterSetParams(
+           enabled_disk_templates=enabled_disk_templates,
+           ipolicy={constants.IPOLICY_DTS: enabled_disk_templates})
+    self.ExecOpCodeExpectOpPrereqError(op, "Cannot disable disk template")
+
   def testVgNameNoLvmDiskTemplateEnabled(self):
     vg_name = "test_vg"
     self.cfg.SetEnabledDiskTemplates([constants.DT_DISKLESS])
@@ -1102,6 +1132,12 @@ class TestLUClusterVerifyConfig(CmdlibTestCase):
       "following instances have a non-existing primary-node")
     self.assertFalse(result)
 
+  def testDanglingDisk(self):
+    self.cfg.AddOrphanDisk()
+    op = opcodes.OpClusterVerifyConfig()
+    result = self.ExecOpCode(op)
+    self.assertTrue(result)
+
 
 class TestLUClusterVerifyGroup(CmdlibTestCase):
   def testEmptyNodeGroup(self):
@@ -1153,6 +1189,19 @@ class TestLUClusterVerifyGroup(CmdlibTestCase):
 
     self.ExecOpCode(op)
 
+  def testVerifyNodeDrbdSuccess(self):
+    ninfo = self.cfg.AddNewNode()
+    disk = self.cfg.CreateDisk(dev_type=constants.DT_DRBD8,
+                                primary_node=self.master,
+                                secondary_node=ninfo)
+    instance = self.cfg.AddNewInstance(disks=[disk])
+    instanceinfo = self.cfg.GetAllInstancesInfo()
+    disks_info = self.cfg.GetAllDisksInfo()
+    drbd_map = {ninfo.uuid: {0: disk.uuid}}
+    minors = verify.LUClusterVerifyGroup._ComputeDrbdMinors(
+      ninfo, instanceinfo, disks_info, drbd_map, lambda *args: None)
+    self.assertEquals(minors, {0: (disk.uuid, instance.uuid, False)})
+
 
 class TestLUClusterVerifyClientCerts(CmdlibTestCase):
 
@@ -1300,7 +1349,7 @@ class TestLUClusterVerifyGroupMethods(CmdlibTestCase):
     lu._exclusive_storage = False
     lu.master_node = self.master_uuid
     lu.group_info = self.group
-    cluster.LUClusterVerifyGroup.all_node_info = \
+    verify.LUClusterVerifyGroup.all_node_info = \
       property(fget=lambda _: self.cfg.GetAllNodesInfo())
 
 
@@ -1419,7 +1468,7 @@ class TestLUClusterVerifyGroupUpdateVerifyNodeLVM(
   def testValidNodeResultExclusiveStorage(self, lu):
     lu._exclusive_storage = True
     lu._UpdateVerifyNodeLVM(self.master, self.VALID_NRESULT, "mock_vg",
-                            cluster.LUClusterVerifyGroup.NodeImage())
+                            verify.LUClusterVerifyGroup.NodeImage())
     self.mcpu.assertLogIsEmpty()
 
 
@@ -1470,7 +1519,7 @@ class TestLUClusterVerifyGroupVerifyGroupLVM(TestLUClusterVerifyGroupMethods):
   @withLockedLU
   def testNoPvInfo(self, lu):
     lu._exclusive_storage = True
-    nimg = cluster.LUClusterVerifyGroup.NodeImage()
+    nimg = verify.LUClusterVerifyGroup.NodeImage()
     lu._VerifyGroupLVM({self.master.uuid: nimg}, "mock_vg")
     self.mcpu.assertLogIsEmpty()
 
@@ -1478,10 +1527,10 @@ class TestLUClusterVerifyGroupVerifyGroupLVM(TestLUClusterVerifyGroupMethods):
   def testValidPvInfos(self, lu):
     lu._exclusive_storage = True
     node2 = self.cfg.AddNewNode()
-    nimg1 = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master.uuid)
+    nimg1 = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master.uuid)
     nimg1.pv_min = 10000
     nimg1.pv_max = 10010
-    nimg2 = cluster.LUClusterVerifyGroup.NodeImage(uuid=node2.uuid)
+    nimg2 = verify.LUClusterVerifyGroup.NodeImage(uuid=node2.uuid)
     nimg2.pv_min = 9998
     nimg2.pv_max = 10005
     lu._VerifyGroupLVM({self.master.uuid: nimg1, node2.uuid: nimg2}, "mock_vg")
@@ -1597,7 +1646,7 @@ class TestLUClusterVerifyGroupVerifyInstance(TestLUClusterVerifyGroupMethods):
     self.diskless_inst = self.cfg.AddNewInstance(disks=[])
 
     self.master_img = \
-      cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+      verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
     self.master_img.volumes = ["/".join(disk.logical_id)
                                for inst in [self.running_inst,
                                             self.diskless_inst]
@@ -1608,7 +1657,7 @@ class TestLUClusterVerifyGroupVerifyInstance(TestLUClusterVerifyGroupMethods):
       ["/".join(disk.logical_id) for disk in drbd_inst_disks[0].children])
     self.master_img.instances = [self.running_inst.uuid]
     self.node1_img = \
-      cluster.LUClusterVerifyGroup.NodeImage(uuid=self.node1.uuid)
+      verify.LUClusterVerifyGroup.NodeImage(uuid=self.node1.uuid)
     self.node1_img.volumes = \
       ["/".join(disk.logical_id) for disk in drbd_inst_disks[0].children]
     self.node_imgs = {
@@ -1673,6 +1722,7 @@ class TestLUClusterVerifyGroupVerifyInstance(TestLUClusterVerifyGroupMethods):
 
   @withLockedLU
   def testNotOkayDiskStatus(self, lu):
+    self.diskstatus[self.master_uuid][0][1].is_degraded = True
     self.diskstatus[self.master_uuid][0][1].ldisk_status = constants.LDS_FAULTY
     lu._VerifyInstance(self.running_inst, self.node_imgs, self.diskstatus)
     self.mcpu.assertLogContainsRegex("instance .* state is 'faulty'")
@@ -1682,7 +1732,7 @@ class TestLUClusterVerifyGroupVerifyInstance(TestLUClusterVerifyGroupMethods):
     self.master.ndparams[constants.ND_EXCLUSIVE_STORAGE] = True
     lu._VerifyInstance(self.drbd_inst, self.node_imgs, self.diskstatus)
     self.mcpu.assertLogContainsRegex(
-      "instance has template drbd, which is not supported")
+        "disk types? drbd, which are not supported")
 
   @withLockedLU
   def testExclusiveStorageWithValidInstance(self, lu):
@@ -1713,7 +1763,7 @@ class TestLUClusterVerifyGroupVerifyOrphanVolumes(
         TestLUClusterVerifyGroupMethods):
   @withLockedLU
   def testOrphanedVolume(self, lu):
-    master_img = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    master_img = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
     master_img.volumes = [
       "mock_vg/disk_0",  # Required, present, no error
       "mock_vg/disk_1",  # Unknown, present, orphan
@@ -1751,14 +1801,14 @@ class TestLUClusterVerifyGroupVerifyNPlusOneMemory(
     inst2 = self.cfg.AddNewInstance()
     inst3 = self.cfg.AddNewInstance()
 
-    node1_img = cluster.LUClusterVerifyGroup.NodeImage(uuid=node1.uuid)
+    node1_img = verify.LUClusterVerifyGroup.NodeImage(uuid=node1.uuid)
     node1_img.sbp = {
       self.master_uuid: [inst1.uuid, inst2.uuid, inst3.uuid]
     }
 
-    node2_img = cluster.LUClusterVerifyGroup.NodeImage(uuid=node2.uuid)
+    node2_img = verify.LUClusterVerifyGroup.NodeImage(uuid=node2.uuid)
 
-    node3_img = cluster.LUClusterVerifyGroup.NodeImage(uuid=node3.uuid)
+    node3_img = verify.LUClusterVerifyGroup.NodeImage(uuid=node3.uuid)
     node3_img.offline = True
 
     node_imgs = {
@@ -1881,7 +1931,7 @@ class TestLUClusterVerifyGroupVerifyNodeOs(TestLUClusterVerifyGroupMethods):
     for ndata in [{}, {constants.NV_OSLIST: ""}, {constants.NV_OSLIST: [""]},
                   {constants.NV_OSLIST: [["1", "2"]]}]:
       self.mcpu.ClearLogMessages()
-      nimage = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+      nimage = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
       lu._UpdateNodeOS(self.master, ndata, nimage)
       self.mcpu.assertLogContainsRegex("node hasn't returned valid OS data")
 
@@ -1896,15 +1946,15 @@ class TestLUClusterVerifyGroupVerifyNodeOs(TestLUClusterVerifyGroupMethods):
          True]
       ]
     }
-    nimage = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    nimage = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
     lu._UpdateNodeOS(self.master, ndata, nimage)
     self.mcpu.assertLogIsEmpty()
 
   @withLockedLU
   def testVerifyNodeOs(self, lu):
     node = self.cfg.AddNewNode()
-    nimg_root = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
-    nimg = cluster.LUClusterVerifyGroup.NodeImage(uuid=node.uuid)
+    nimg_root = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    nimg = verify.LUClusterVerifyGroup.NodeImage(uuid=node.uuid)
 
     nimg_root.os_fail = False
     nimg_root.oslist = {
@@ -2032,7 +2082,7 @@ class TestLUClusterVerifyGroupUpdateNodeVolumes(
   TestLUClusterVerifyGroupMethods):
   def setUp(self):
     super(TestLUClusterVerifyGroupUpdateNodeVolumes, self).setUp()
-    self.nimg = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    self.nimg = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
 
   @withLockedLU
   def testNoVgName(self, lu):
@@ -2066,7 +2116,7 @@ class TestLUClusterVerifyGroupUpdateNodeInstances(
   TestLUClusterVerifyGroupMethods):
   def setUp(self):
     super(TestLUClusterVerifyGroupUpdateNodeInstances, self).setUp()
-    self.nimg = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    self.nimg = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
 
   @withLockedLU
   def testInvalidNodeResult(self, lu):
@@ -2085,7 +2135,7 @@ class TestLUClusterVerifyGroupUpdateNodeInstances(
 class TestLUClusterVerifyGroupUpdateNodeInfo(TestLUClusterVerifyGroupMethods):
   def setUp(self):
     super(TestLUClusterVerifyGroupUpdateNodeInfo, self).setUp()
-    self.nimg = cluster.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
+    self.nimg = verify.LUClusterVerifyGroup.NodeImage(uuid=self.master_uuid)
     self.valid_hvresult = {constants.NV_HVINFO: {"memory_free": 1024}}
 
   @withLockedLU
@@ -2155,15 +2205,15 @@ class TestLUClusterVerifyGroupCollectDiskInfo(TestLUClusterVerifyGroupMethods):
                               secondary_node=self.node2,
                               disk_template=constants.DT_DRBD8)
 
-    self.node1_img = cluster.LUClusterVerifyGroup.NodeImage(
+    self.node1_img = verify.LUClusterVerifyGroup.NodeImage(
                        uuid=self.node1.uuid)
     self.node1_img.pinst = [self.diskless_inst.uuid]
     self.node1_img.sinst = []
-    self.node2_img = cluster.LUClusterVerifyGroup.NodeImage(
+    self.node2_img = verify.LUClusterVerifyGroup.NodeImage(
                        uuid=self.node2.uuid)
     self.node2_img.pinst = [self.plain_inst.uuid]
     self.node2_img.sinst = [self.drbd_inst.uuid]
-    self.node3_img = cluster.LUClusterVerifyGroup.NodeImage(
+    self.node3_img = verify.LUClusterVerifyGroup.NodeImage(
                        uuid=self.node3.uuid)
     self.node3_img.pinst = [self.drbd_inst.uuid]
     self.node3_img.sinst = []
@@ -2340,7 +2390,7 @@ class TestLUClusterRenewCrypto(CmdlibTestCase):
       self.cfg.AddNewNode()
     self.rpc.call_node_crypto_tokens = self._CompletelySuccessfulRpc
 
-    op = opcodes.OpClusterRenewCrypto()
+    op = opcodes.OpClusterRenewCrypto(node_certificates=True)
     self.ExecOpCode(op)
 
     self._AssertCertFiles(pathutils)
@@ -2350,6 +2400,7 @@ class TestLUClusterRenewCrypto(CmdlibTestCase):
     self.assertEqual(num_nodes + 1, len(cluster.candidate_certs))
     nodes = self.cfg.GetAllNodesInfo()
     master_uuid = self.cfg.GetMasterNode()
+
     for (node_uuid, _) in nodes.items():
       if node_uuid == master_uuid:
         # The master digest is from the actual test certificate.
@@ -2386,7 +2437,7 @@ class TestLUClusterRenewCrypto(CmdlibTestCase):
                          if node_uuid != master_uuid][1]
     self.rpc.call_node_crypto_tokens = self._partiallyFailingRpc
 
-    op = opcodes.OpClusterRenewCrypto()
+    op = opcodes.OpClusterRenewCrypto(node_certificates=True)
     self.ExecOpCode(op)
 
     self._AssertCertFiles(pathutils)
@@ -2414,7 +2465,7 @@ class TestLUClusterRenewCrypto(CmdlibTestCase):
       self.cfg.AddNewNode(offline=(i==offline_index))
     self.rpc.call_node_crypto_tokens = self._CompletelySuccessfulRpc
 
-    op = opcodes.OpClusterRenewCrypto()
+    op = opcodes.OpClusterRenewCrypto(node_certificates=True)
     self.ExecOpCode(op)
 
     self._AssertCertFiles(pathutils)
@@ -2461,7 +2512,7 @@ class TestLUClusterRenewCrypto(CmdlibTestCase):
     # Add one non-master node
     self.cfg.AddNewNode()
 
-    op = opcodes.OpClusterRenewCrypto()
+    op = opcodes.OpClusterRenewCrypto(node_certificates=True)
     self.ExecOpCode(op)
 
     self._AssertCertFiles(pathutils)
