@@ -30,14 +30,15 @@
 
 """Script for testing ganeti.utils.process"""
 
-import unittest
-import tempfile
-import shutil
 import os
-import stat
-import time
 import select
+import shutil
 import signal
+import stat
+import subprocess
+import tempfile
+import time
+import unittest
 
 from ganeti import constants
 from ganeti import utils
@@ -268,6 +269,16 @@ class TestRunCmd(testutils.GanetiTestCase):
     self.assertEqual(result.signal, 15)
     self.assertEqual(result.output, "")
 
+  def testTimeoutFlagTrue(self):
+    result = utils.RunCmd(["sleep", "2"], timeout=0.1)
+    self.assertTrue(result.failed)
+    self.assertTrue(result.failed_by_timeout)
+
+  def testTimeoutFlagFalse(self):
+    result = utils.RunCmd(["false"], timeout=5)
+    self.assertTrue(result.failed)
+    self.assertFalse(result.failed_by_timeout)
+
   def testTimeoutClean(self):
     cmd = ("trap 'exit 0' TERM; echo >&%d; read < %s" %
            (self.proc_ready_helper.write_fd, self.fifo_file))
@@ -732,5 +743,15 @@ class RunInSeparateProcess(unittest.TestCase):
                       utils.RunInSeparateProcess, _exc)
 
 
+class GetCmdline(unittest.TestCase):
+  def test(self):
+    sample_cmd = "sleep 2; true"
+    pid = subprocess.Popen(sample_cmd, shell=True).pid
+    cmdline = utils.GetProcCmdline(pid)
+    # As the popen will quote and pass on the sample_cmd, it should be returned
+    # by the function as an element in the list of arguments
+    self.assertTrue(sample_cmd in cmdline)
+
+
 if __name__ == "__main__":
   testutils.GanetiTestProgram()
